/***************************************************************************
                          integrator.h  -  description
                             -------------------
    begin                : Mon Mar 19 2001
    copyright            : (C) 2001 by Ivan Martinez
    email                : ivanmr@altavista.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef INTEGRATOR_H
#define INTEGRATOR_H

#include "fixedoperationblock.h"

/**
  *@author Ivan Martinez
  */

#include "fixedinput.h"
#include "fixedoutput.h"
#include "blockvariable.h"

/**
  * CIntegrator implements discrete integration with an IIR filter:
  * State(t) = StateFactor * State(t-1) + InputFactor * Input(t)
  * Output(t) = State(t)
  */
class CIntegrator : public CFixedOperationBlock  {
public:
	/** Type of the factors */
	enum EFactorType {CONSTANT, VARIABLE};
	
	/** Constructor */
	CIntegrator();
	
	/** Destructor */
	~CIntegrator();
	
	/** Returns "CGain" */
	virtual string Class() const;
	
	/** Returns the input factor when it's a constant */
	float ConstantInputFactor() const;
	
	/** Returns the state factor when it's a constant */
	float ConstantStateFactor() const;
	
	/** Returns TRUE if the argument is "CGain" or one of its
		base classes, otherwise FALSE */
	virtual bool DescendantOf(const string&) const;
	
	/** Returns the initial state value */
	float InitialState() const;
	
	/** Returns the input */
	CFixedInput* Input() const;
	
	/** Returns the output */
	CFixedOutput* Output() const;

	/** Sets the input factor as a constant */
	bool SetConstantInputFactor(float);

	/** Sets the state factor as a constant */
	bool SetConstantStateFactor(float);

	/** Sets the initial state value */
	bool SetInitialState(float);
	
	/** Sets the input factor as a variable */
	bool SetVariableInputFactor(const string&);
	
	/** Sets the state factor as a variable */
	bool SetVariableStateFactor(const string&);
	
	/** Returns the input factor when it's a variable */
	const string VariableInputFactor() const;

	/** Returns the state factor when it's a variable */
	const string VariableStateFactor() const;
protected: // Protected methods
  /** If IllegalCIdentifier(Name()) generates the error 500.
		If not,	sets the identifier */
  virtual void Arrange(void *);
	
	/** Generates the function that executes the block */
	virtual void GenerateFunctions(ostream&, void* = NULL) const;

  /** Return the variable that stores the state */
  CBlockVariable* StateVariable() const;
private: // Private attributes
	/** Type of the input factor */
	EFactorType m_InputFactorType;
	
  /** Input factor when it's a constant */
  float m_rConstantInputFactor;

  /** State factor when it's a constant */
  float m_rConstantStateFactor;

	/** Initial state value */
	float m_rInitialState;
	
	/** Type of the state factor */
	EFactorType m_StateFactorType;

  /** Input factor when it's a variable */
  string m_VariableInputFactor;

  /** State factor when it's a variable */
  string m_VariableStateFactor;
};

#endif
