/***************************************************************************
                          codegenerator.h  -  description
                             -------------------
    begin                : Fri May 12 2000
    copyright            : (C) 2000 by Ivan Martinez
    email                : ivanmr@altavista.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CODEGENERATOR_H
#define CODEGENERATOR_H

/**
  *@author Ivan Martinez
  */

#include <string>
#include <iostream>
#include <set>
#include <codegeneratorpointers.h>
#include <errors.h>

class CCodeGenerator;

class CCodeGenerator {
public:
	/** Set of strings type */
	typedef set<string,less<string> > CStringsSet;

	/** Constructor */
	CCodeGenerator();

	/** Destructor */
	virtual ~CCodeGenerator();

	/** If the parameter is a suitable container,
		sets it as the container of this component and returns true.
		If it isn't, generates the error 101 and retuns false */
	virtual bool AcceptsContainer(const CCodeGenerator*) = 0;

	/** Adds a component to this container. Returns if succeeded */
	bool AddComponent(CCodeGenerator*);

	/** Calls Arrange, then AllArrange in the contained components and
		then ComponentsDependingArrange. Returns if the calls succeeded.
		Stops when a call fails */
	bool AllArrange(void* = NULL);

	/** Calls AllReportErrors in the contained components with the same
		argument, then adds the errors report of this component to the
	argument, and finally clears the errors report of this component */
	void AllReportErrors(CErrors&);

	/** Returns the name of the class */
	virtual string Class() const = 0;

	/** Returns the non-fixed contained components */
	const CCodeGeneratorPointers& Components() const;

	/** Calls AllGenerateCode(ostream&...) in the contained components */
	void ComponentsGenerateCode(ostream&, const CCodeGenerator*, const string&,
		void* = NULL) const;

	/** Calls AllGenerateCode(CStringsSet&...) in the contained components */
	void ComponentsGenerateCode(CStringsSet&, const CCodeGenerator*,
		const string&, void* = NULL) const;

	/** Returns the container of this component */
	const CCodeGenerator* Container() const;

	/** Calls DeleteObjects in the contained components and then deletes
		them */
	void DeleteObjects();

	/** Returns if the argument is "CCodeGenerator" */
	virtual bool DescendantOf(const string&) const;

	/** Returns if the errors report of this component is not empty */
	bool ErrorsReported() const;

	/** Returns the errors report of this component */
	const CErrors& Errors() const;

	/** Returns the fixed contained components */
	const CCodeGeneratorPointers& FixedComponents() const;

	/** Calls ComponentsGenerateCode(ostream&...) and then
		GenerateCode(ostream&...) */
	void AllGenerateCode(ostream&, const CCodeGenerator*, const string&,
		void* = NULL) const;

	/** Calls ComponentsGenerateCode(CStringsSet&...) and then
		GenerateCode(CStringsSet&...) */
	void AllGenerateCode(CStringsSet&, const CCodeGenerator*,
		const string&, void* = NULL) const;

	/** Returns the identifier */
	const string& Identifier() const;

	/** Calls RepeatedValue with the same arguments in the contained
		components. Returns if one of the calls returned true */
	virtual bool RepeatedValue(const CCodeGenerator*, const string&,
		const string&) const;

	/** Clears the argument, calls AllReportErrors with it, and returns
		if it contais errors after the call */
	bool ReportErrors(CErrors&);
protected:
	/** Adds an error in this component to the errors report */
	void AddError(unsigned int);

	/** Adds an error in any component to the errors report */
	void AddError(unsigned int, const CCodeGenerator*);

  /** Adds a fixed component to this container. Returns if succeeded */
  void AddFixedComponent(CCodeGenerator*);
	
	/** Performs the updating of the internal values that do not depend
		on the contained components */
	virtual void Arrange(void* = NULL);

	/** Returns the non-fixed contained components */
	CCodeGeneratorPointers& Components();

	/** Performs the updating of the internal values that depend
		on the contained components */
	virtual void ComponentsDependingArrange(void* = NULL);

	/** Calls AllGenerateCode(ostream&...) with the same arguments
		in the contained components */
	void ComponentsGenerateCode(ostream&, const string&, void* = NULL) const;

	/** Calls AllGenerateCode(CStringsSet&...) with the same arguments
		in the contained components */
	void ComponentsGenerateCode(CStringsSet&, const string&,
		void* = NULL) const;

	/** Returns the errors report of this component */
	CErrors& Errors();

	/** Returns the fixed contained components */
	CCodeGeneratorPointers& FixedComponents();

	/** Generates code for the section of the component indicated in the
		arguments */
	virtual void GenerateCode(ostream&, const CCodeGenerator*, const string&,
		void* = NULL) const = 0;

	/** Generates code for the section of the component indicated in the
		arguments. Lines repeated in the same section are eliminated */
	virtual void GenerateCode(CStringsSet&, const CCodeGenerator*,
		const string&, void* = NULL) const;

	/** Returns if the argument is an illegal C identifier */
	bool IllegalCIdentifier(const string&) const;

	/** Sets the container of this component */
	void SetContainer(const CCodeGenerator*);
	
	/** Sets the identifier */
	void SetIdentifier(const string&);
private: // Private attributes
	/** Identifier */
	string m_Identifier;

	/** Non-fixed contained components */
	CCodeGeneratorPointers m_Components;

	/** Fixed contained components */
	CCodeGeneratorPointers m_FixedComponents;

	/** Container */
	const CCodeGenerator* m_pContainer;

	/** Errors report */
	CErrors m_Errors;
};

#endif
