/***************************************************************************
                          integrator.cpp  -  description
                             -------------------
    begin                : Mon Mar 19 2001
    copyright            : (C) 2001 by Ivan Martinez
    email                : ivanmr@altavista.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <integrator.h>

#include <loop.h>
#include <dynamicsystem.h>
#include <strstream>

CIntegrator::CIntegrator()
{
	CFixedInput* pInput;
	CFixedOutput* pOutput;
	CBlockVariable* pState;
	
	pInput = new CFixedInput(this);
	pInput->SetName("input");
	AddFixedComponent(pInput);
	
	pOutput = new CFixedOutput(this);
	pOutput->SetName("output");
	AddFixedComponent(pOutput);
	
	pState = new CBlockVariable();
	pState->AcceptsContainer(this);
	pState->SetName("state");
	AddFixedComponent(pState);
	
	m_InputFactorType = CONSTANT;
	m_rConstantInputFactor = 1;
	m_VariableInputFactor = "";
	m_StateFactorType = CONSTANT;
	m_rConstantStateFactor = 1;
	m_VariableStateFactor = "";
}

CIntegrator::~CIntegrator()
{
}

void CIntegrator::Arrange(void *pExtraParameter)
{
	strstream Initialization;
	
	if (IllegalCIdentifier(Name()))
		AddError(500);
	else
		if (IllegalCIdentifier(IOsType()))
			AddError(3400);
		else
		{
			SetIdentifier("block_" + Container()->Identifier() + "_" + Name());
			StateVariable()->SetType(IOsType());
			Initialization << "  state = " << InitialState() << ";" << ends;
			StateVariable()->SetInitialization(Initialization.str());
		}
}

string CIntegrator::Class() const
{
	return "CIntegrator";
}

float CIntegrator::ConstantInputFactor() const
{
	if (m_InputFactorType == CONSTANT)
		return m_rConstantInputFactor;
	else
		return 0;
}

float CIntegrator::ConstantStateFactor() const
{
	if (m_StateFactorType == CONSTANT)
		return m_rConstantStateFactor;
	else
		return 0;
}

bool CIntegrator::DescendantOf(const string& ClassName) const
{
	if (ClassName == "CIntegrator")
		return true;
	else
		return CFixedOperationBlock::DescendantOf(ClassName);
}

void CIntegrator::GenerateFunctions(ostream& Stream,
	void* pExtraParameter = NULL) const
{
	CStringsSet StringsSet;

	Stream << "void " << Identifier() << "(void)" <<	endl;
	Stream << "{" << endl;

	if ((m_InputFactorType == VARIABLE) || (m_StateFactorType == VARIABLE))
	{
		StringsSet.clear();
		Container()->Container()->ComponentsGenerateCode(StringsSet, this,
			"flags declaration", pExtraParameter);
		copy(StringsSet.begin(), StringsSet.end(),
			ostream_iterator<string>(Stream, "\n"));
		Container()->Container()->ComponentsGenerateCode(Stream, this,
			"local system variable declarations", pExtraParameter);
		Container()->ComponentsGenerateCode(Stream, this,
			"local loop variable declarations", pExtraParameter);
			
		Container()->Container()->ComponentsGenerateCode(Stream, this,
			"system variables reading", pExtraParameter);
		Container()->ComponentsGenerateCode(Stream, this,
			"loop variables reading", pExtraParameter);
	}

	Stream << "  " << StateVariable()->GlobalIdentifier() << " = ";
	if (m_StateFactorType == CONSTANT)
		Stream << ConstantStateFactor();
	else
		Stream << VariableStateFactor();
	Stream << " * " << StateVariable()->GlobalIdentifier() << " + ";
	if (m_InputFactorType == CONSTANT)
		Stream << ConstantInputFactor();
	else
		Stream << VariableInputFactor();
	Stream << " * " << Input()->GlobalIdentifier() << ";" << endl;
	Stream << "  " << Output()->GlobalIdentifier() << " = " <<
		StateVariable()->GlobalIdentifier() << ";" << endl;
		
	if ((m_InputFactorType == VARIABLE) || (m_StateFactorType == VARIABLE))
	{
		Container()->ComponentsGenerateCode(Stream, this,
			"loop variables writing", pExtraParameter);
		Container()->Container()->ComponentsGenerateCode(Stream, this,
			"system variables writing", pExtraParameter);
	}

	Stream << "}" << endl;
}

float CIntegrator::InitialState() const
{
	return m_rInitialState;
}

CFixedInput* CIntegrator::Input() const
{
	return (CFixedInput*) *(FixedComponents().Begin());
}

CFixedOutput* CIntegrator::Output() const
{
	return (CFixedOutput*) FixedComponents()[1];
}

bool CIntegrator::SetConstantInputFactor(float rFactor)
{
	m_InputFactorType = CONSTANT;
	m_rConstantInputFactor = rFactor;
	return true;
}

bool CIntegrator::SetConstantStateFactor(float rFactor)
{
	m_StateFactorType = CONSTANT;
	m_rConstantStateFactor = rFactor;
	return true;
}

bool CIntegrator::SetInitialState(float State)
{
	m_rInitialState = State;
}

bool CIntegrator::SetVariableInputFactor(const string& Factor)
{
	if (IllegalCIdentifier(Factor))
	{
		AddError(3700);
		return false;
	}	
	else
	{
		m_InputFactorType = VARIABLE;
		m_VariableInputFactor = Factor;
		return true;
	}
}

bool CIntegrator::SetVariableStateFactor(const string& Factor)
{
	if (IllegalCIdentifier(Factor))
	{
		AddError(3701);
		return false;
	}	
	else
	{
		m_StateFactorType = VARIABLE;
		m_VariableStateFactor = Factor;
		return true;
	}
}

CBlockVariable* CIntegrator::StateVariable() const
{
	return (CBlockVariable*) FixedComponents()[2];
}

const string CIntegrator::VariableInputFactor() const
{
	if (m_InputFactorType == CONSTANT)
		return m_VariableInputFactor;
	else
		return "";
}

const string CIntegrator::VariableStateFactor() const
{
	if (m_StateFactorType == CONSTANT)
		return m_VariableStateFactor;
	else
		return "";
}