/***************************************************************************
                          loggingfifo.cpp  -  description
                             -------------------
    begin                : Tue May 23 2000
    copyright            : (C) 2000 by Ivan Martinez
    email                : ivanmr@altavista.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <loggingfifo.h>

#include <dualvariable.h>
#include <strstream>
#include <systemvariable.h>

CLoggingFifo::CLoggingFifo()
{
	SetSamplingRate(1);
}

CLoggingFifo::~CLoggingFifo()
{
}

void CLoggingFifo::Arrange(void* ExtraParameterL = NULL)
{
  strstream ConversionStreamL;

	SetMode(CRTFifo::WRITE);

	ConversionStreamL << Number() << ends;

	SetIdentifier(string("rtf") + ConversionStreamL.str() + "_" +
		Container()->Identifier());
	SetSamplingCounterIdentifier("samp_count_" +
		Container()->GlobalIdentifier());
	SetBuffer("&" + Container()->GlobalIdentifier());
	SetBufferSize("sizeof(" + Container()->GlobalIdentifier() + ")");
}

string CLoggingFifo::Class() const
{
	return "CLoggingFifo";
}

bool CLoggingFifo::DescendantOf(const string& ClassNameL) const
{
	if (ClassNameL == "CLoggingFifo")
		return true;
	else
		return CVariableFifo::DescendantOf(ClassNameL);
}

const CDualVariable* CLoggingFifo::Container() const
{
	return (CDualVariable*)CRTFifo::Container();
}

void CLoggingFifo::GenerateCode(ostream& StreamL,
	const CCodeGenerator* CallerPL,	const string& SectionL,
	void* ExtraParameterL = NULL) const
{
	if (CallerPL->DescendantOf("CDynamicSystem"))
	{
		if (SectionL == "global variables declaration")
			GenerateSamplingCounterDeclaration(StreamL, ExtraParameterL);
		else	
			if (SectionL == "module I/Os declaration")
				GenerateDeclaration(StreamL, ExtraParameterL);
			else
				if (SectionL == "global variables initialization")
					GenerateSamplingCounterInitialization(StreamL, ExtraParameterL);
				else
					if (SectionL == "module I/Os initialization")
						GenerateInitialization(StreamL, ExtraParameterL);
					else
						if (SectionL == "module I/Os destruction")
							GenerateDestruction(StreamL, ExtraParameterL);	
	}
	else
		if (CallerPL->DescendantOf("CLoop"))
			if (SectionL == "variables log")
				GenerateFifoWriting(StreamL, ExtraParameterL);
			else
				if (SectionL == "global variables log")
					GenerateLockedFifoWriting(StreamL, ExtraParameterL);
}

void CLoggingFifo::GenerateLockedFifoWriting(ostream& StreamL,
	void* ExtraParameterL = NULL) const
{
	if (Container()->Class() == "CSystemVariable")
		if (SamplingRate() == 1)
		{
			StreamL << "      rtl_spin_lock_irqsave(&" <<
				((CSystemVariable*)Container())->LockIdentifier() << ", flags);" <<
				endl;
			StreamL << "      write(" << Identifier() << ", " << Buffer() << ", " <<
				BufferSize() << ");" << endl;
			StreamL << "      rtl_spin_unlock_irqrestore(&" <<
				((CSystemVariable*)Container())->LockIdentifier() << ", flags);" <<
				endl;
		}
		else
		{
			StreamL << "      " << SamplingCounterIdentifier() << " += " <<
				SamplingRate() << ";" << endl;
			StreamL << "      if (" << SamplingCounterIdentifier() << " >= 1)" <<
				endl;
			StreamL << "      {" << endl;
			StreamL << "        rtl_spin_lock_irqsave(&" <<
				((CSystemVariable*)Container())->LockIdentifier() << ", flags);" <<
				endl;
			StreamL << "        write(" << Identifier() << ", " << Buffer() <<
				", " << BufferSize() << ");" << endl;
			StreamL << "        rtl_spin_unlock_irqrestore(&" <<
				((CSystemVariable*)Container())->LockIdentifier() << ", flags);" <<
				endl;
			StreamL << "        " << SamplingCounterIdentifier() << " -= 1;" <<
				endl;
			StreamL << "      }" << endl;
		}
}

bool CLoggingFifo::SetNumber(unsigned int uNumber)
{
  strstream Stream;

  Stream << uNumber << ends;
	CVariableFifo::SetNumber(Stream.str());

	if (uNumber >= 0)
		return true;
	else
	{
		AddError(400);
		return false;
	}
}

unsigned int CLoggingFifo::Number() const
{
	strstream Stream;
	long long Number;

	Stream << CVariableFifo::Number();
	Stream >> Number;
	return Number;
}

bool CLoggingFifo::SetSize(int iSize)
{
  strstream Stream;

  Stream << iSize << ends;
	CVariableFifo::SetSize(Stream.str());

	if (iSize > 0)
		return true;
	else
	{
		AddError(401);
		return false;
	}
}

int CLoggingFifo::Size() const
{
	strstream Stream;
	int iSize;

	Stream << CVariableFifo::Size();
	Stream >> iSize;
	return iSize;
}

bool CLoggingFifo::AcceptsContainer(const CCodeGenerator* pContainer)
{
	if (pContainer->DescendantOf("CDualVariable"))
	{
		if (pContainer->Components().DescendantsOf("CLoggingFifo"))
		{
			AddError(1900);
			return false;
		}
		else
		{
			SetContainer(pContainer);
			return true;
		}
	}
	else
	{
		AddError(101);
		return false;
	}
}

bool CLoggingFifo::SetSamplingRate(float rRate)
{
	if ((rRate >= 0) && (rRate <= 1))
	{
		m_rSamplingRate = rRate;
		return true;
	}
	else
	{
		AddError(1901);
		return false;
	}	
}

float CLoggingFifo::SamplingRate() const
{
	return m_rSamplingRate;
}

void CLoggingFifo::GenerateSamplingCounterDeclaration(ostream& Stream,
	void* pExtraParameter) const
{
	if (SamplingRate() < 1)
		Stream << "float " << SamplingCounterIdentifier() << ";" << endl;
}

const string& CLoggingFifo::SamplingCounterIdentifier() const
{
	return m_SamplingCounterIdentifier;
}

void CLoggingFifo::SetSamplingCounterIdentifier(const string& Identifier)
{
	m_SamplingCounterIdentifier = Identifier;
}
/**  */
void CLoggingFifo::GenerateSamplingCounterInitialization(ostream& Stream,
	void* pExtraParameter = NULL) const
{
	if (SamplingRate() < 1)
		Stream << "  " << SamplingCounterIdentifier() << " = 0;" << endl;
}

void CLoggingFifo::GenerateFifoWriting(ostream& Stream,
	void *pExtraParameter = NULL) const
{
	if (SamplingRate() == 1)
		Stream << "      write(" << Identifier() << ", " << Buffer() << ", " <<
			BufferSize() << ");" << endl;
	else
	{
		Stream << "      " << SamplingCounterIdentifier() << " += " <<
			SamplingRate() << ";" << endl;
		Stream << "      if (" << SamplingCounterIdentifier() << " >= 1)" <<
			endl;
		Stream << "      {" << endl;
		Stream << "        write(" << Identifier() << ", " << Buffer() << ", " <<
			BufferSize() << ");" << endl;
		Stream << "        " << SamplingCounterIdentifier() << " -= 1;" <<
			endl;
		Stream << "      }" << endl;
	}
}
