/***************************************************************************
                          loop.h  -  description
                             -------------------
    begin                : Thu Feb 10 2000
    copyright            : (C) 2000 by Ivan Martinez
    email                : ivanmr@altavista.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef LOOP_H
#define LOOP_H

#include <rttask.h>

/**
  *@author Ivan Martinez
  */

#include <function.h>
#include <time.h>
#include <loopvariable.h>
#include <iostream>
#include <reportingfifo.h>

class CDynamicSystem;

class CLoop : public CRTTask {
public:
	/** Constructor */
	CLoop();

	/** Destructor */
	~CLoop();

	/** If the paramater is DescendantOf("CDynamicSystem"),
		sets it as the container of this component and returns true.
		If it isn't, generates the error 101 and retuns false */
	virtual bool AcceptsContainer(const CCodeGenerator*);

	/** Returns "CLoop" */
	virtual string Class() const;

	/** Returns the container of this component */
	const CDynamicSystem* Container() const;

	/** Returns TRUE if the argument is "CLoop" or one of its base
		classes, otherwise FALSE */
	virtual bool DescendantOf(const string&) const;

	/** Returns the IRQ number when the loop is triggered by an interrupt */
	unsigned int IRQNumber() const;

	/** Returns the name of this component */
	const string& Name() const;

	/** Returns the period in nanoseconds when the loop is periodic */
	long long Period() const;

	/** Returns the priority of the loop */
	unsigned int Priority() const;

	/** If the first argument is not this component, the second is
		"loop name"	and the third is Identifier(), returns TRUE meaning that
		there is a repetition of names between loops. If not, calls the
		function with the same arguments in the contained components and
		returns if one of the calls returned TRUE. */
	virtual bool RepeatedValue(const CCodeGenerator*, const string&,
		const string&) const;

	/** Sets the IRQ number when the loop is triggered by an interrupt */
	bool SetIRQNumber(unsigned int);

	/** Sets the name of this component */
	bool SetName(const string&);

	/** Sets the period in nanoseconds when the loop is periodic */
	bool SetPeriod(long long);

	/** Sets the priority */
	bool SetPriority(unsigned int);

	/** Sets the scheduling type */
	bool SetScheduling(eScheduling);
protected:
	/** If IllegalCIdentifier(Identifier()) generates the error 100.
		If not, sets the identifier of the function that executes the loop,
		its argument, the identifier of "pthread_attr_t", the	identifier of
		"sched_param", and the starting time of the execution of the loop */
	virtual void Arrange(void* = NULL);

	/** If there is a repetition of names between loops, generates the
		error 1000. Tries to ordenate the blocks of the loop for execution. If
		there is a redundant initialization of an input, generates the
		error 1401. If it cannot ordenate all the blocks, generates the error
		502. */
	virtual void ComponentsDependingArrange(void* = 0);

	/** Calls a code generation function depending on the arguments */
	virtual void GenerateCode(ostream&, const CCodeGenerator*, const string&,
		void* = NULL) const;

	/** Calls a code generation function depending on the arguments.
		Lines repeated in the same section are eliminated */
	virtual void GenerateCode(CStringsSet&, const CCodeGenerator*,
		const string&, void* = NULL) const;

	/** Generates the function that implements the execution of the loop */
	virtual void GenerateFunctions(ostream&, void* = NULL) const;

	/** Generates the call to the initialization function */
	virtual void GenerateInitializationCall(ostream&, void* = NULL) const;

	/** Generates the function that performs the variables initialization */
	virtual void GenerateInitializationFunctions(ostream&, void* = NULL) const;

	/** Generates the suspension of the execution of the loop it is
		triggered by an interrupt */
	virtual void GenerateSuspend(ostream&, void* = NULL) const;
};

#endif
