/***************************************************************************
                          rttask.h  -  description
                             -------------------
    begin                : Fri May 12 2000
    copyright            : (C) 2000 by Ivan Martinez
    email                : ivanmr@altavista.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This library is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef RTTASK_H
#define RTTASK_H

#include <codegenerator.h>

/**
  *@author Ivan Martinez
  */

class CRTModule;

class CRTTask : public CCodeGenerator {
public:
	/** Scheduling types */
	enum eScheduling {PERIOD, INTERRUPT};

	/** Constructor */
	CRTTask();

	/** Destructor */
	~CRTTask();

	/** Returns the container of this component */
	const CRTModule* Container() const;

	/** Returns if the argument is "CRTTask" or "CCodeGenerator" */
	virtual bool DescendantOf(const string&) const;

	/** Returns if floating point operations are allowed in the task */
	const string& FloatingPoint() const;

	/** Returns the argument for the function executed in realtime */
	const string& FunctionArgument() const;

	/** Returns the identifier of the function executed in realtime */
	const string& FunctionIdentifier() const;

	/** Returns the parameter of the function executed in realtime */
	const string& FunctionParameterIdentifier() const;

	/** Returns the IRQ number when the task is an interrupt handler */
	const string& IRQNumber() const;

	/** Returns the period in nanoseconds when the task is periodic */
	const string& Period() const;

	/** Returns the priority of the realtime task */
	const string& Priority() const;

	/** Returns the identifier of pthread_attr_t */
	const string& PThreadAttrIdentifier() const;

	/** Returns the identifier of the sched_param structure */
	const string& SchedParamIdentifier() const;

	/** Returns the scheduling type */
	enum eScheduling Scheduling() const;

	/** Returns the execution starting time in nanoseconds */
	const string& StartTime() const;
protected:
	/** Generates realtime thread destruction */
	virtual void GenerateDestruction(ostream&, void* = NULL) const;

	/** Generates includes declarations */
	virtual void GenerateIncludes(CStringsSet&, void* = NULL) const;

	/** Generates realtime thread initialization */
	virtual void GenerateInitialization(ostream&, void* = NULL) const;

	/** Generates declaration of initialization variables */
	virtual void GenerateInitializationVariablesDeclaration(CStringsSet&,
		void* = NULL) const;

	/** Generates realtime thread suspension */
	virtual void GenerateSuspend(ostream&, void* = NULL) const;

	/** Generates realtime thread declaration */
	virtual void GenerateThreadDeclaration(ostream&, void* = NULL) const;

	/** Generates realtime thread resumption */
	virtual void GenerateWakeUp(ostream&, void* = NULL) const;

	/** Sets floating point operations in the task */
	void SetFloatingPoint(const string&);

	/** Sets the argument for the function executed in realtime */
	void SetFunctionArgument(const string&);

	/** Sets the identifier of the function  */
	void SetFunctionIdentifier(const string&);

	/** Parameter of the function executed in realtime */
	void SetFunctionParameterIdentifier(const string&);

	/** Sets the IRQ number when the task is an interrupt handler */
	void SetIRQNumber(const string&);

	/** Sets the period in nanosecods when the task is periodic */
	void SetPeriod(const string&);

	/** Sets the priority of the realtime task */
	void SetPriority(const string&);

	/** Sets the identifier of pthread_attr_t */
	void SetPThreadAttrIdentifier(const string&);

	/** Sets the identifier of sched_param structure */
	void SetSchedParamIdentifier(const string&);

	/** Sets the scheduling type */
	void SetScheduling(eScheduling);

	/** Sets the execution starting time in nanoseconds */
	void SetStartTime(const string& StartTimeL);
private: // Private attributes
	/** pthread_attr_t identifier */
	string m_PThreadAttrIdentifier;

	/** sched_param structure identifier */
	string m_SchedParamIdentifier;

	/** Priority of the realtime task */
	string m_Priority;

	/** Identifier of the function executed in realtime */
	string m_FunctionIdentifier;

	/** Argument for the function executed in realtime */
	string m_FunctionArgument;

	/** Period when the task is periodic */
	string m_Period;

	/** Execution starting time */
	string m_StartTime;

	/** Parameter of the function executed in realtime */
	string m_FunctionParameterIdentifier;

	/** Scheduling type */
	eScheduling m_Scheduling;

	/** Returns if floating point operations are allowed in the task */
	string m_FloatingPoint;

	/** Returns the IRQ number when the task is an interrupt handler */
	string m_IRQNumber;
};

#endif
