/*
 *  linux_ipc.c -- IPC primitives for communication with Real-Time Linux tasks
 *
 *  Copyright (C) 1997 Jerry Epplin.  All rights reserved.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  History:
 *   15-Aug-97 jhe  V0.1 Original.  RT-FIFO enhancements.
 */

#include <stdio.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/fs.h>
#include <asm/rt_time.h>
#include "linux_ipc.h"

typedef struct
{
  int fd;
} RT_IPC_FIFO;		/* structure holding rt_ipc rt-fifo-specific data */

static RT_IPC_FIFO ipc_fifos[IPC_RTF_NO];	/* rt_ipc-specific fifo data */

static const char dev_root_name[] = "/dev/rtf";

/*************************************************************************
 * rtf_ipc_create -- create an rt_ipc rt-fifo
 *
 * Creates the rt_ipc rt-fifo 'fifo'.  This has the same capabilities as
 * the standard RT-Linux rt-fifos but with a cleaner API.  'size' is the size
 * of the fifo in bytes.  If 'rtl_to_linux' is 1, the fifo is used for
 * transferring data from RT-Linux tasks to a Linux process.  If 'rtl_to_linux'
 * is 0, the fifo is used for transferring data from a Linux process to
 * one or more RT-Linux tasks.  Note that the rt_ipc rtf_ipc_create() call
 * should be made before this call, as otherwise this rtf_ipc_create() will
 * return -EINVAL.
 *
 * Returns 0 if successful.  If the return value is less than zero, it is
 * -ENODEV if 'fifo' is greater than or equal to IPC_RTF_NO or RTF_NO, or
 * it is -EINVAL if the corresponding rt_ipc rtf_ipc_create() has not been
 * called first, or it is -EBUSY if the fifo is already open, or it is
 * the negative of the error code result of a call to mknod(2).
 * See the man pages for that function for the error values.
 *************************************************************************/
int rtf_ipc_create(unsigned int fifo, int size, int rtl_to_linux)
{
  int ret = 0;
  char dev_name[sizeof(dev_root_name) + 2];
  if (fifo >= IPC_RTF_NO)
    ret = -ENODEV;
  else
  {
    sprintf(dev_name, "%s%d", dev_root_name, fifo);
    unlink(dev_name);	/* remove it if it already exists */
    if (mknod(dev_name, S_IFCHR|0644, MKDEV(63, fifo)) == -1)
      ret = -errno;
    else
      if ((ipc_fifos[fifo].fd = open(dev_name, rtl_to_linux ? O_RDONLY : O_WRONLY)) == -1)
        ret = -errno;
  }
  return ret;
}

/*************************************************************************
 * rtf_ipc_destroy -- destroy an rt_ipc rt-fifo
 *
 * Removes the rt_ipc rt-fifo 'fifo' previously created with rt_ipc_create().
 *
 * Returns 0 if successful.  If the return value is less than zero, it is
 * the negative of the error code result of a call to close(2) or unlink(2).
 * See the man pages for those functions for the error values.
 *************************************************************************/
int rtf_ipc_destroy(unsigned int fifo)
{
  int ret = 0;
  char dev_name[sizeof(dev_root_name) + 2];
  sprintf(dev_name, "%s%d", dev_root_name, fifo);
  if ((ret = close(ipc_fifos[fifo].fd)) == -1)
    ret = -errno;
  else
    if ((ret = unlink(dev_name)) == -1)
      ret = -errno;
  return ret;
}

/*************************************************************************
 * rtf_receive -- get data from an rt_ipc rt-fifo
 *
 * Gets data from the rt-fifo 'fifo'.  The data, up to size 'count', is put
 * into 'buf'.  If 'timeout' is RT_NO_WAIT, the function returns immediately
 * even if 'count' bytes are not available.  If 'timeout' is RT_WAIT_FOREVER,
 * the function blocks until 'count' bytes are available.  If 'timeout' is
 * any other value, it represents the length of time after which the function
 * will return with a timeout after unsuccessfully waiting for data.  In any
 * case, as many bytes as possible are returned, even if a timeout occurs or
 * RT_NO_WAIT is specified.
 *
 * If the return value is greater than or equal to zero, it represents the
 * number of bytes received.  This might be less than 'count' if the function
 * timed out or RT_NO_WAIT was specified and less than 'count' bytes were
 * available.  If the return value is less than zero, it is the negative of
 * the error code result of a call to select(2) or read(2).  See the man
 * pages for those functions for the error values.
 *************************************************************************/
int rtf_receive(unsigned int fifo, void *buf, int count, RTIME timeout)
{
  fd_set rfds;
  struct timeval tv, *ptv;
  int ret = 0, bytes_still_to_get;

  /* set up the timeout structure */
  switch (timeout)
  {
  case RT_NO_WAIT:
    tv.tv_sec = tv.tv_usec = 0;
    ptv = &tv;
    break;
  case RT_WAIT_FOREVER:
    ptv = NULL;
    break;
  default:
    tv.tv_sec = timeout % RT_TICKS_PER_SEC;
    tv.tv_usec = (timeout * 1000000L) / RT_TICKS_PER_SEC;
    ptv = &tv;
    break;
  }

  /* loop until all requested data has been received, or timeout occurs */
  for (bytes_still_to_get=count ; bytes_still_to_get != 0 ; )
  {
    FD_ZERO(&rfds);
    FD_SET(ipc_fifos[fifo].fd, &rfds);
    if ((ret = select(FD_SETSIZE, &rfds, NULL, NULL, ptv)) == -1)
    {
      ret = -errno;
      break;
    }
    if (ret == 0)
      break;		/* timeout occurred */
    if ((ret = read(ipc_fifos[fifo].fd, buf, bytes_still_to_get)) == -1)
    {
      ret = -errno;
      break;
    }
    bytes_still_to_get -= ret;
    buf += ret;
  }
  if (ret < 0)
    return ret;
  else
    return count - bytes_still_to_get;
}

/*************************************************************************
 * rtf_send -- send data to an rt_ipc rt-fifo
 *
 * Sends data to the rt-fifo 'fifo'.  The data, up to size 'count', is taken
 * from 'buf'.  If 'timeout' is RT_NO_WAIT, the function returns immediately
 * even if 'count' bytes are not available.  If 'timeout' is RT_WAIT_FOREVER,
 * the function blocks until 'count' bytes are available.  If 'timeout' is
 * any other value, it represents the length of time after which the function
 * will return with a timeout after unsuccessfully waiting to send data.  In
 * any case, as many bytes as possible are sent, even if a timeout occurs or
 * RT_NO_WAIT is specified.
 *
 * If the return value is greater than or equal to zero, it represents the
 * number of bytes sent.  This might be less than 'count' if the function
 * timed out or RT_NO_WAIT was specified and less than 'count' bytes could
 * be sent.  If the return value is less than zero, it is the negative of
 * the error code result of a call to select(2) or write(2).  See the man
 * pages for those functions for the error values.
 *************************************************************************/
int rtf_send(unsigned int fifo, void *buf, int count, RTIME timeout)
{
  fd_set wfds;
  struct timeval tv, *ptv;
  int ret = 0, bytes_still_to_send;

  /* set up the timeout structure */
  switch (timeout)
  {
  case RT_NO_WAIT:
    tv.tv_sec = tv.tv_usec = 0;
    ptv = &tv;
    break;
  case RT_WAIT_FOREVER:
    ptv = NULL;
    break;
  default:
    tv.tv_sec = timeout % RT_TICKS_PER_SEC;
    tv.tv_usec = (timeout * 1000000L) / RT_TICKS_PER_SEC;
    ptv = &tv;
    break;
  }

  /* loop until all requested data has been sent, or timeout occurs */
  for (bytes_still_to_send=count ; bytes_still_to_send != 0 ; )
  {
    FD_ZERO(&wfds);
    FD_SET(ipc_fifos[fifo].fd, &wfds);
    if ((ret = select(FD_SETSIZE, NULL, &wfds, NULL, ptv)) == -1)
    {
      ret = -errno;
      break;
    }
    if (ret == 0)
      break;		/* timeout occurred */
    if ((ret = write(ipc_fifos[fifo].fd, buf, bytes_still_to_send)) == -1)
    {
      ret = -errno;
      break;
    }
    bytes_still_to_send -= ret;
    buf += ret;
  }
  if (ret < 0)
    return ret;
  else
    return count - bytes_still_to_send;
}
