#ifndef _RT_REGISTRY_H
#define _RT_REGISTRY_H
/*
 * rt_registry.h
 * Copyright (C) 1997, 1998 Till Siering
 *
 */

#include <linux/sched.h>    /* Imports NULL and more */
#include <linux/sysctl.h>   /* Imports sysctl() Interface */
#include <linux/rt_sched.h> /* Imports RT_TASK */

/*
 * maximum size for a string
 */
#define RT_REG_STRING_SIZE  256 

/*
 * maximum number of registered rt-tasks
 */
#define RT_REG_SIZE 20

/*
 * set the start time for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  RTIME start_time      the start time to be set
 * Out:
 *  1                     on success
 *  0                     if there is no rt-task with descriptor id.
 *  -EINVAL               on other errors.
 * Post:
 *  the start_time is set.
 */
extern int rt_reg_set_start_time(int id, RTIME start_time);

/*
 * get the start time for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 * Out:
 *  RTIME start_time      on success
 *  0                     if there is no rt-task with descriptor id.
 *  -EINVAL               on other errors.
 */
extern RTIME rt_reg_get_start_time(int id);

/*
 * set the period for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  RTIME period          the period to be set
 * Out:
 *  1                     on success
 *  0                     if there is no rt-task with descriptor id.
 *  -EINVAL               on other errors.
 * Post:
 *  the period is set.
 */
extern int rt_reg_set_period(int id, RTIME period);

/*
 * get the period for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  RT_REGISTRY *this_ptr the rt-registry
 * Out:
 *  RTIME period          on success
 *  0                     if there is no rt-task with descriptor id.
 *  -EINVAL               on other errors.
 */
extern RTIME rt_reg_get_period(int id);

/*
 * set priority for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  int priority          the priority to be set
 * Out:
 *  1                     on success
 *  0                     if there is no rt-task with descriptor id.
 *  -EINVAL               on other errors.
 * Post:
 *  the priority is set.
 */
extern int rt_reg_set_priority(int id, int priority);

/*
 * get priority for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 * Out:
 *  int priority          on success
 *  0                     if there is no rt-task with descriptor id.
 *  -EINVAL               on other errors.
 */
extern int rt_reg_get_priority(int id);

/*
 * set the info information string for rt-task with given
 * descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  char *info            the info to be set
 * Out:
 *  1                     on success
 *  0                     if there is no rt-task with descriptor id.
 *  -EINVAL               on other errors.
 * Post:
 *  the priority is set.
 */
extern int rt_reg_set_info(int id, char *info);

/*
 * get the info for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 * Out:
 *  char *info            on success
 *  NULL                  if there is no rt-task with descriptor id.
 */
extern char* rt_reg_get_info(int id);

/*
 * registers a rt-task and returns its descriptor.
 * In:
 *  RT_TASK *rt_task_ptr,
 *  char    *rt_task_name;
 *  int (* on_start_time_change) (RTIME old_start_time,
 *                                RTIME new_start_time),
 *  int (* on_period_change)     (RTIME old_period,
 *                                RTIME new_period),
 *  int (* on_priority_change)   (int old_priority,
 *                                int new_priority),
 *  int (* on_info_change)       (char *old_info_string,
 *                                char *new_info_string)
 * Out:
 *  int descrpitor the descrpitor for the rt-task.
 *  -1             if the rt-task can not be registered.
 * Post:
 *  the rt-task is registered. The value of the descriptor
 *  corresponds to the name of the directory in /proc/sys/rt_linux
 *  in which the attributes of the rt-task will be placed.
 */
extern int rt_reg_register_task(
        RT_TASK *rt_task_ptr,
        char    *rt_task_name,
        int (* on_start_time_change) (RTIME old_start_time,
                                      RTIME *new_start_time),
        int (* on_period_change)     (RTIME old_period,
                                      RTIME *new_period),
        int (* on_priority_change)   (int old_priority,
                                      int *new_priority),
        int (* on_info_change)       (char *old_info_string,
                                      char *new_info_string));
/*
 * unregisters a rt-task with given descriptor id.
 * In:
 *  int id  the descriptor for the rt-task
 * Post:
 *  the rt-task is unregistered, which means its attributes
 *  will no longer be visible in the proc-file-system.
 */
extern void rt_reg_un_register_task(int id);

#endif /* _RT_REGISTRY_H */


