#define MODULE
/*
 * rt_registry.c
 * Copyright (C) 1997, 1998 Till Siering
 *
 */

#include <linux/module.h>   /* Imports Module Definitions */
#include <linux/version.h>  /* Imports version number */
#include <linux/config.h>   /* Imports some often used functions */
#include <linux/malloc.h>   /* Imports kmalloc, the kernel malloc */
#include <linux/mm.h>       /* Imports memset and friends */
#include <linux/string.h>   /* Imports kernel string functions */

#include <linux/kernel.h>   /* Imports some often used functions */
#include <linux/timer.h>    /* Imports struct timeval */
#include <linux/ctype.h>    /* Imports size_t, isspace, ... */

#include "rt_registry.h"

/*
 * needed by the sysctl Interface
 */
#define CTL_RT          1
#define CTL_RT_SYSTEM   2
#define CTL_RT_ENTRY    3

/*
 * Need two more ctl-table entries one for rt_system directory and
 * one to mark the end of the whole ctl-table.
 */
#define RT_TABLE_SIZE   (RT_REG_SIZE + 2)
#define RT_REG_MAGIC    0x47114711
#define TMPBUFLEN 50

typedef struct rt_reg_entry_struct {
    int       magic;

    /*
     * is a rt-task registered in this entry
     */
    int       empty;                          

    /*
     * the real-time-task
     */
    RT_TASK   *rt_task_ptr;
    
    /*
     * Time to start the rt-task
     */
    RTIME     start_time;

    /*
     * period of the task
     */
    RTIME     period;
    
    /*
     * priority of the rt-task
     */
    int       priority;

    /*
     * name of the rt-task in this entry
     */
    char      rt_task_name[RT_REG_STRING_SIZE];

    /*
     * info description of rt-task in this entry
     */
    char      rt_task_info[RT_REG_STRING_SIZE];

    /*
     * name of the directory for this entry
     */
    char      directory_name[RT_REG_STRING_SIZE];
    
    /*
     * entries for proc-files in the sysctl interface
     */
    ctl_table ctl_entries[6];

    /*
     * this handler gets called, when priority was changed.
     */
    int (* on_priority_change) (int old_priority,
                                int *new_priority);

    /*
     * this handler gets called, when period was changed.
     */
    int (* on_period_change) (RTIME old_period,
                              RTIME *new_period);

    /*
     * this handler gets called, when start_time was changed.
     */
    int (* on_start_time_change) (RTIME old_start_time,
                                  RTIME *new_start_time);

    /*
     * this handler gets called, when info string was changed.
     */
    int (* on_info_change) (char *old_info_string,
                            char *new_info_string);
} RT_REG_ENTRY;

typedef struct rt_reg_system_struct {
    /*
     * BSD time the registry was loaded
     */
    struct timeval unix_startup_time;

    /*
     * RT-time the registry was loaded
     */
    RTIME          rt_startup_time;

    /*
     * proc-files to show startup times
     */
    ctl_table      entries[3];
} RT_REG_SYSTEM_INFO;

typedef struct rt_registry_struct {
    int                     magic;
    RT_REG_SYSTEM_INFO      rt_system;
    RT_REG_ENTRY            rt_entry[RT_TABLE_SIZE];
    ctl_table               rt_directory[RT_TABLE_SIZE];
    ctl_table               root_directory[2];
    struct ctl_table_header *rt_table_header;

    /*
     * set the start time for rt-task with given descriptor.
     */
    int (* set_start_time)(struct rt_registry_struct *this_ptr,
                           int id,
                           RTIME start_time);
    /*
     * get the start time for rt-task with given descriptor.
     */
    RTIME (* get_start_time)(struct rt_registry_struct *this_ptr,
                             int id);

    /*
     * set the period for rt-task with given descriptor.
     */
    int (* set_period)(struct rt_registry_struct *this_ptr,
                       int id,
                       RTIME period);
    /*
     * get the period for rt-task with given descriptor.
     */
    RTIME (* get_period)(struct rt_registry_struct *this_ptr,
                         int id);

    /*
     * set priority for rt-task with given descriptor.
     */
    int (* set_priority)(struct rt_registry_struct *this_ptr,
                         int id,
                         int priority);
    /*
     * get priority for rt-task with given descriptor.
     */
    int (* get_priority)(struct rt_registry_struct *this_ptr,
                         int id);

    /*
     * set the info information string for rt-task with given
     * descriptor.
     */
    int (* set_info)(struct rt_registry_struct *this_ptr,
                     int id,
                     char *info);
    /*
     * get the info for rt-task with given descriptor.
     */
    char *(* get_info)(struct rt_registry_struct *this_ptr,
                        int id);

    /*
     * registers a rt-task and returns its descriptor.
     */
    int (* register_task) (struct rt_registry_struct *this_ptr,
        RT_TASK *rt_task_ptr,
        char    *rt_task_name,
        int (* on_start_time_change) (RTIME old_start_time,
                                      RTIME *new_start_time),
        int (* on_period_change)     (RTIME old_period,
                                      RTIME *new_period),
        int (* on_priority_change)   (int old_priority,
                                      int *new_priority),
        int (* on_info_change)       (char *old_info_string,
                                      char *new_info_string));
    /*
     * unregisters a rt-task.
     */
    void (* un_register_task)(struct rt_registry_struct *this_ptr,
                              int id);
} RT_REGISTRY;

static RT_REGISTRY *rt_registry_ptr = NULL;


/*
 * writes the startup time in BSD style to user space.
 * In:
 *  int write          flag for read or write.
 *                      write == 1 => write
 *                      write == 0 => read
 * InOut:
 *  ctl_table *table    ctl_table entry
 *  struct file *filep  associated file struct
 *  void *buffer        associated data buffer
 *  size_t *lenp        how much data is read or written
 * Out:
 *  0 on succes
 * Post:
 *  *lenp contains the size of data read or written
 *  *filep->f_pos ist incremented by *lenp.
 */
static int
proc_do_unix_startup_time(ctl_table *table,
                          int write,
                          struct file *filp,
                          void *buffer,
                          size_t *lenp)
{
    int len = 0;
    struct timeval *startup_time = NULL;
    char buf[TMPBUFLEN];

    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : proc_do_unix_startup_time\n");
    #endif
        
    if (!table->data || !table->maxlen || !*lenp ||
        (filp->f_pos && !write)) {
        *lenp = 0;
        return 0;
    }
    /*
     * The unix startup time is always read-only,
     * thus it can just be copied to user-space.
     */
    if(!write) {
        startup_time = (struct timeval *) table->data;
        len = sprintf(buf,"%d %d\n",startup_time->tv_sec,
                                    startup_time->tv_usec);
        if (len > *lenp)
            len = *lenp;
        memcpy_tofs(buffer, buf, len);
    } else {
        len = 0;
    }
    *lenp = len;
    filp->f_pos += *lenp;

    return 0;

} /* End of proc_do_unix_startup_time */

/*
 * writes the startup time in rt-ticks to user space.
 * The startup time is divided in a low- and a high-word,
 * because writing long long directly does not work.
 * In:
 *  int write           flag for read or write.
 *                      write == 1 => write
 *                      write == 0 => read
 * InOut:
 *  ctl_table *table    ctl_table entry
 *  struct file *filep  associated file struct
 *  void *buffer        associated data buffer
 *  size_t *lenp        how much data is read or written
 * Out:
 *  0 on succes
 * Post:
 *  *lenp contains the size of data read or written
 *  *filep->f_pos ist incremented by *lenp.
 */
static int
proc_do_rt_startup_time(ctl_table *table,
                        int write,
                        struct file *filp,
                        void *buffer,
                        size_t *lenp)
{
    int len = 0;
    RTIME rt_startup_time;
    unsigned long low_word = 0, high_word  = 0;
    char buf[TMPBUFLEN];
        
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : proc_do_rt_startup_time\n");
    #endif

    if (!table->data || !table->maxlen || !*lenp ||
        (filp->f_pos && !write)) {
        *lenp = 0;
        return 0;
    }
    /*
     * The startup time in rt-ticks is always read-only,
     * thus it can just be copied to user-space.
     */
    if(!write) {
        rt_startup_time = *((RTIME *) table->data);

        /*
         * writing long long to user space does not work using
         * sprintf(buf,"%Lu\n",rt_time);
         * therefore I write the low word and high word separate.
         */
        low_word = low_word | rt_startup_time ;

        /*
         * don't use sizeof(long) instead of 32.
         * 32           => right shift is filled with 0 from left
         * sizeof(long) => right shift is filled with 1 from left
         */
        rt_startup_time = rt_startup_time >> 32;
        high_word = high_word | rt_startup_time;
        len = sprintf(buf,"%lu %lu\n",high_word, low_word);
        if (len > *lenp)
            len = *lenp;
        memcpy_tofs(buffer, buf, len);
    } else {
        len = 0;
    }
    *lenp = len;
    filp->f_pos += *lenp;

    return 0;

} /* End of proc_do_rt_startup_time */

/*
 * gets called when the name of a rt-taks is modified from user
 * space and writes the name user space.
 * In:
 *  int write          flag for read or write.
 *                      write == 1 => write
 *                      write == 0 => read
 * InOut:
 *  ctl_table *table    ctl_table entry
 *  struct file *filep  associated file struct
 *  void *buffer        associated data buffer
 *  size_t *lenp        how much data is read or written
 * Out:
 *  0 on succes
 * Post:
 *  *lenp contains the size of data read or written
 *  *filep->f_pos ist incremented by *lenp.
 */
static int
proc_do_task_name_change(ctl_table *table,
                         int write,
                         struct file *filp,
                         void *buffer,
                         size_t *lenp)
{
    RT_REG_ENTRY *rt_entry_ptr = NULL;
    int len = 0;
    char buf[RT_REG_STRING_SIZE + 1];
        
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : proc_do_task_name_change\n");
    #endif

    if (!table->data || !table->maxlen || !*lenp ||
        (filp->f_pos && !write)) {
        *lenp = 0;
        return 0;
    }
    rt_entry_ptr = (RT_REG_ENTRY *)table->data;
    if(rt_entry_ptr->magic != RT_REG_MAGIC) {
        *lenp = 0;
        return 0;
    }
    if(!write) {
        len = sprintf(buf,"%s\n",rt_entry_ptr->rt_task_name);
        if(len > *lenp)
            len = *lenp;
        memcpy_tofs(buffer, buf, len);
    } else {
        len = 0;
    }
    *lenp = len;
    filp->f_pos += *lenp;

    return 0;

} /* End of proc_do_task_name_change */

/*
 * gets called when the info of a rt-task is modified
 * from user space and writes the info to user space.
 * In:
 *  int write          flag for read or write.
 *                     write == 1 => write
 *                     write == 0 => read
 * InOut:
 *  ctl_table *table    ctl_table entry
 *  struct file *filep  associated file struct
 *  void *buffer        associated data buffer
 *  size_t *lenp        how much data is read or written
 * Out:
 *  0 on succes
 * Post:
 *  *lenp contains the size of data read or written
 *  *filep->f_pos ist incremented by *lenp.
 */
static int
proc_do_task_info_change(ctl_table *table,
                         int write,
                         struct file *filp,
                         void *buffer,
                         size_t *lenp)
{
    RT_REG_ENTRY *rt_entry_ptr = NULL;
    char buf[RT_REG_STRING_SIZE];
    int len = 0;
        
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : proc_do_task_info_change\n");
    #endif

    if (!table->data || !table->maxlen || !*lenp ||
        (filp->f_pos && !write)) {
        *lenp = 0;
        return 0;
    }
    rt_entry_ptr = (RT_REG_ENTRY *)table->data;
    if(rt_entry_ptr->magic != RT_REG_MAGIC) {
        *lenp = 0;
        return 0;
    }
    if(write) {
        /*
         * read data out of buffer.
         */
        len = *lenp;
        if(len > RT_REG_STRING_SIZE) {
            len = RT_REG_STRING_SIZE;
        }
        memcpy_fromfs((void *)buf, buffer, len);
        buf[len] = 0;
        /*
         * trigger handler and update the info string to
         * whatever the rt-task has decided to set it to.
         */
        if(NULL != rt_entry_ptr->on_info_change) {
            if(!rt_entry_ptr->on_info_change(rt_entry_ptr->rt_task_info,
                    (char *)&buf))
            {
                sprintf(rt_entry_ptr->rt_task_info, "%s", buf);
            }
        }
    } else {
        len = sprintf(buf, "%s\n", rt_entry_ptr->rt_task_info);
        if(len > *lenp)
            len = *lenp;
        memcpy_tofs(buffer, buf, len);
    }
    *lenp = len;
    filp->f_pos += *lenp;

    return 0;

} /* End of proc_do_task_info_change */

/*
 * handles requests to the period of an rt-task.
 * Attention *lenp must be set to size of data read or
 * written and must be set to *filp->f_pos += *lenp.
 * In:
 *  int write           flag for read or write.
 *                      write == 1 => write
 *                      write == 0 => read
 * InOut:
 *  ctl_table *table    ctl_table entry
 *  struct file *filep  associated file struct
 *  void *buffer        associated data buffer
 *  size_t *lenp        how much data is read or written
 * Out:
 *  0 on succes
 * Post:
 *  *lenp contains the size of data read or written
 *  *filep->f_pos ist incremented by *lenp.
 */
static int
proc_do_period_change(ctl_table *table,
                      int write,
                      struct file *filp,
                      void *buffer,
                      size_t *lenp)
{
    RT_REG_ENTRY *rt_entry_ptr  = NULL;
    int len                     = 0;
    RTIME period                = -1;
    unsigned long long tmp_time = 0;    /* needed for conversion of */ 
    unsigned long high_word     = 0;    /* high and low word to     */
    unsigned long low_word      = 0;    /* RTIME start_time         */
    char *p                     = NULL;
    char buf[TMPBUFLEN];
        
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : proc_do_period_change\n");
    #endif

    if (!table->data || !table->maxlen || !*lenp ||
        (filp->f_pos && !write) ||
        rt_registry_ptr == NULL ||
        rt_registry_ptr->magic != RT_REG_MAGIC) {
        *lenp = 0;
        return 0;
    }
    rt_entry_ptr = table->data;
    if(rt_entry_ptr->magic != RT_REG_MAGIC) {
        *lenp = 0;
        return 0;
    }
    if(write) {
        /*
         * read data out of buffer.
         */
        len = *lenp;
        if(len > TMPBUFLEN - 1) {
            len = TMPBUFLEN - 1;
        }
        memcpy_fromfs((void *)buf, buffer, len);
        buf[len] = 0;

        /*
         * check if first byte of data is valid
         * and get high word from data.
         */
        p = buf;
        while(isspace(*p)) {
            p++;
        }
        if(!isdigit(*p)) {
            *lenp = 0;
            return -EINVAL;
        }
        high_word = simple_strtoul(p, &p, 0);
        while(isspace(*p)) {
            p++;
        }
        /*
         * check if first byte of data is valid
         * and get low word from data.
         */
        if(!isdigit(*p)) {
            *lenp = 0;
            return -EINVAL;
        }
        low_word = simple_strtoul(p, &p, 0);

        /*
         * tmp_time used here because variables of type RTIME
         * are signed and shift does not work as i want it.
         */
        tmp_time = tmp_time | high_word;
        tmp_time = tmp_time << 32;
        tmp_time = tmp_time | low_word;
        period = tmp_time;
        len = p - buf;

        /*
         * trigger handler and update the period to whatever the
         * rt-task has decided to set it to.
         */
        if(NULL != rt_entry_ptr->on_period_change) {
            if(!rt_entry_ptr->on_period_change(rt_entry_ptr->period,
                    &period)) {
                rt_entry_ptr->period = period;
            }
        }
    } else {
        period = rt_entry_ptr->period;

        /*
         * writing long long to user space does not work using
         * sprintf(buf,"%Lu\n",period);
         * therefore I write the low word and high word seperate.
         */
         low_word = low_word | period ;

        /*
         * don't use sizeof(long) instead of 32.
         * 32           => right shift is filled with 0 from left
         * sizeof(long) => right shift is filled with 1 from left
         */
        period = period >> 32;
        high_word = high_word | period;
        len = sprintf(buf,"%lu %lu\n",high_word, low_word);
        if (len > *lenp)
            len = *lenp;
        memcpy_tofs(buffer, buf, len);
    }
    *lenp = len;
    filp->f_pos += *lenp;

    return 0;

} /* End of proc_do_period_change */

/*
 * handles requests to the start-time of an rt-task.
 * Attention *lenp must be set to size of data read or
 * written and must be set to *filp->f_pos += *lenp.
 * In:
 *  int write           flag for read or write.
 *                      write == 1 => write
 *                      write == 0 => read
 * InOut:
 *  ctl_table *table    ctl_table entry
 *  struct file *filep  associated file struct
 *  void *buffer        associated data buffer
 *  size_t *lenp        how much data is read or written
 * Out:
 *  0 on succes
 * Post:
 *  *lenp contains the size of data read or written
 *  *filep->f_pos ist incremented by *lenp.
 */
static int
proc_do_start_time_change(ctl_table *table,
                          int write,
                          struct file *filp,
                          void *buffer,
                          size_t *lenp)
{
    RT_REG_ENTRY *rt_entry_ptr      = NULL;
    int len                     = 0;
    RTIME start_time            = -1;
    unsigned long long tmp_time = 0;    /* needed for conversion of */ 
    unsigned long high_word     = 0;    /* high and low word to     */
    unsigned long low_word      = 0;    /* RTIME start_time         */
    char *p                     = NULL;
    char buf[TMPBUFLEN];
        
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : proc_do_start_time_change\n");
    #endif

    if (!table->data || !table->maxlen || !*lenp ||
        (filp->f_pos && !write)) {
        *lenp = 0;
        return 0;
    }
    rt_entry_ptr = table->data;;
    if(rt_entry_ptr->magic != RT_REG_MAGIC) {
        *lenp = 0;
        return 0;
    }
    if(write) {
        /*
         * read data out of buffer.
         */
        len = *lenp;
        if(len > TMPBUFLEN - 1) {
            len = TMPBUFLEN - 1;
        }
        memcpy_fromfs((void *)buf, buffer, len);
        buf[len] = 0;

        /*
         * check if first byte of data is valid
         * and get high word from data.
         */
        p = buf;
        while(isspace(*p)) {
            p++;
        }
        if(!isdigit(*p)) {
            *lenp = 0;
            return -EINVAL;
        }
        high_word = simple_strtoul(p, &p, 0);
        while(isspace(*p)) {
            p++;
        }
        /*
         * check if first byte of data is valid
         * and get low word from data.
         */
        if(!isdigit(*p)) {
            *lenp = 0;
            return -EINVAL;
        }
        low_word = simple_strtoul(p, &p, 0);

        /*
         * tmp_time used here because variable of type RTIME
         * are signed and shift does not work as i want it.
         */
        tmp_time = tmp_time | high_word;
        tmp_time = tmp_time << 32;
        tmp_time = tmp_time | low_word;
        start_time = tmp_time;
        len = p - buf;

        /*
         * trigger handler and set start_time to whatever value
         * the rt-task has decided to set it to.
         */
        if(NULL != rt_entry_ptr->on_start_time_change) {
            if(!rt_entry_ptr->on_start_time_change(
                    rt_entry_ptr->start_time,
                    &start_time)) {
                rt_entry_ptr->start_time = start_time;
            }
        }
    } else {
        start_time = rt_entry_ptr->start_time;

        /*
         * writing long long to user space does not work using
         * sprintf(buf,"%Lu\n",start_time);
         * therefore I write the low word and high word seperate.
         */
         low_word = low_word | start_time ;

        /*
         * don't use sizeof(long) instead of 32.
         * 32           => right shift is filled with 0 from left
         * sizeof(long) => right shift is filled with 1 from left
         */
        start_time = start_time >> 32;
        high_word = high_word | start_time;
        len = sprintf(buf,"%lu %lu\n",high_word, low_word);
        if (len > *lenp)
            len = *lenp;
        memcpy_tofs(buffer, buf, len);
    }
    *lenp = len;
    filp->f_pos += *lenp;

    return 0;

} /* End of proc_do_start_time_change */

/*
 * handles requests to the priority of an rt-task.
 * Attention *lenp must be set to size of data read or
 * written and must be set to *filp->f_pos += *lenp.
 * In:
 *  int write           flag for read or write.
 *                      write == 1 => write
 *                      write == 0 => read
 * InOut:
 *  ctl_table *table    ctl_table entry
 *  struct file *filep  associated file struct
 *  void *buffer        associated data buffer
 *  size_t *lenp        how much data is read or written
 * Out:
 *  0 on succes
 * Post:
 *  *lenp contains the size of data read or written
 *  *filep->f_pos ist incremented by *lenp.
 */
static int
proc_do_priority_change(ctl_table *table,
                        int write,
                        struct file *filp,
                        void *buffer,
                        size_t *lenp)
{
    RT_REG_ENTRY *rt_entry_ptr = NULL;
    int priority               = -1;
    char *p                    = NULL;
    char buf[TMPBUFLEN];

    /*
     * how much was successfully read/written
     */
    int len                    = 0;
        
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : proc_do_priority_change\n");
    #endif

    if (!table->data || !table->maxlen || !*lenp ||
        (filp->f_pos && !write) ||
        rt_registry_ptr == NULL ||
        rt_registry_ptr->magic != RT_REG_MAGIC) {
        *lenp = 0;
        return 0;
    }
    rt_entry_ptr = table->data;;
    if(rt_entry_ptr->magic != RT_REG_MAGIC) {
        *lenp = 0;
        return 0;
    }
    if(write) {
        /*
         * read data out of buffer.
         */
        len = *lenp;
        if(len > TMPBUFLEN - 1) {
            len = TMPBUFLEN - 1;
        }
        memcpy_fromfs((void *)buf, buffer, len);
        buf[len] = 0;

        /*
         * check if data is negative.
         */
        p = buf;
        while(isspace(*p)) {
            p++;
        }
        if(!isdigit(*p)) {
            *lenp = 0;
            return -EINVAL;
        }
        /*
         * convert data to signed integer.
         */
        priority = simple_strtoul(p, &p, 0);
        len = p - buf;

        /*
         * trigger handler and set priority to whatever value
         * the rt-task has decided to set it to.
         */
        if(NULL != rt_entry_ptr->on_priority_change) {
            if(!rt_entry_ptr->on_priority_change(rt_entry_ptr->priority,
                    &priority)) {
                rt_entry_ptr->priority = priority;
            }
        }

    } else {
        priority = rt_entry_ptr->priority;
        len = sprintf(buf,"%d\n",priority);
        if (len > *lenp)
            len = *lenp;
        memcpy_tofs(buffer, buf, len);
    }
    *lenp = len;
    filp->f_pos += *lenp;

    return 0;

} /* End of proc_do_priority_change */

/*--------- usefull routines to manipulate sysctl interface ----------*/

/*
 * sets all bytes in a ctl_table to 0. A ctl_table with all bytes
 * set to 0 marks the end of an array of ctl_tables.
 * InOut:
 *  ctl_table *table
 * Out:
 *  -EINVAL if the table is NULL.
 *  0 on success.
 * Post:
 *  all bytes in the ctl_table are set to 0.
 */
static int
rt_sysctl_mkend(ctl_table *table)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : rt_sysctl_mkend\n");
    #endif

    if(NULL == table) {
        return -EINVAL;
    }
    memset(table, 0, sizeof(ctl_table));

    return 0;

} /* End of rt_sysctl_mkend */

/*
 * creates a ctl_table which represents a directory.
 * In:
 *  int ctl_name      some token used internally (can be any integer)
 *  char *name        name for the directory
 *  mode_t mode       the access mode for the directory
 *  ctl_table *child  clt_table for files or subdirectories in this
 *                    directory
 * InOut:
 *  ctl_table *table 
 * Out:
 *  -EINVAL on error
 *  0       on success
 * Post:
 *  the ctl_table is filled with requested information.
 */
static int
rt_sysctl_mkdir(ctl_table *table,
                int ctl_name,
                char *name,
                mode_t mode,
                ctl_table *child)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : rt_sysctl_mkdir\n");
    #endif

    if(NULL == table || NULL == child || NULL == name || ctl_name == 0) {
        return -EINVAL;
    }
    table->ctl_name     = ctl_name;    
    table->procname     = name;
    table->data         = NULL;
    table->maxlen       = 0;
    table->mode         = mode;
    table->child        = child;
    table->proc_handler = NULL;
    table->strategy     = NULL;
    table->de           = NULL;
    table->extra1       = NULL;
    table->extra2       = NULL;

    return 0;

} /* rt_sysctl_mkdir */

/*
 * creates a ctl_table which represents a file.
 * In:
 *  int ctl_name               some token used internally
 *                             (can be any integer)
 *  char *name                 name for the file
 *  mode_t mode                the access mode for the file
 *  void *data                 data associated with this file
 *  int maxlen                 size of data
 *  proc_handler *proc_handler callback which gets triggered, when the
 *                             file is accessed proc filesystem.
 *  ctl_handler  *strategy     callback which gets triggered, when the
 *                             file is accessed over sysctl system call.
 * InOut:
 *  ctl_table *table 
 * Out:
 *  -EINVAL on error
 *  0       on success
 * Post:
 *  the ctl_table is filled with the requested information.
 */
static int
rt_sysctl_mkfile(ctl_table *table,
                 int ctl_name,
                 char *name,
                 mode_t mode,
                 void *data,
                 int maxlen,
                 proc_handler *proc_handler,
                 ctl_handler *strategy)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : rt_sysctl_mkfile\n");
    #endif

    if(NULL == table || NULL == name || ctl_name == 0) {
        return -EINVAL;
    }
    table->ctl_name     = ctl_name;
    table->procname     = name;
    table->data         = data;
    table->maxlen       = maxlen;
    table->mode         = mode;
    table->child        = NULL;
    table->proc_handler = proc_handler;
    table->strategy     = strategy;
    table->de           = NULL;
    table->extra1       = NULL;
    table->extra2       = NULL;

    return 0;

} /* End of rt_sysctl_mkfile */

/*---------------------- routines on rt-entries ----------------------*/

/*
 * initializes a rt_entry.
 * InOut:
 *  RT_REG_ENTRY *rt_entry_ptr
 * Out:
 *  -EINVAL if rt_entry is NULL
 *  0 on success
 * Post:
 *  attributes of rt_entry are initialized.
 */
static int
rt_entry_init(RT_REG_ENTRY *rt_entry_ptr)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : rt_entry_init\n");
    #endif

    if(NULL == rt_entry_ptr) {
        return -EINVAL;
    }
    rt_entry_ptr->magic                = RT_REG_MAGIC;
    rt_entry_ptr->empty                = 1;
    rt_entry_ptr->rt_task_ptr          = NULL;
    sprintf(rt_entry_ptr->rt_task_name," ");
    rt_entry_ptr->priority             = -1;
    rt_entry_ptr->start_time           = -1;
    rt_entry_ptr->period               = -1;
    sprintf(rt_entry_ptr->rt_task_info," ");
    rt_entry_ptr->on_priority_change   = NULL;
    rt_entry_ptr->on_period_change     = NULL;
    rt_entry_ptr->on_start_time_change = NULL;
    rt_entry_ptr->on_info_change       = NULL;

    return 0;

} /* End of rt_entry_init */

/*------------------- methods of the rt-registry ---------------------*/

/*
 * sets the registered start time for an rt-tasks.
 * In:
 *  int id                descriptor for the rt-task
 *  RTIME start_time      the start time to be set
 * InOut:
 *  RT_REGISTRY *this_ptr the rt-registry
 * Out:
 *  1                     on success
 *  -EINVAL               if this_ptr is NULL
 *  0                     if there is no rt-task with descriptor id.
 * Post:
 *  the start_time is set.
 */
static int
l_set_start_time(struct rt_registry_struct *this_ptr,
                 int id,
                 RTIME start_time) 
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_set_start_time\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC) {
        return EINVAL;
    }
    if(!this_ptr->rt_entry[id].empty) {
        this_ptr->rt_entry[id].start_time = start_time;
        return 1;
    }
    return 0;

} /* End of l_set_start_time */

/*
 * get the registered start time for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  RT_REGISTRY *this_ptr the rt-registry
 * Out:
 *  RTIME start_time      on success
 *  -EINVAL               if this_ptr is NULL
 *  0                     if there is no rt-task with descriptor id.
 */
static RTIME
l_get_start_time(struct rt_registry_struct *this_ptr,
                 int id)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_get_start_time\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC) {
        return -EINVAL;
    }
    if(!this_ptr->rt_entry[id].empty) {
        return this_ptr->rt_entry[id].start_time;
    }
    return 0;

} /* End of l_get_start_time */

/*
 * sets the registered period for an rt-tasks with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  RTIME period          the period to be set
 * InOut:
 *  RT_REGISTRY *this_ptr the rt-registry
 * Out:
 *  1                     on success
 *  -EINVAL               if this_ptr is NULL
 *  0                     if there is no rt-task with descriptor id.
 * Post:
 *  the period is set.
 */
static int
l_set_period(struct rt_registry_struct *this_ptr,
             int id,
             RTIME period)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_set_period\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC) {
        return -EINVAL;
    }
    if(!this_ptr->rt_entry[id].empty) {
        this_ptr->rt_entry[id].period = period;
        return 1;
    }
    return 0;

} /* End of l_set_period */

/*
 * get the registered period for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  RT_REGISTRY *this_ptr the rt-registry
 * Out:
 *  RTIME period          on success
 *  -EINVAL               if this_ptr is NULL
 *  0                     if there is no rt-task with descriptor id.
 */
static RTIME
l_get_period(struct rt_registry_struct *this_ptr,
             int id)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_get_period\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC) {
        return -EINVAL;
    }
    if(!this_ptr->rt_entry[id].empty) {
        return this_ptr->rt_entry[id].period;
    }
    return 0;

} /* End of l_get_period */

/*
 * sets the registered priority for an rt-tasks with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  int priority          the priority to be set
 * InOut:
 *  RT_REGISTRY *this_ptr the rt-registry
 * Out:
 *  1                     on success
 *  -EINVAL               if this_ptr is NULL
 *  0                     if there is no rt-task with descriptor id.
 * Post:
 *  the priority is set.
 */
static int
l_set_priority(struct rt_registry_struct *this_ptr,
               int id,
               int priority)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_set_priority\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC) {
        return -EINVAL;
    }
    if(!this_ptr->rt_entry[id].empty) {
        this_ptr->rt_entry[id].priority = priority;
        return 1;
    }
    return 0;

} /* End of l_set_priority */

/*
 * get the registered priority for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  RT_REGISTRY *this_ptr the rt-registry
 * Out:
 *  int priority          on success
 *  -EINVAL               if this_ptr is NULL
 *  0                     if there is no rt-task with descriptor id.
 */
static int
l_get_priority(struct rt_registry_struct *this_ptr,
               int id)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_get_priority\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC) {
        return -EINVAL;
    }
    if(!this_ptr->rt_entry[id].empty) {
        return this_ptr->rt_entry[id].priority;
    }
    return 0;

} /* End of l_get_priority */

/*
 * sets the registered info for an rt-tasks with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  char *info            the info to be set
 * InOut:
 *  RT_REGISTRY *this_ptr the rt-registry
 * Out:
 *  1                     on success
 *  -EINVAL               if this_ptr is NULL
 *  0                     if there is no rt-task with descriptor id.
 * Post:
 *  the priority is set.
 */
static int
l_set_info(struct rt_registry_struct *this_ptr,
           int id,
           char *info)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_set_info\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC) {
        return -EINVAL;
    }
    if(!this_ptr->rt_entry[id].empty) {
        sprintf(this_ptr->rt_entry[id].rt_task_info,
                "%s",
                info);
        return 1;
    }
    return 0;

} /* End of l_set_info */

/*
 * get the registered info for rt-task with given descriptor.
 * In:
 *  int id                descriptor for the rt-task
 *  RT_REGISTRY *this_ptr the rt-registry
 * Out:
 *  char *info            on success
 *  NULL                  if there is no rt-task with descriptor id.
 */
static char *
l_get_info(struct rt_registry_struct *this_ptr,
           int id)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_get_info\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC) {
        return NULL;
    }
    return this_ptr->rt_entry[id].rt_task_info; 

} /* End of l_get_info */

/*
 * registers a rt-task in the first empty slot in the
 * rt-registry.
 * In:
 *  RT_TASK *rt_task_ptr,
 *  char    *rt_task_name;
 *  int (* on_start_time_change) (RTIME old_start_time,
 *                                RTIME new_start_time),
 *  int (* on_period_change)     (RTIME old_period,
 *                                RTIME new_period),
 *  int (* on_priority_change)   (int old_priority,
 *                                int new_priority)
 *  int (* on_info_change)       (char *old_info_string,
 *                                char *new_info_string)
 * InOut:
 *  struct rt_registry_struct *this_ptr,
 * Out:
 *  int slot the slot number in which the rt-task is
 *           registered.
 *  -1       if there is not empty slot.
 * Post:
 *  the slot is occupied.
 */
static int
l_register_task(struct rt_registry_struct *this_ptr,
        RT_TASK *rt_task_ptr,
        char    *rt_task_name,
        int (* on_start_time_change) (RTIME old_start_time,
                                      RTIME *new_start_time),
        int (* on_period_change)     (RTIME old_period,
                                      RTIME *new_period),
        int (* on_priority_change)   (int old_priority,
                                      int *new_priority),
        int (* on_info_change)       (char *old_info_string,
                                      char *new_info_string))
{
    int slot;
    int priority;
    RTIME period;
    RTIME start_time;

    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_register_task\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC) {
        return -1;
    }
    /*
     * find empty slot.
     */
    for(slot = 0; slot < (RT_TABLE_SIZE - 2); slot++) {
        if(RT_REG_MAGIC == this_ptr->rt_entry[slot].magic
           && this_ptr->rt_entry[slot].empty) {
            /*
             * fill in slot and return slot number.
             */
            start_time = rt_task_ptr->resume_time;
            period = rt_task_ptr->period;
            priority = rt_task_ptr->priority;

            this_ptr->rt_entry[slot].empty = 0;
            this_ptr->rt_entry[slot].rt_task_ptr = rt_task_ptr;
            if(NULL != rt_task_name) {
                if(strlen(rt_task_name) > (RT_REG_STRING_SIZE - 1)) {
                    strncpy(this_ptr->rt_entry[slot].rt_task_name,
                            rt_task_name,
                            RT_REG_STRING_SIZE - 1);
                    this_ptr->rt_entry[slot].rt_task_name[RT_REG_STRING_SIZE - 1]
                            = 0;
                } else {
                    sprintf(this_ptr->rt_entry[slot].rt_task_name,
                            "%s",
                            rt_task_name);
                }
            }
            this_ptr->rt_entry[slot].start_time = start_time;
            this_ptr->rt_entry[slot].on_start_time_change =
                    on_start_time_change;

            this_ptr->rt_entry[slot].period = period;
            this_ptr->rt_entry[slot].on_period_change =
                    on_period_change;

            this_ptr->rt_entry[slot].priority = priority;
            this_ptr->rt_entry[slot].on_priority_change =
                    on_priority_change;
            this_ptr->rt_entry[slot].on_info_change =
                    on_info_change;

            return slot;
        }
    }

    return -1;

} /* End of l_register_task */

/*
 * unregisters a rt-task with given descriptor id.
 * In:
 *  int id                the descriptor for the rt-task
 * InOut:
 *  RT_REGISTRY *this_ptr the rt-registry
 * Post:
 *  the rt-task is unregistered, which means that the slot
 *  in the rt-registry is initialized again.
 */
static void
l_unregister_task(struct rt_registry_struct *this_ptr,
                  int id)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : l_unregister_task\n");
    #endif

    if(NULL == this_ptr || this_ptr->magic != RT_REG_MAGIC
       || 0 > id || (RT_TABLE_SIZE - 2) <= id) {
        return;
    }
    rt_entry_init(&this_ptr->rt_entry[id]);

} /* End of l_unregister_task */

/*
 * Create the real-time task registry.
 * Out:
 *  RT_REGISTRY *rt_registry_ptr  Registry which was created and initialized.
 * Post:
 *  The registry is initialized. Memory was kmalloced for it.
 */
static RT_REGISTRY *
rt_registry_create()
{
    RT_REGISTRY *tmp_ptr = NULL;
    int i                = 0;

    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : rt_registry_create\n");
    #endif

    if(NULL == (tmp_ptr = kmalloc(sizeof(struct rt_registry_struct),
                                   GFP_KERNEL))) {
        printk("rt_registry.c failed memory allocation\n");
        return NULL;
    }
    /*
     * initialize the rt-registry
     */
    tmp_ptr->magic            = RT_REG_MAGIC;
    tmp_ptr->register_task    = l_register_task;
    tmp_ptr->un_register_task = l_unregister_task;
    tmp_ptr->set_start_time   = l_set_start_time;
    tmp_ptr->get_start_time   = l_get_start_time;
    tmp_ptr->set_period       = l_set_period;
    tmp_ptr->get_period       = l_get_period;
    tmp_ptr->set_priority     = l_set_priority;
    tmp_ptr->get_priority     = l_get_priority;
    tmp_ptr->get_info         = l_get_info;
    tmp_ptr->set_info         = l_set_info;
    
    /*
     * Create files for each rt-task.
     * The files are named 
     * /proc/sys/rt_linux/i/start_time
     * /proc/sys/rt_linux/i/period
     * /proc/sys/rt_linux/i/priority
     * where i is in between
     * 0 <= i <= (RT_TABLE_SIZE -2)
     */
    for(i = 0; i < (RT_TABLE_SIZE - 2); i++) {
        rt_entry_init(&tmp_ptr->rt_entry[i]);

        /*
         * Create file for rt-tasks start_time.
         */
        rt_sysctl_mkfile(&tmp_ptr->rt_entry[i].ctl_entries[0],
                         CTL_RT_ENTRY,
                         "start_time",
                         0666,
                         &tmp_ptr->rt_entry[i],
                         sizeof(RT_REG_ENTRY),
                         &proc_do_start_time_change,
                         NULL);

        /*
         * Create file for rt-tasks period.
         */
        rt_sysctl_mkfile(&tmp_ptr->rt_entry[i].ctl_entries[1],
                         CTL_RT_ENTRY,
                         "period",
                         0666,
                         &tmp_ptr->rt_entry[i],
                         sizeof(RT_REG_ENTRY),
                         &proc_do_period_change,
                         NULL);

        /*
         * Create file for rt-tasks priority.
         */
        rt_sysctl_mkfile(&tmp_ptr->rt_entry[i].ctl_entries[2],
                         CTL_RT_ENTRY,
                         "priority",
                         0666,
                         &tmp_ptr->rt_entry[i],
                         sizeof(RT_REG_ENTRY),
                         &proc_do_priority_change,
                         NULL);

        /*
         * Create file for rt-tasks name.
         */
        rt_sysctl_mkfile(&tmp_ptr->rt_entry[i].ctl_entries[3],
                         CTL_RT_ENTRY,
                         "name",
                         0444,
                         &tmp_ptr->rt_entry[i],
                         sizeof(RT_REG_ENTRY),
                         &proc_do_task_name_change,
                         NULL);

        /*
         * Create file for rt-tasks info.
         */
        rt_sysctl_mkfile(&tmp_ptr->rt_entry[i].ctl_entries[4],
                         CTL_RT_ENTRY,
                         "info",
                         0666,
                         &tmp_ptr->rt_entry[i],
                         sizeof(RT_REG_ENTRY),
                         &proc_do_task_info_change,
                         NULL);

        /*
         * Create end of entries ctl_table.
         */
        rt_sysctl_mkend(&tmp_ptr->rt_entry[i].ctl_entries[5]);

        /*
         * Create directory for rt-task in /proc/sys/rt_linux
         * and put all of the above files in it.
         */
        sprintf(tmp_ptr->rt_entry[i].directory_name,"%d",i);
        rt_sysctl_mkdir(&tmp_ptr->rt_directory[i],
                        CTL_RT,
                        tmp_ptr->rt_entry[i].directory_name,
                        0555,
                        tmp_ptr->rt_entry[i].ctl_entries);
    }

    /*
     * Create files with startup-time information.
     * This files will show up in 
     * /proc/sys/rt_linux/rt_system/rt_startup_time
     * /proc/sys/rt_linux/rt_system/unix_startup_time
     */
    do_gettimeofday(&tmp_ptr->rt_system.unix_startup_time);
    tmp_ptr->rt_system.rt_startup_time = rt_get_time();
    rt_sysctl_mkfile(&tmp_ptr->rt_system.entries[0],
                     CTL_RT_SYSTEM,
                     "rt_startup_time",
                     0444,
                     &tmp_ptr->rt_system.rt_startup_time,
                     sizeof(RTIME),
                     &proc_do_rt_startup_time,
                     NULL);
    rt_sysctl_mkfile(&tmp_ptr->rt_system.entries[1],
                     CTL_RT_SYSTEM,
                     "unix_startup_time",
                     0444,
                     &tmp_ptr->rt_system.unix_startup_time,
                     sizeof(struct timeval),
                     &proc_do_unix_startup_time,
                     NULL);
    rt_sysctl_mkend(&tmp_ptr->rt_system.entries[2]);

    /*
     * Create directory in /proc/sys/rt_linux/rt_system.
     * In this directory the startup-time information
     * files are kept.
     */
    rt_sysctl_mkdir(&tmp_ptr->rt_directory[RT_TABLE_SIZE -2],
                    CTL_RT,
                    "rt_system",
                    0555,
                    tmp_ptr->rt_system.entries);
    rt_sysctl_mkend(&tmp_ptr->rt_directory[RT_TABLE_SIZE - 1]);

    /*
     * Create directory /proc/sys/rt_linux/
     */
    rt_sysctl_mkdir(&tmp_ptr->root_directory[0],
                    CTL_KERN,
                    "rt_linux",
                    0555,
                    tmp_ptr->rt_directory);
    rt_sysctl_mkend(&tmp_ptr->root_directory[1]);

    /*
     * Register the whole directory tree.
     */
    tmp_ptr->rt_table_header =
            register_sysctl_table(tmp_ptr->root_directory, 0);
    if(NULL == tmp_ptr->rt_table_header) {
        kfree(tmp_ptr);
        printk("%s %s %s\n",
                "rt_registry.c:",
                "Directories in /proc/sys for",
                "rt_linux could not be created.");
        return NULL;
    }
    return tmp_ptr;

} /* End of rt_registry_create */

/*
 * Delete the real-time task registry.
 * InOut:
 *  RT_REGISTRY *this_ptr; Registry which will be deleted.
 * Post:
 *  The registry is deleted.
 */
static void
rt_registry_delete(RT_REGISTRY *this_ptr)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : rt_registry_delete\n");
    #endif

    if(NULL != this_ptr && RT_REG_MAGIC == this_ptr->magic) {
        unregister_sysctl_table(this_ptr->rt_table_header);
        kfree(this_ptr);
    }

} /* End of rt_registry_delete */


/*
 * set the start time for rt-task with given descriptor.
 */
int
rt_reg_set_start_time(int id, RTIME start_time)
{
    return rt_registry_ptr->set_start_time(rt_registry_ptr,
            id,
            start_time);
} /* End of rt_reg_set_start_time */

/*
 * get the start time for rt-task with given descriptor.
 */
RTIME
rt_reg_get_start_time(int id)
{
    return rt_registry_ptr->get_start_time(rt_registry_ptr,
            id);
} /* End of rt_reg_get_start_time */

/*
 * set the period for rt-task with given descriptor.
 */
int
rt_reg_set_period(int id, RTIME period)
{
    return rt_registry_ptr->set_period(rt_registry_ptr,
            id,
            period);
} /* End of rt_reg_set_period */

/*
 * get the period for rt-task with given descriptor.
 */
RTIME
rt_reg_get_period(int id)
{
    return rt_registry_ptr->get_period(rt_registry_ptr,
            id);
} /* End of rt_reg_get_period */

/*
 * set priority for rt-task with given descriptor.
 */
int
rt_reg_set_priority(int id, int priority)
{
    return rt_registry_ptr->set_priority(rt_registry_ptr,
            id,
            priority);
} /* End of rt_reg_set_priority */

/*
 * get priority for rt-task with given descriptor.
 */
int
rt_reg_get_priority(int id)
{
    return rt_registry_ptr->get_priority(rt_registry_ptr,
            id);
} /* End of rt_reg_get_priority */

/*
 * set the info information string for rt-task with given
 * descriptor.
 */
int
rt_reg_set_info(int id, char *info)
{
    return rt_registry_ptr->set_info(rt_registry_ptr,
            id,
            info);
} /* End of rt_reg_set_info */

/*
 * get the info for rt-task with given descriptor.
 */
char*
rt_reg_get_info(int id)
{
    return rt_registry_ptr->get_info(rt_registry_ptr,
            id);
} /* End of rt_reg_get_info */

/*
 * registers a rt-task and returns its descriptor.
 */
int
rt_reg_register_task(RT_TASK *rt_task_ptr,
        char    *rt_task_name,
        int (* on_start_time_change) (RTIME old_start_time,
                                      RTIME *new_start_time),
        int (* on_period_change)     (RTIME old_period,
                                      RTIME *new_period),
        int (* on_priority_change)   (int old_priority,
                                      int *new_priority),
        int (* on_info_change)       (char *old_info_string,
                                      char *new_info_string))
{
    return rt_registry_ptr->register_task(rt_registry_ptr,
        rt_task_ptr,
        rt_task_name,
        on_start_time_change,
        on_period_change,
        on_priority_change,
        on_info_change);
} /* End of rt_reg_register_task */

/*
 * unregisters a rt-task.
 */
void
rt_reg_un_register_task(int id)
{
    rt_registry_ptr->un_register_task(rt_registry_ptr, id);
} /* End of rt_reg_un_register_task */

/*------------------- standard kernel module routines ----------------*/

int
init_module(void)
{
    #ifdef RT_REG_DEBUG
        printk("rt_registry.c : init_module\n");
    #endif

    /*
     * Initialize the global rt-registry.
     */
    rt_registry_ptr = rt_registry_create();
    if(NULL == rt_registry_ptr) {
        printk("rt_registry.c : failed to load module\n");
        return 1;
    }
    return 0;

} /* End of init_module */

void
cleanup_module(void)
{
    rt_registry_delete(rt_registry_ptr);

} /* End of cleanup_module */


