/************************************************************************
 *                                                                      *
 *  acl6126.c  ---  real time daq kernel functions.                     *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include "acl6126.h"

/***********************************************************************
 * some general things                                                 *
 ***********************************************************************/

/* Debug macros for DAQ board ACL6126
 */
//#define DEBUG
#ifdef DEBUG
    #define DPRINTK(args...)    printk(## args)
#else
    #define DPRINTK(args...)
#endif

/************************************************************************
 *                                                                      * 
 *  Settings on the DAQ board the following driver definitions are      *
 *  valid for:                                                          *
 *                                                                      *
 *      JP#     Pos     Meaning                                         *
 *      JP1      15    interrupt                                        *
 *      JP1-1    BI    -5 min voltage                                   *
 *      JP1-2    BI    +5 max voltage                                   *
 *      JP1-3   -5V    DAC voltage reference                            *
 *      JP2-1    BI    -5 min voltage                                   *
 *      JP2-2    BI    +5 max voltage                                   *
 *      JP2-3   -5V    DAC voltage reference                            *
 *      JP3-1    BI    -5 min voltage                                   *
 *      JP3-2    BI    +5 max voltage                                   *
 *      JP3-3   -5V    DAC voltage reference                            *
 *      JP4-1    BI    -5 min voltage                                   *
 *      JP4-2    BI    +5 max voltage                                   *
 *      JP4-3   -5V    DAC voltage reference                            *
 *      JP5-1    BI    -5 min voltage                                   *
 *      JP5-2    BI    +5 max voltage                                   *
 *      JP5-3   -5V    DAC voltage reference                            *
 *      JP6-1    BI    -5 min voltage                                   *
 *      JP6-2    BI    +5 max voltage                                   *
 *      JP6-3   -5V    DAC voltage reference                            *
 *                                                                      * 
 ************************************************************************/

#define ACL6126_NAME  "acl6126: "

/************************************************************************
 *  acl6126 initialisation                                              * 
 ************************************************************************/

/* base addr and range 
 */
#define ACL6126_BASE        0x2c0    /* i/o base address  */
#define ACL6126_RANGE       0x10     /* i/o address range */


/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

/* AO register range and offsets
 */
#define ACL6126_AO_ADDR        0x0000  /* not used, but anyway   */
#define ACL6126_AO_ZERO        0x7fff  /* zero Volt at output    */
#define ACL6126_AO_CHANNELS    6       /* number of channels     */

/* DI register range and offsets
 */
#define ACL6126_DI_PORTS       2       /* number of input ports  */
#define ACL6126_DI_DADDR       { 15, 14 } /* d in addresses      */
#define ACL6126_DI_NBITS       8       /* number of channels     */

/* DO register range and offsets
 */
#define ACL6126_DO_PORTS       2       /* number of output ports */
#define ACL6126_DO_DADDR       { 13, 12 } /* d out addresses     */
#define ACL6126_DO_NBITS       8       /* number of channels     */


/************************************************************************
 * global values                                                        * 
 ************************************************************************/

struct
    {
    unsigned char di_port[ACL6126_DI_PORTS];
    unsigned char do_port[ACL6126_DO_PORTS];
    unsigned short int ao_active;
    unsigned short int base_io;
    } 
    acl6126;


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* select channel for successive DA conversion
 */
void acl6126_ao_dac_select( unsigned short int channel )
    {
    acl6126.ao_active = channel % ACL6126_AO_CHANNELS;
    }

/* set a single analog value given that channel is already
 * selected and its range set value mapped to 16bit.
 */
void acl6126_ao_set_value0( unsigned short int value )
    {
    /* value is a 16bit number, but board assumes
     * x x x x D11 D10 D9 D8 | D7 D6 D5 D4 D3 D2 D1 D0 
     * so take the upper 12 bit of unsigned short int.
     * hbyte: take first 4 bits and shift down by 12,
     * lbyte: take next 8 bits and shift down by 4.
     */
    outb( (unsigned char) ( ( value & 0xf000 ) >> 12 ), /* high byte */
        acl6126.base_io + (acl6126.ao_active << 1) );
    outb( (unsigned char) ( ( value & 0x0ff0 ) >> 4 ),  /* low byte */
        acl6126.base_io + ( acl6126.ao_active << 1) + 1 );
    }

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
void acl6126_ao_set_value1( unsigned short int value, 
    unsigned short int channel )
    {
    acl6126_ao_dac_select( channel );
    acl6126_ao_set_value0( value );
    }


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
unsigned char acl6126_di_get_byte( unsigned short int nport )
    {
    unsigned short int di_addr[ACL6126_DI_PORTS] = ACL6126_DI_DADDR;

    nport %= ACL6126_DI_PORTS; /* just in case */

    acl6126.di_port[nport] = inb( acl6126.base_io + di_addr[nport] );
            
    return acl6126.di_port[nport];    
    }

/* get bit #"nbit" from port "nport".
 */ 
unsigned short int acl6126_di_get_bit( unsigned short int nport, 
    unsigned short int nbit )
    {
    nbit  %= ACL6126_DI_NBITS; /* just in case */

    return ( (unsigned short int) 
        ( ( ( acl6126_di_get_byte( nport ) & ( 1 << nbit ) ) >> nbit ) ) );
    } 

/* set byte "byte" on port "nport"
 */ 
void acl6126_do_set_byte( unsigned char byte,
    unsigned short int nport )
    {
    unsigned short int do_addr[ACL6126_DO_PORTS] = ACL6126_DO_DADDR;

    nport %= ACL6126_DO_PORTS; /* just in case */

    acl6126.do_port[nport] = byte; /* keep a copy */

    outb( acl6126.do_port[nport], acl6126.base_io + do_addr[nport] );
    }

/* set bit "bit" with #"nbit" on port "nport".
 */ 
void acl6126_do_set_bit( unsigned short int bit, 
    unsigned short int nport, unsigned short int nbit )
    {
    nport %= ACL6126_DO_PORTS; /* just in case */
    nbit  %= ACL6126_DO_NBITS; /* just in case */
    bit   &= 0x01;           /* just in case */

    acl6126.do_port[nport] &= ~(1 << nbit);
    acl6126.do_port[nport] |=  (bit << nbit); 
    
    acl6126_do_set_byte( acl6126.do_port[nport], nport );
    }


/***********************************************************************
 *  board defaults                                                     *
 ***********************************************************************/

static void acl6126_defaults( void )
    {
    int i;
        
    /* init values for analog output 
     */
    for ( i=0; i<ACL6126_AO_CHANNELS; i++ )
        acl6126_ao_set_value1( ACL6126_AO_ZERO, i );   

    /* init values for digital output 
     */
    for ( i=0; i<ACL6126_DO_PORTS; i++ )
        acl6126_do_set_byte( 0x00, i );   
    }


/***********************************************************************
 * Linux kernel functions                                              *
 ***********************************************************************/

/*the mandatory init_module.
 */
int init_module( void )
    {
    acl6126.base_io = ACL6126_BASE;

    /* request base io addresses
     */
    if ( check_region( acl6126.base_io, ACL6126_RANGE ) < 0 ) 
        {
     	printk( "%s could not request base io addresss\n", ACL6126_NAME );
        return -EIO;
        }
    request_region( acl6126.base_io, ACL6126_RANGE, ACL6126_NAME );
    printk( "%s requested base io address 0x%x\n", ACL6126_NAME, 
        acl6126.base_io );

    /* Set acl6126 defaults
     */
    acl6126_defaults();

    return 0;
    } 

/*the mandatory cleanup_module.
 */
void cleanup_module( void )
    {
    /* Set acl6126 defaults
     */
    acl6126_defaults();
    
    release_region( acl6126.base_io, ACL6126_RANGE );
    printk( "%s released base io address 0x%x\n", ACL6126_NAME, 
        acl6126.base_io );
    
    printk( "%s DAQ board released\n", ACL6126_NAME );
    }
