/************************************************************************
 *                                                                      *
 *  bmc1000.c  ---  real time daq kernel functions.                     *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/io.h>
#include "bmc1000.h"

/***********************************************************************
 * some general things                                                 *
 ***********************************************************************/

/* Debug macros for DAQ board BMC1000
 */
//#define DEBUG
#ifdef DEBUG
    #define DPRINTK(args...)    printk(## args)
#else
    #define DPRINTK(args...)
#endif

/************************************************************************
 *                                                                      * 
 *  Settings on the DAQ board the following driver definitions are      *
 *  valid for:                                                          *
 *                                                                      *
 *      JP#    Pos     Meaning                                          *
 *      JP1    C       TRIG0 connected to DI0                           *
 *      JP2    NC      GATE0 not connected                              * 
 *      JP3    INT     internal voltage reference                       *
 *      JP4    INT     internal voltage reference                       *
 *      JP5    -5V     DAC voltage reference                            *
 *      JP6    3       default                                          *
 *                                                                      * 
 ************************************************************************/

#define BMC1000_NAME  "bmc1000: "

/************************************************************************
 *  bmc1000 initialisation                                                * 
 ************************************************************************/

/* for base addr and range
 */
#define BMC1000_BASE  0x230    /* io base */
#define BMC1000_RANGE 0x10     /* address range */


/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

/* AI register range and offsets
 */
#define BMC1000_AI_ADDR        0x08    /* AI read address        */
#define BMC1000_AI_MUX         0x04    /* start AD conversion    */
#define BMC1000_AI_START       0xffff  /* start AD conversion    */
#define BMC1000_AI_CHANNELS    16      /* in single ended        */
#define BMC1000_AI_ADCDELAY    10      /* 10 usecs conversion    */
#define BMC1000_AI_MUXDELAY    10      /* 10 usecs settling      */

/* AO register range and offsets
 */
#define BMC1000_AO_ADDR        0x0c    /* DAC wirte address      */
#define BMC1000_AO_ZERO        0x7fff  /* zero Volt at output    */
#define BMC1000_AO_CHANNELS    2       /* number of channels     */

/* DIO register range and offsets, similar to 8255 but different though
 */
#define BMC1000_DIO_CHIPS      2       /* number of DIO nchips    */
#define BMC1000_DIO_PPERC      2       /* number of ports/nchip   */
#define BMC1000_DIO_PORTS      ( BMC1000_DIO_CHIPS*BMC1000_DIO_PPERC )
#define BMC1000_DIO_NBITS      8       /* number of bits/PORT     */
#define BMC1000_DIO_BASES      { 0x06, 0x07 } /* chips base addr  */
#define BMC1000_DIO_DADDR      { {0x00, 0x01}, {0x02, 0x03} }
#define BMC1000_DIO_OCODE      0x80;          /* output code */
#define BMC1000_DIO_ICODE      { 0x90, 0x82 } /* input code  */


/************************************************************************
 * global values                                                        * 
 ************************************************************************/

struct
    {
    unsigned char dio_port[BMC1000_DIO_PORTS];
    unsigned char dio_type[BMC1000_DIO_PORTS];
    unsigned char dio_code[BMC1000_DIO_PORTS];
    unsigned short int ai_active;
    unsigned short int ai_range[BMC1000_AI_CHANNELS];    
    unsigned short int ao_active;
    unsigned short int base_io;
    } 
    bmc1000;


/***********************************************************************
 * Analog input basic functions ...                                    *
 ***********************************************************************/

/* select channel for successive conversion
 */
void bmc1000_ai_adc_select( unsigned short int channel )
    {
    channel %= BMC1000_AI_CHANNELS;
    bmc1000.ai_active = channel;

    /* program MUX for channel
     */
    outb( (unsigned char) ( channel + 8 + bmc1000.ai_range[channel] ), 
        bmc1000.base_io + BMC1000_AI_MUX );

    /* We have to waiting here for 10us, mux settling time.
     */
    udelay( BMC1000_AI_MUXDELAY );
    }

/* program ADC for successive conversion, with both channel 
 * and range as defined in header file.
 */
void bmc1000_ai_adc_program( unsigned short int channel,
    unsigned short int range )
    {
    channel %= BMC1000_AI_CHANNELS;
    bmc1000.ai_range[channel] = range;
    
    bmc1000_ai_adc_select( channel );
    }

/* get a single analog value mapped to 16bit independently of ADC,
 * assuming that ADC is terminated ( or/and value is in FIFO).
 */
inline unsigned short int bmc1000_ai_get_value( void )
    {
    return inw( bmc1000.base_io + BMC1000_AI_ADDR );
    }

/* get a single analog value given that channel is already
 * selected and its range set, mode is software.
 */
int bmc1000_ai_get_value0( unsigned short int *value )
    {
    /* Start conversion by READING from BMC1000_AI_ADDR
     */
    (void) inw( bmc1000.base_io + BMC1000_AI_ADDR );

    /* wait for EOC
     */
    udelay( BMC1000_AI_ADCDELAY );

    *value =  bmc1000_ai_get_value();
    return 0;
    }

/* get a single analog value( channel ) given that its range set.
 */
int bmc1000_ai_get_value1( unsigned short int *value,
    unsigned short int channel )
    {
    bmc1000_ai_adc_select( channel );
    
    return bmc1000_ai_get_value0( value );
    }

/* get a single analog value( channel, range ).
 */
int bmc1000_ai_get_value2( unsigned short int *value,
    unsigned short int channel, unsigned short int range )
    {
    bmc1000_ai_adc_program( channel, range );

    return bmc1000_ai_get_value0( value );
    }


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* select channel for successive DA conversion
 */
void bmc1000_ao_dac_select( unsigned short int channel )
    {
    bmc1000.ao_active = channel;
    }

/* set a single analog value given that channel is already
 * selected and ints range set value mapped to 16bit.
 */
void bmc1000_ao_set_value0( unsigned short int value )
    {
    /* outw value, using bmc1000.ao_active
     */
    outw( value, bmc1000.base_io+bmc1000.ao_active*2+BMC1000_AO_ADDR );
    }

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
void bmc1000_ao_set_value1( unsigned short int value, 
    unsigned short int channel )
    {
    bmc1000_ao_dac_select( channel );

    bmc1000_ao_set_value0( value );
    }


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/*bmc1000_dio_set_byte  ---  set byte at port 0-3 by "byte".
 */ 
void bmc1000_dio_set_byte( unsigned char byte,
    unsigned short int nport )
    {
    int nchan, nchip;
    unsigned short int dio_daddr[BMC1000_DIO_CHIPS][BMC1000_DIO_PPERC] = 
        BMC1000_DIO_DADDR;

    nport %= BMC1000_DIO_PORTS;        /* just in case      */
    nchip =  nport/BMC1000_DIO_PPERC;  /* number of chip    */
    nchan =  nport%BMC1000_DIO_PPERC;  /* number of channel */

    if ( bmc1000.dio_type[nport] == BMC1000_DIO_OUTPUT )
        {
        bmc1000.dio_port[nport] = byte;
        
        outb( bmc1000.dio_port[nport], 
            bmc1000.base_io+dio_daddr[nchip][nchan] );
        }
    }

/*set bit at port, channel 0-7, by "bit" 0 or 1.
 */ 
void bmc1000_dio_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit )
    {
    nport %= BMC1000_DIO_PORTS; /* just in case */
    nbit  %= BMC1000_DIO_NBITS; /* just in case */
    bit   &= 0x01;            /* just in case */

    bmc1000.dio_port[nport] &= ~(1 << nbit);
    bmc1000.dio_port[nport] |=  (bit << nbit); 

    if ( bmc1000.dio_type[nport] == BMC1000_DIO_OUTPUT ) /* just in case */
        bmc1000_dio_set_byte( bmc1000.dio_port[nport], nport );
    }
    
/* bmc1000_dio_get_byte  ---  get byte at port 0-3 into "byte"
 */ 
unsigned char bmc1000_dio_get_byte( unsigned short int nport )
    {
    int nchan, nchip;
    unsigned short int dio_daddr[BMC1000_DIO_CHIPS][BMC1000_DIO_PPERC] = 
        BMC1000_DIO_DADDR;

    nport %= BMC1000_DIO_PORTS;        /* just in case      */
    nchip =  nport/BMC1000_DIO_PPERC;  /* number of chip    */
    nchan =  nport%BMC1000_DIO_PPERC;  /* number of channel */
    
    if ( bmc1000.dio_type[nport] == BMC1000_DIO_INPUT ) /* just in case */
        bmc1000.dio_port[nport] = 
            inb( bmc1000.base_io+dio_daddr[nchip][nchan] );
            
    return bmc1000.dio_port[nport];
    }

/* bmc1000_dio_get_bit  ---  get bit at port 0-5, channel 0-7,
 * into "bit".
 */ 
unsigned short int bmc1000_dio_get_bit( unsigned short int nport,
    unsigned short int nbit )
    {
    nport %= BMC1000_DIO_PORTS; /* just in case */
    nbit  %= BMC1000_DIO_NBITS; /* just in case */

    if ( bmc1000.dio_type[nport] == BMC1000_DIO_INPUT ) /* just in case */
        bmc1000.dio_port[nport] = bmc1000_dio_get_byte( nport );

    return (unsigned short int) 
        ( ( ( bmc1000.dio_port[nport] ) & ( 1<<nbit ) ) >> nbit );
    }

/* config ports as INPUT or OUTPUT, if output, set to zero.
 */ 
void bmc1000_dio_conf_port( unsigned short int nport,
    unsigned char type )
    {
    unsigned short int dio_caddr[BMC1000_DIO_CHIPS] = BMC1000_DIO_BASES;
    unsigned char dio_icode[BMC1000_DIO_PPERC] = BMC1000_DIO_ICODE;
    unsigned char ctrl;
    int nchip, nchan, i;

    nport %= BMC1000_DIO_PORTS;        /* just in case */
    nchip =  nport/BMC1000_DIO_PPERC;  /* number of chip */
    nchan =  nport%BMC1000_DIO_PPERC;  /* number of channel */
    
    /* configure port
     */
    if ( type == BMC1000_DIO_INPUT )
        {
        bmc1000.dio_type[nport] = BMC1000_DIO_INPUT;  /* type for input */
        bmc1000.dio_code[nport] = dio_icode[nchan]; /* code for input */
        }
    else if ( type == BMC1000_DIO_OUTPUT )
        {
        bmc1000.dio_type[nport] = BMC1000_DIO_OUTPUT; /* type for output */
        bmc1000.dio_code[nport] = BMC1000_DIO_OCODE; /* code for output */
        }
    bmc1000.dio_port[nport] = 0x00; /* set byte to 0  */

    /* program nchip by composing port types
     */
    ctrl = 0x00;
    for ( i=0; i<BMC1000_DIO_PPERC; i++ )
        ctrl |= bmc1000.dio_code[i+nchip*BMC1000_DIO_PPERC];
    outb( ctrl, bmc1000.base_io+dio_caddr[nchip] );
    
    /* set initial values
     */
    if ( type == BMC1000_DIO_OUTPUT )
        bmc1000_dio_set_byte( bmc1000.dio_port[nport], nport );

    if ( type == BMC1000_DIO_INPUT )
        bmc1000.dio_port[nport] = bmc1000_dio_get_byte( nport );
    }


/***********************************************************************
 *  board defaults                                                     *
 ***********************************************************************/

static void bmc1000_defaults( void )
    {
    int i;
        
    /* init ranges for analog inputs, if possible
     */
    for ( i=0; i<BMC1000_AI_CHANNELS; i++ )
        bmc1000_ai_adc_program( i, BMC1000_AI_PM5 );   

    /* configure digital io ports
     */
    for ( i=0; i<BMC1000_DIO_PORTS; i++ )
        bmc1000_dio_conf_port( i, BMC1000_DIO_INPUT );   

    /* init values for analog output 
     */
    for ( i=0; i<BMC1000_AO_CHANNELS; i++ )
        bmc1000_ao_set_value1( BMC1000_AO_ZERO, i );   
    }


/***********************************************************************
 * Linux kernel functions                                              *
 ***********************************************************************/

/*the mandatory init_module.
 */
int init_module( void )
    {
    /* request PCI io addresses
     */
    bmc1000.base_io = BMC1000_BASE;
    if ( check_region( bmc1000.base_io, BMC1000_RANGE ) < 0 ) 
        {
     	printk( "%s could not request base io addresss\n", BMC1000_NAME );
        return -EIO;
        }
    request_region( bmc1000.base_io, BMC1000_RANGE, BMC1000_NAME );
    printk( "%s requested base io address 0x%x\n", BMC1000_NAME, 
        bmc1000.base_io );

    /* Set bmc1000 defaults
     */
    bmc1000_defaults();

    return 0;
    } 

/*the mandatory cleanup_module.
 */
void cleanup_module( void )
    {
    /* Set bmc1000 defaults
     */
    bmc1000_defaults();
    
    release_region( bmc1000.base_io, BMC1000_RANGE );
    printk( "%s released base io address 0x%x\n", BMC1000_NAME, 
        bmc1000.base_io );
    
    printk( "%s DAQ board released\n", BMC1000_NAME );
    }
