/************************************************************************
 *                                                                      *
 *  board.c  ---  real time daq kernel functions.                       *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/io.h>
#include "board.h"

/***********************************************************************
 * some general things                                                 *
 ***********************************************************************/

/* Debug macros for DAQ board BOARD
 */
//#define DEBUG
#ifdef DEBUG
    #define DPRINTK(args...)    printk(## args)
#else
    #define DPRINTK(args...)
#endif

/************************************************************************
 *                                                                      * 
 *  Settings on the DAQ board the following driver definitions are      *
 *  valid for:                                                          *
 *                                                                      *
 *    JP#    Pos       Meaning                                          *
 *                                                                      * 
 ************************************************************************/

#define BOARD_NAME  "board: "

/************************************************************************
 *  board initialisation                                                * 
 ************************************************************************/

/* for PCI boards, vendor and device id. 
 */
#define BOARD_VENDOR_ID 0x00001234  /* as used in pci_find_device */
#define BOARD_DEVICE_ID 0x00005678  /* as used in pci_find_device */

/* for ISA boards, base addr and irq
 */
#define BOARD_BASE  0x300   /* io base */
#define BOARD_IRQ   0       /* no IRQ  */

/* for both in general a range
 */
#define BOARD_RANGE 0x0100  /* address range */


/************************************************************************
 *  control stuff                                                       * 
 ************************************************************************/

/* dma defines
 */
#define BOARD_DMA_ADDR      0x00    /* DMA address for select */
#define BOARD_DMA_CHANNEL   3       /* select DMA channel     */

/* irq and int defines
 */
#define BOARD_INT_MASK_ADDR     0x00    /* interrupt mask register       */
#define BOARD_INT_POLARITY_ADDR 0x00    /* interrupt polariy register    */
#define BOARD_INT_SOURCE_ADDR   0x00    /* interrupt source register     */
#define BOARD_INT_ENABLE_ADDR   0x00    /* interrupt en/disable register */
#define BOARD_IRQ_ADDR          0       /* IRQ address                   */


/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

/* AI register range and offsets
 */
#define BOARD_AI_ADDR       0x0000  /* AI read address      */
#define BOARD_AI_START      0xffff  /* start AD conversion  */
#define BOARD_AI_BUSY       0x10    /* bit 4: 1 if ADC busy */
#define BOARD_AI_CHANNELS   16      /* in single ended      */

/* AO register range and offsets
 */
#define BOARD_AO_ADDR       0x0000  /* DAC wirte address   */
#define BOARD_AO_ZERO       0x7fff  /* zero Volt at output */
#define BOARD_AO_CHANNELS   4       /* number of channels  */

/* DI register range and offsets
 */
#define BOARD_DI_PORTS      2       /* number of input ports */
#define BOARD_DI_DADDR      { 0x0000, 0x0001 } /* addresses  */
#define BOARD_DI_NBITS      8       /* number of channels    */

/* DO register range and offsets
 */
#define BOARD_DO_PORTS      2       /* number of output ports */
#define BOARD_DO_DADDR      { 0x0002, 0x0003 } /* addresses   */
#define BOARD_DO_NBITS      8       /* number of channels     */

/* DIO register range and offsets (8255 dio chip)
 */
#define BOARD_DIO_CHIPS     2       /* number of DIO nchips    */
#define BOARD_DIO_PPERC     3       /* number of ports/nchip   */
#define BOARD_DIO_PORTS     ( BOARD_DIO_CHIPS*BOARD_DIO_PPERC )
#define BOARD_DIO_NBITS     8       /* number of bits/PORT     */
#define BOARD_DIO_BASES     { 0x00c0, 0x00d0 } /* chip base addr*/
#define BOARD_DIO_DADDR     { 0x0000, 0x0004, 0x0008 }
#define BOARD_DIO_CADDR     0x000c  /* chips control register  */
#define BOARD_DIO_OCODE     0x80;   /* output code */
#define BOARD_DIO_ICODE     { 0x90, 0x82, 0x89 } /* input code */


/************************************************************************
 *  timer/counter stuff                                                 * 
 ************************************************************************/

/* counter chips (8254 timer/counter chip) 
 */
#define BOARD_COUNTER_CHIPS 1       /* number of counter chips     */
#define BOARD_COUNTER_CPERC 3       /* number of counter per chip  */
#define BOARD_COUNTER_TOTAL ( BOARD_COUNTER_CPERC*BOARD_COUNTER_CHIPS )
#define BOARD_COUNTER_BASES { 0x00e0 } /* chip address             */
#define BOARD_COUNTER_DADDR { 0x0000, 0x0004, 0x0008 }
#define BOARD_COUNTER_CADDR 0x000c  /* COUNTER control             */

/* counter control register words (8254 timer/counter) 
 */
#define BOARD_COUNTER_BIN   0x00    /* binary count mode             */
#define BOARD_COUNTER_BCD   0x01    /* BCD count mode                */
#define BOARD_COUNTER_LATCH 0x00    /* counter latch instruction     */
#define BOARD_COUNTER_LO    0x10    /* read/write low counter byte   */
#define BOARD_COUNTER_HI    0x20    /* read/write high counter byte  */
#define BOARD_COUNTER_LOHI  0x30    /* read/write low, then high     */
#define BOARD_COUNTER_SEL   { 0x00, 0x40, 0x80 } /* select counter i */
#define BOARD_COUNTER_RBACK 0xc0    /* counter read back             */

/* in general, there are some 8254 timer/counter with several counters 
 * each. Some counters are used as such, others are used as timers, single
 * or by combination of two counters. For AI, a timer is usually used
 * to serve as pacer. For an application one has:
 */
#define BOARD_TIMER_BASE    125     /* Clock time in nanosecs        */
#define BOARD_TIMER_16      1       /* available 16bit timers        */
#define BOARD_TIMER_16C     {0}     /* counter used for 16 bit timer */
#define BOARD_TIMER_32      1       /* available 32bit timers        */
#define BOARD_TIMER_32C     {1,2}   /* counter used for 32 bit timer */

#define BOARD_PACER_ADDR   10


/************************************************************************
 * global values                                                        * 
 ************************************************************************/

struct
    {
    unsigned char di_port[BOARD_DI_PORTS];
    unsigned char do_port[BOARD_DO_PORTS];
    unsigned char dio_port[BOARD_DIO_PORTS];
    unsigned char dio_type[BOARD_DIO_PORTS];
    unsigned char dio_code[BOARD_DIO_PORTS];
    unsigned short int ai_mode;
    unsigned short int ai_active;
    unsigned short int ai_range[BOARD_AI_CHANNELS];    
    unsigned short int ao_active;
    unsigned short int ao_range[BOARD_AO_CHANNELS];
    unsigned long  int base_phys;
    unsigned short int base_io;
    unsigned short int irq;
    unsigned short int dma;
    unsigned char mode_status;
    unsigned char irq_status;
    unsigned char int_status;
    unsigned char dma_status;
    } 
    board;


/***********************************************************************
 * Control functions ...                                               *
 ***********************************************************************/

/* program interrupt with "mask", source and its polarity
 */
void board_int_program( unsigned char mask, unsigned char source,
    unsigned char polarity )
    {
    outb( mask,	board.base_io+BOARD_INT_MASK_ADDR );
    outb( source, board.base_io+BOARD_INT_SOURCE_ADDR );
    outb( polarity, board.base_io+BOARD_INT_POLARITY_ADDR );
    }

/* set interrupt status, enable or disable
 */
void board_irq_status( unsigned char state  )
    {
    outb( state, board.base_io+BOARD_INT_ENABLE_ADDR );
    }

/* set number of interrupt line, for ISA
 */
void board_irq_set( void )
    {
    outb( board.irq, board.base_io+BOARD_IRQ_ADDR );
    }

/* get number of interrupt line, irq handler can be defined outside
 */
int board_irq_get( void )
    {
    return board.irq;
    }

/*irq handler for board 
 */
unsigned int board_irq_handler( unsigned int irq_number, 
    struct pt_regs *p )
    {
    /* nothing done her, use this template in your real time application !
     */
    
    return 0;
    }

/* dma enable/disable
 */ 
void board_dma_status( unsigned char state )
    {
    outb( state, board.base_io+BOARD_DMA_ADDR );
    }

/* start counter "counter" with divisor and mode
 */
void board_counter_set( unsigned int counter, 
    unsigned int divisor, unsigned char mode )
    {
    int nchan, nchip;
    unsigned short int counter_caddr[BOARD_COUNTER_CHIPS] = BOARD_COUNTER_BASES;
    unsigned short int counter_daddr[BOARD_COUNTER_CPERC] = BOARD_COUNTER_DADDR;
    unsigned char counter_sel[BOARD_COUNTER_CPERC] = BOARD_COUNTER_SEL;
    
    counter %= BOARD_COUNTER_TOTAL;
    nchip   = counter / BOARD_COUNTER_CPERC;
    nchan   = counter % BOARD_COUNTER_CPERC;
    
    DPRINTK("counter[%d] with %d \n",counter,divisor);

    outb( (mode|counter_sel[nchan]|BOARD_COUNTER_LOHI|BOARD_COUNTER_BIN), 
        board.base_io+counter_caddr[nchip]+BOARD_COUNTER_CADDR );
    outb( (unsigned char)(divisor),	   
        board.base_io+counter_caddr[nchip]+counter_daddr[nchan] );
    outb( (unsigned char)(divisor>>8),
        board.base_io+counter_caddr[nchip]+counter_daddr[nchan] );
    }

/* start timer using coutners with "time" in nanosecs and mode !
 */
unsigned int board_timer16_start( unsigned int timer,
    unsigned int time, unsigned char mode )
    {
    unsigned int divisor;
    int timer_16c[BOARD_TIMER_16] = BOARD_TIMER_16C;

    divisor = time/BOARD_TIMER_BASE;
    
    timer %= BOARD_TIMER_16;
    board_counter_set( timer_16c[timer], time/BOARD_TIMER_BASE, mode );
    
    return divisor*BOARD_TIMER_BASE;
    }

/* stop timer using coutners
 */
void board_timer16_stop( unsigned int timer )
    {
    int timer_16c[BOARD_TIMER_16] = BOARD_TIMER_16C;
    
    timer %= BOARD_TIMER_16;
    board_counter_set( timer_16c[timer], 0x0000, 0x00 );
    }

/* start timer using two coutners, with "time" in nanosecs and mode!
 */
unsigned int board_timer32_start( unsigned int timer,
    unsigned int time, unsigned char mode )
    {
    unsigned int divisor,d1,d2;
    int timer_32c[BOARD_TIMER_32*2] = BOARD_TIMER_32C;
    
    timer  %= BOARD_TIMER_32;
    divisor = time/BOARD_TIMER_BASE;
    DPRINTK("divisor[%d] = %d\n",timer,divisor);
    
    d1 = divisor/0x0ffff+1;
    if ( d1 < 0x00002 ) d1 = 2;
    if ( d1 > 0x0ffff ) d1 = 0x0ffff;
    d2 = divisor/d1;
    if ( d2 < 0x00002 ) d2 = 2;
    if ( d2 > 0x0ffff ) d2 = 0x0ffff;

    time    = d1*d2*BOARD_TIMER_BASE;   
    DPRINTK("timer[%d] at %d ns\n",timer,time);
    
    board_counter_set( timer_32c[timer*2], 
        (unsigned short int)(d1), mode );
    board_counter_set( timer_32c[timer*2+1], 
        (unsigned short int)(d2), mode );

    return time;
    }

/* stop timer using coutners
 */
void board_timer32_stop( unsigned int timer )
    {
    int timer_32c[BOARD_TIMER_32*2] = BOARD_TIMER_32C;
    
    timer %= BOARD_TIMER_32;
    board_counter_set( timer_32c[timer*2],   0x0000, 0x00 );
    board_counter_set( timer_32c[timer*2+1], 0x0000, 0x00 );
    }

/* program pacer to use timer
 */
void board_pacer_set( unsigned char mode )
    {
    outb( mode, board.base_io+BOARD_PACER_ADDR );
    }


/***********************************************************************
 * Analog input basic functions ...                                    *
 ***********************************************************************/

/* select adc mode
 */
void board_ai_set_mode( unsigned char mode )
    {
    board.ai_mode = mode;

    /* program baord for mode
     */
    }

/* select channel for successive conversion
 */
void board_ai_adc_select( unsigned short int channel )
    {
    channel %= BOARD_AI_CHANNELS;
    board.ai_active = channel;
    
    /* program MUX for to use this channel
     */
    }

/* program ADC for successive conversion, with both channel 
 * and range as defined in header file.
 */
void board_ai_adc_program( unsigned short int channel,
    unsigned short int range )
    {
    channel %= BOARD_AI_CHANNELS;
    board.ai_range[channel] = range;

    board_ai_adc_select( channel );
    
    /* program ADC range/gain for this channel 
     * and set thi channel active on board ! 
     */
    }

/* get a single analog value mapped to 16bit independently of ADC,
 * assuming that ADC is terminated ( or/and value is in FIFO).
 */
inline unsigned short int board_ai_get_value( void )
    {
    unsigned short int input;

    /* get values from register (or FIFO)
     */
    input = inw( board.base_io+BOARD_AI_ADDR );

    return ( input << 4 ); /* 12 to 16bit */
    }

/* get a single analog value given that channel is already
 * selected and ints range set, mode is software.
 */
int board_ai_get_value0( unsigned short int *value )
    {
    /* start ADC
     */
    outw( BOARD_AI_START, board.base_io+BOARD_AI_ADDR );

    /* wait for EOC, if not return error
     */

    *value = board_ai_get_value();
    return 0;
    }

/* get a single analog value( channel ) given that its range set.
 */
int board_ai_get_value1( unsigned short int *value,
    unsigned short int channel )
    {
    board_ai_adc_select( channel );
    
    return board_ai_get_value0( value );
    }

/* get a single analog value( channel, range ).
 */
int board_ai_get_value2( unsigned short int *value, 
    unsigned short int channel, unsigned short int range )
    {
    board_ai_adc_program( channel, range );

    return board_ai_get_value0( value );
    }

/* start a channel scan with list, number of channels and sampling time[us]
 * mode has to be set appropriately, extern (pre,post,middle) or internal
 */
int board_ai_start_scan( unsigned short int *list,
    unsigned short int N, unsigned int time, unsigned char mode )
    {
    /* Program cannel list on board
     */
     
    /* Stop timer and clear FIFO
     */
     
    /* Program PIC for scan
     */
     
    /* Start timer according to mode
     */
    if ( mode == BOARD_AI_EXTPACER )
        board_timer32_start( BOARD_AI_EXTPACER, time, BOARD_COUNTER_MODE2 );
    else if ( mode == BOARD_AI_INTPACER )
        board_timer32_start( BOARD_AI_INTPACER, time, BOARD_COUNTER_MODE2 );
   
    return 0;
    }

/* Once the scan has been started, the application is responsible
 * for picking up values from the fifos periodically. If an interupt
 * is available, this can be done by the respective routine. The
 * rest depends on the application.
 */
int board_ai_get_scan( unsigned short int *value )
    {
    /* check end of FIFO if yes return 0, if not empty, then */
    
    *value = board_ai_get_value();
    
    return 1;
    }

/* Stop analogue input scan, wbasically stop the timer.
 */
void board_ai_stop_scan( void )
    {
    board_timer32_stop( BOARD_AI_EXTPACER );
    board_timer32_stop( BOARD_AI_INTPACER );
    }


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* activate channel for successive DA conversion
 */
void board_ao_dac_select( unsigned short int channel )
    {
    board.ao_active = channel;
    
    /* program DAC (MUX) for to use this channel
     */
    }

/* program DAC for successive DA conversion with both channel 
 * and range as defined in header file.
 */
void board_ao_dac_program( unsigned short int channel,
    unsigned short int range )
    {
    channel %= BOARD_AO_CHANNELS;
    board.ao_range[channel] = range;

    /* program DAC range/gain for this channel 
     * and set this channel active on board ! 
     */
    board_ao_dac_select( channel );
    }

/* set a single analog value given that channel is already
 * selected and ints range set value mapped to 16bit.
 */
void board_ao_set_value0( unsigned short int value )
    {
    /* outb or outw value, use board.ao_active
     */
    outw( value, board.base_io+BOARD_AO_ADDR );
    }

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
void board_ao_set_value1( unsigned short int value, 
    unsigned short int channel )
    {
    board_ao_dac_select( channel );
    board_ao_set_value0( value );
    }

/* set a single analog value (channel,range).
 * value is mapped to 16bit independently of DAC.
 */
void board_ao_set_value2( unsigned short int value,
    unsigned short int channel, unsigned short int range )
    {
    board_ao_dac_program( channel, range );
    board_ao_set_value0( value );
    }


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
unsigned char board_di_get_byte( unsigned short int nport )
    {
    unsigned short int di_addr[BOARD_DI_PORTS] = BOARD_DI_DADDR;

    nport %= BOARD_DI_PORTS; /* just in case */

    board.di_port[nport] = inb( board.base_io + di_addr[nport] );
            
    return board.di_port[nport];    
    }

/* get bit #"nbit" from port "nport".
 */ 
unsigned short int board_di_get_bit( unsigned short int nport, 
    unsigned short int nbit )
    {
    nbit  %= BOARD_DI_NBITS; /* just in case */

    return ( (unsigned short int) 
        ( ( ( board_di_get_byte( nport ) & ( 1 << nbit ) ) >> nbit ) ) );
    } 

/* set byte "byte" on port "nport"
 */ 
void board_do_set_byte( unsigned char byte, 
    unsigned short int nport )
    {
    unsigned short int do_addr[BOARD_DO_PORTS] = BOARD_DO_DADDR;

    nport %= BOARD_DO_PORTS; /* just in case */

    board.do_port[nport] = byte; /* keep a copy */

    outb( board.do_port[nport], board.base_io + do_addr[nport] );
    }

/* set bit "bit" with #"nbit" on port "nport".
 */ 
void board_do_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit )
    {
    nport %= BOARD_DO_PORTS; /* just in case */
    nbit  %= BOARD_DO_NBITS; /* just in case */
    bit   &= 0x01;           /* just in case */

    board.do_port[nport] &= ~(1 << nbit);
    board.do_port[nport] |=  (bit << nbit); 
    
    board_do_set_byte( board.do_port[nport], nport );
    }

/*board_dio_set_byte  ---  set byte at port 0-5 by "byte".
 */ 
void board_dio_set_byte( unsigned char byte,
    unsigned short int nport )
    {
    int nchan, nchip;
    unsigned short int dio_caddr[BOARD_DIO_CHIPS] = BOARD_DIO_BASES;
    unsigned short int dio_daddr[BOARD_DIO_PPERC] = BOARD_DIO_DADDR;

    nport %= BOARD_DIO_PORTS;        /* just in case      */
    nchip =  nport/BOARD_DIO_PPERC;  /* number of chip    */
    nchan =  nport%BOARD_DIO_PPERC;  /* number of channel */

    if ( board.dio_type[nport] == BOARD_DIO_OUTPUT )
        {
        board.dio_port[nport] = byte;
        
        outb( board.dio_port[nport], 
            board.base_io+dio_caddr[nchip]+dio_daddr[nchan] );
        }
    }

/*set bit at port, channel 0-7, by "bit" 0 or 1.
 */ 
void board_dio_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit )
    {
    nport %= BOARD_DIO_PORTS; /* just in case */
    nbit  %= BOARD_DIO_NBITS; /* just in case */
    bit   &= 0x01;            /* just in case */

    board.dio_port[nport] &= ~(1 << nbit);
    board.dio_port[nport] |=  (bit << nbit); 

    if ( board.dio_type[nport] == BOARD_DIO_OUTPUT ) /* just in case */
        board_dio_set_byte( board.dio_port[nport], nport );
    }
    
/* board_dio_get_byte  ---  get byte at port 0-5 into "byte"
 */ 
unsigned char board_dio_get_byte( unsigned short int nport )
    {
    int nchan, nchip;
    unsigned short int dio_caddr[BOARD_DIO_CHIPS] = BOARD_DIO_BASES;
    unsigned short int dio_daddr[BOARD_DIO_PPERC] = BOARD_DIO_DADDR;

    nport %= BOARD_DIO_PORTS;        /* just in case      */
    nchip =  nport/BOARD_DIO_PPERC;  /* number of chip    */
    nchan =  nport%BOARD_DIO_PPERC;  /* number of channel */
    
    if ( board.dio_type[nport] == BOARD_DIO_INPUT ) /* just in case */
        board.dio_port[nport] = 
            inb( board.base_io+dio_caddr[nchip]+dio_daddr[nchan] );
            
    return board.dio_port[nport];
    }

/* board_dio_get_bit  ---  get bit at port 0-5, channel 0-7,
 * into "bit".
 */ 
unsigned short int board_dio_get_bit( unsigned short int nport,
    unsigned short int nbit )
    {
    nport %= BOARD_DIO_PORTS; /* just in case */
    nbit  %= BOARD_DIO_NBITS; /* just in case */

    if ( board.dio_type[nport] == BOARD_DIO_INPUT ) /* just in case */
        board.dio_port[nport] = board_dio_get_byte( nport );

    return (unsigned short int) 
        ( ( ( board.dio_port[nport] ) & ( 1<<nbit ) ) >> nbit );
    }

/* config ports as INPUT or OUTPUT, if output, set to zero.
 */ 
void board_dio_conf_port( unsigned short int nport,
    unsigned char type )
    {
    unsigned short int dio_caddr[BOARD_DIO_CHIPS] = BOARD_DIO_BASES;
    unsigned char dio_icode[BOARD_DIO_PPERC] = BOARD_DIO_ICODE;
    unsigned char ctrl;
    int nchip, nchan, i;

    nport %= BOARD_DIO_PORTS;        /* just in case */
    nchip =  nport/BOARD_DIO_PPERC;  /* number of chip */
    nchan =  nport%BOARD_DIO_PPERC;  /* number of channel */
    
    /* configure port
     */
    if ( type == BOARD_DIO_INPUT )
        {
        board.dio_type[nport] = BOARD_DIO_INPUT;  /* type for input */
        board.dio_code[nport] = dio_icode[nchan]; /* code for input */
        }
    else if ( type == BOARD_DIO_OUTPUT )
        {
        board.dio_type[nport] = BOARD_DIO_OUTPUT; /* type for output */
        board.dio_code[nport] = BOARD_DIO_OCODE; /* code for output */
        }
    board.dio_port[nport] = 0x00; /* set byte to 0  */

    /* program nchip by composing port types
     */
    ctrl = 0x00;
    for ( i=0; i<BOARD_DIO_PPERC; i++ )
        ctrl |= board.dio_code[i+nchip*BOARD_DIO_PPERC];
    outb( ctrl, board.base_io+dio_caddr[nchip]+BOARD_DIO_CADDR );
    
    /* set initial values
     */
    if ( type == BOARD_DIO_OUTPUT )
        board_dio_set_byte( board.dio_port[nport], nport );

    if ( type == BOARD_DIO_INPUT )
        board.dio_port[nport] = board_dio_get_byte( nport );
    }


/***********************************************************************
 *  board defaults                                                     *
 ***********************************************************************/

static void board_defaults( void )
    {
    int i;
        
    /* init ranges for analog inputs, if possible
     */
    for ( i=0; i<BOARD_AI_CHANNELS; i++ )
        board_ai_adc_program( i, BOARD_AI_PM5 );   

    /* init ranges for analog outputs, if possible
     */
    for ( i=0; i<BOARD_AO_CHANNELS; i++ )
        board_ao_dac_program( i, BOARD_AO_PM5 );   

    /* configure digital io ports
     */
    for ( i=0; i<BOARD_DIO_PORTS; i++ )
        board_dio_conf_port( i, BOARD_DIO_INPUT );   

    /* init values for analog output 
     */
    for ( i=0; i<BOARD_AO_CHANNELS; i++ )
        board_ao_set_value1( BOARD_AO_ZERO, i );   

    /* init values for digital output 
     */
    for ( i=0; i<BOARD_DO_PORTS; i++ )
        board_do_set_byte( 0x00, i );   
    }


/***********************************************************************
 * Linux kernel functions                                              *
 ***********************************************************************/

/*the mandatory init_module.
 */
int init_module( void )
    {
    struct pci_dev *dev;

    /* Check presence of PCI bus, ca se discute.
     */
    if( !pci_present() ) 
        {
        printk("%s PCI bus not present!\n", BOARD_NAME );
        return -ENODEV;
        }
    
    /* find the first device of this kind
     */
    dev =  pci_find_device( BOARD_VENDOR_ID, BOARD_DEVICE_ID, NULL ); 
    if ( !dev )
        {
        printk("%s no %s with vendor ID %x and device ID %x found\n", 
            BOARD_NAME, BOARD_NAME, BOARD_VENDOR_ID, BOARD_DEVICE_ID );
        return -ENODEV;
        }
    printk("%s %s with vendor ID %x and device ID %x found\n", 
        BOARD_NAME, BOARD_NAME, BOARD_VENDOR_ID, BOARD_DEVICE_ID );

    /* request PCI io addresses
     */
    board.base_phys = dev->base_address[0] & PCI_BASE_ADDRESS_IO_MASK;
    if ( check_region( board.base_phys, BOARD_RANGE ) < 0 ) 
        {
     	printk( "%s could not request base io addresss\n", BOARD_NAME );
        return -EIO;
        }
    request_region( board.base_phys, BOARD_RANGE, BOARD_NAME );

    board.base_io = (unsigned short int)(board.base_phys);
    printk( "%s requested base io address 0x%x\n", BOARD_NAME, 
        board.base_io );

    board.irq = dev->irq;
    printk("%s IRQ is selected to be %d (0 = no interrupt)\n", 
        BOARD_NAME, board.irq );

    /* Set board defaults
     */
    board_defaults();

    return 0;
    } 

/*the mandatory cleanup_module.
 */
void cleanup_module( void )
    {
    /* Set board defaults
     */
    board_defaults();
    
    release_region( board.base_phys, BOARD_RANGE );
    printk( "%s released base io address 0x%x\n", BOARD_NAME, 
        board.base_io );
    
    printk( "%s DAQ board released\n", BOARD_NAME );
    }
