/************************************************************************
 *                                                                      *
 *  board.h  ---  daq inputs header file for board.c.                   *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *  control stuff                                                       * 
 ************************************************************************/

/* dma
 */
#define BOARD_DMA_ENABLE    0x00    /* enable DMA transfer  */
#define BOARD_DMA_DISABLE   0x00    /* disable DMA transfer */

/* interrupt mask
 */
#define BOARD_INTX_DISABLE  0x00    /* disable INT generation */
#define BOARD_INT0_ENABLE   0x01    /* enable INT0 generation */
#define BOARD_INT1_ENABLE   0x02    /* enable INT1 generation */
#define BOARD_INT2_ENABLE   0x04    /* enable INT2 generation */
#define BOARD_INT3_ENABLE   0x08    /* enable INT3 generation */
#define BOARD_INTX_ENABLE   0x0f    /* enable all INTs        */

/* interrupt polarity
 */
#define BOARD_INTX_INVERSE  0x00    /* select polarity all inverse      */
#define BOARD_INT0_NONINV   0x01    /* select polarity INT0 non-inverse */
#define BOARD_INT1_NONINV   0x02    /* select polarity INT1 non-inverse */
#define BOARD_INT2_NONINV   0x03    /* select polarity INT2 non-inverse */
#define BOARD_INT3_NONINV   0x04    /* select polarity INT3 non-inverse */
#define BOARD_INTX_NONINV   0x0f    /* select polarity all non-inverse  */

/* interrupt source
 */
#define BOARD_INT0_SOURCE1  0x04    /* select source */
#define BOARD_INT0_SOURCE2  0x04    /* select source */
#define BOARD_INT1_SOURCE1  0x00    /* select source */
#define BOARD_INT1_SOURCE1  0x00    /* select source */

/* interrupt dis/enable
 */
#define BOARD_IRQ_DISABLE   0x00    /* disable all interrupts */
#define BOARD_IRQ_ENABLE    0x01    /* enable interrupt       */

/* counter modes
 */
#define BOARD_COUNTER_MODE0 0x00    /* interrupt on terminal count */
#define BOARD_COUNTER_MODE1 0x02    /* programmable oneshot        */
#define BOARD_COUNTER_MODE2 0x04    /* rate generator              */
#define BOARD_COUNTER_MODE3 0x06    /* square wave generator       */
#define BOARD_COUNTER_MODE4 0x08    /* software triggered pulse    */
#define BOARD_COUNTER_MODE5 0x0a    /* hardware triggered pulse    */


/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

/* AI range
 */
#define BOARD_AI_PM1    0x0140      /* [  -1.0,    1.0   ] Volt */
#define BOARD_AI_PM5    0x0000      /* [  -5.0,    5.0   ] Volt */
#define BOARD_AI_PM10   0x0100      /* [ -10.0,   10.0   ] Volt */
#define BOARD_AI_0P1    0x0240      /* [   0.0,    1.0   ] Volt */
#define BOARD_AI_0P10   0x0200      /* [   0.0,   10.0   ] Volt */

#define BOARD_AI_FIFO   1024        /* input FIFO size */

#define BOARD_AI_EXTPACER   0       /* use timer O as external pacer */
#define BOARD_AI_INTPACER   1       /* use timer 1 as internal pacer */

/* ADC mode
 */
#define BOARD_SOFTWARE_MODE     0x00    /* software trigger mode */
#define BOARD_EXTERNAL_MODE     0x02    /* external trigger mode */
#define BOARD_INTERNAL_MODE     0x03    /* internal trigger mode */

/* AO range
 */
#define BOARD_AO_PM5    0x0000      /* [  -5.0,    5.0   ] Volt */
#define BOARD_AO_PM10   0x0100      /* [ -10.0,   10.0   ] Volt */

/* DIO port type
 */
#define BOARD_DIO_INPUT     0x00    /* dio port is an output */
#define BOARD_DIO_OUTPUT    0x01    /* dio port is an input  */


/***********************************************************************
 * Control functions ...                                               *
 ***********************************************************************/

/* program interrupt with "mask", source and its polarity
 */
extern void board_int_program( unsigned char mask, unsigned char source,
    unsigned char polarity );

/* set interrupt status, enable or disable
 */
extern void board_irq_status( unsigned char state );

/* set number of interrupt line, for ISA
 */
extern void board_irq_set( void );

/* get number of interrupt line, irq handler can be defined outside
 */
extern int board_irq_get( void );

/* dma enable/disable
 */ 
extern void board_dma_status( unsigned char state );

/* start counter "counter" with divisor and mode
 */
extern void board_counter_set( unsigned int counter, 
    unsigned int divisor, unsigned char mode );

/* start timer using coutners with "time" in nanosecs and mode !
 */
extern unsigned int board_timer16_start( unsigned int timer,
    unsigned int time, unsigned char mode );

/* stop timer using coutners
 */
extern void board_timer16_stop( unsigned int timer );

/* start timer using two coutners, with "time" in nanosecs and mode!
 */
extern unsigned int board_timer32_start( unsigned int timer,
    unsigned int time, unsigned char mode );

/* stop timer using coutners
 */
extern void board_timer32_stop( unsigned int timer );

/* program pacer to use timer
 */
extern void board_pacer_set( unsigned char mode );


/***********************************************************************
 * Analog input basic functions ...                                    *
 ***********************************************************************/

/* select adc mode
 */
extern void board_ai_set_mode( unsigned char mode );

/* select channel for successive conversion
 */
extern void board_ai_adc_select( unsigned short int channel );

/* program ADC for successive conversion, with both channel 
 * and range as defined in header file.
 */
extern void board_ai_adc_program( unsigned short int channel,
    unsigned short int range );

/* get a single analog value given that channel is already
 * selected and ints range set, mode is software.
 */
extern int board_ai_get_value0( unsigned short int *value );

/* get a single analog value( channel ); given that its range set.
 */
extern int board_ai_get_value1( unsigned short int *value,
    unsigned short int channel );

/* get a single analog value( channel, range );.
 */
extern int board_ai_get_value2( unsigned short int *value,
    unsigned short int channel, unsigned short int range );

/* start a channel scan with list, number of channels and sampling time[us]
 * mode has to be set appropriately, extern (pre,post,middle) or internal
 */
extern int board_ai_start_scan( unsigned short int *channellist,
    unsigned short int nchannels, unsigned int time, unsigned char mode );

/* Once the scan has been started, the application is responsible
 * for picking up values from the fifos periodically. If an interupt
 * is available, this can be done by the respective routine. The
 * rest depends on the application.
 */
extern int board_ai_get_scan( unsigned short int *value );

/* Stop analogue input scan, wbasically stop the timer.
 */
extern void board_ai_stop_scan( void );


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* activate channel for successive DA conversion
 */
extern void board_ao_dac_select( unsigned short int channel );

/* program DAC for successive DA conversion with both channel 
 * and range as defined in header file.
 */
extern void board_ao_dac_program( unsigned short int channel,
    unsigned short int range );

/* set a single analog value given that channel is already
 * selected and ints range set value mapped to 16bit.
 */
extern void board_ao_set_value0( unsigned short int value );

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
extern void board_ao_set_value1( unsigned short int value, 
    unsigned short int channel );

/* set a single analog value (channel,range).
 * value is mapped to 16bit independently of DAC.
 */
extern void board_ao_set_value2( unsigned short int value,
    unsigned short int channel, unsigned short int range );


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
extern unsigned char board_di_get_byte( unsigned short int nport );

/* get bit #"nbit" from port "nport".
 */ 
extern unsigned short int board_di_get_bit( unsigned short int nport, 
    unsigned short int nbit );

/* set byte "byte" on port "nport"
 */ 
extern void board_do_set_byte( unsigned char byte,
    unsigned short int nport );

/* set bit "bit" with #"nbit" on port "nport".
 */ 
extern void board_do_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit );

/*board_dio_set_byte  ---  set byte at port 0-5 by "byte".
 */ 
extern void board_dio_set_byte( unsigned char byte,
    unsigned short int nport );

/*set bit at port, channel 0-7, by "bit" 0 or 1.
 */ 
extern void board_dio_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit );
    
/* board_dio_get_byte  ---  get byte at port 0-5 into "byte"
 */ 
extern unsigned char board_dio_get_byte( unsigned short int nport );

/* board_dio_get_bit  ---  get bit at port 0-5, channel 0-7,
 * into "bit".
 */ 
extern unsigned short int board_dio_get_bit( unsigned short int nport,
    unsigned short int nbit );

/* config ports as INPUT or OUTPUT, if output, set to zero.
 */ 
extern void board_dio_conf_port( unsigned short int nport,
    unsigned char type );
