/************************************************************************
 *                                                                      *
 *  pci180x.c  ---  real time daq kernel functions.                       *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/io.h>
#include "pci180x.h"

/***********************************************************************
 * some general things                                                 *
 ***********************************************************************/

/* Debug macros for DAQ pci180x PCI180X
 */
//#define DEBUG
#ifdef DEBUG
    #define DPRINTK(args...)    printk(## args)
#else
    #define DPRINTK(args...)
#endif

/************************************************************************
 *                                                                      * 
 *  Settings on the DAQ pci180x the following driver definitions are      *
 *  valid for:                                                          *
 *                                                                      *
 *    JP#    Pos       Meaning                                          *
 *                                                                      * 
 ************************************************************************/

/* select board type here
 */
#define PCI1800     /* PCI1800 or PCI1802, number of channels */
#ifdef PCI1800
    #define PCI180X_NAME  ("pci1800:") 
#endif
#ifdef PCI1802
    #define PCI180X_NAME  ("pci1802:") 
#endif


/************************************************************************
 *  pci180x initialisation                                                * 
 ************************************************************************/

/* PCI- series has no sub ids.  PCI-1202, 1602, 1800, and 1802 
 * have the same vendor id and device id. 
 */
#define PCI180X_VENDOR_ID  0x00001234  /* as used in pci_find_device */
#define PCI180X_DEVICE_ID  0x00005678  /* as used in pci_find_device */


/************************************************************************
 *  control stuff                                                       * 
 ************************************************************************/

/* PCI control register range and offsets
 */
#define PCI180X_PCI_RANGE      0x0040 /* PCI address range */
#define PCI180X_PCI_MBOX_ADDR  0x0000 /* 32bit register    */
#define PCI180X_INT_CTRL_ADDR  0x0038 /* 32bit register    */
#define PCI180X_NVRAM_1_ADDR   0x003c /* 32bit register    */
#define PCI180X_NVRAM_2_ADDR   0x003e /* 8bit register     */
#define PCI180X_NVRAM_3_ADDR   0x003f /* 8bit register     */

/* PCI control register values
 */
#define PCI180X_RESET           0x01 /* PCI reset         */
#define PCI180X_NVRAM_LOAI_LO   0x80 /* load low address  */
#define PCI180X_NVRAM_LOAI_HI   0xa0 /* load high address */
#define PCI180X_NVRAM_READ      0xc0 /* begin read        */
#define PCI180X_NVRAM_WRITE     0xe0 /* begin write       */
#define PCI180X_NVRAM_STATUS    0x80 /* 1->busy, 0->ready */

/* control register range and offset
 */
#define PCI180X_CONTROL_RANGE   0x08    /* control address range */
#define PCI180X_CONTROL_ADDR    0x0000  /* 8/16/32 bits */
#define PCI180X_STATUS_ADDR     0x0000  /* 8/16/32 bits */
#define PCI180X_SW_ADC_ADDR     0x0004  /* 8/16/32 bits */

/* PIC control and oth control values
 */
#define PCI180X_RESET_PIC       0xc000     /* ??? */
#define PCI180X_RECOVERY        0xffff     /* ??? */
#define PCI180X_SET_LOW         0xdfff     /* ??? */
#define PCI180X_SET_HIGH        0x2000     /* ??? */
#define PCI180X_TIMEOUT         1000000    /* int */
#define PCI180X_CONFIG_CHANNEL  0x8400   /* channel config command     */
#define PCI180X_CLEAR_FIFO      0x2000   /* clear fifo, no PIC command */
#define PCI180X_RESET_FIFO      0xa000   /* reset fifo, no PIC command */
#define PCI180X_ADD_2_SCAN      0xd000   /* add to scan PIC command */
#define PCI180X_START_SCAN      0xd400   /* start scan PIC command */
#define PCI180X_ADC_START       0xffff   /* start ADC in software mode */
#define PCI180X_ADC_TIMEOUT     1000000  /* int */
#define PCI180X_ADC_TIME        3000     /* nsecs */

/* status register meanings
 */
#define PCI180X_TIMER2_FLAG  0x01     /* bit 0: 1 if timer is up    */
#define PCI180X_ODM_VERSION  0x02     /* bit 1: 1 if ODM version    */
#define PCI180X_HANDSHAKE    0x04     /* bit 2: PIC Magic scan      */
#define PCI180X_EXT_TRIGGER  0x08     /* bit 3: 1 if timer 1 enable */
#define PCI180X_ADC_BUSY     0x10     /* bit 4: 1 if ADC busy       */
#define PCI180X_FIFO_EMPTY   0x20     /* bit 5: 1 if FIFO is empty  */
#define PCI180X_FIFO_FULL    0x40     /* bit 6: 1 if FIFO falf full */
#define PCI180X_FIFO_HALF    0x80     /* bit 7: 1 if FIFO falf full */


/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

#define PCI180X_AIO_RANGE     0x0004  /* AI read address      */
#define PCI180X_DIO_RANGE     0x0008  /* AI read address      */

/* AI register range and offsets
 */
#define PCI180X_AI_ADDR       0x0000  /* AI read address      */
#define PCI180X_AI_START      0xffff  /* start AD conversion  */
#define PCI180X_AI_BUSY       0x10    /* bit 4: 1 if ADC busy */
#define PCI180X_AI_CHANNELS   16      /* in single ended      */

/* DI register range and offsets
 */
#define PCI180X_DI_PORTS      2       /* number of input ports */
#define PCI180X_DI_DADDR      { 0x0000, 0x0001 } /* addresses  */
#define PCI180X_DI_NBITS      8       /* number of channels    */

/* DO register range and offsets
 */
#define PCI180X_DO_PORTS      2       /* number of output ports */
#define PCI180X_DO_DADDR      { 0x0000, 0x0001 } /* addresses   */
#define PCI180X_DO_NBITS      8       /* number of channels     */


/************************************************************************
 *  timer/counter stuff                                                 * 
 ************************************************************************/

#define PCI180X_COUNTER_RANGE     0x0010 /* timer address range   */

/* counter chips (8254 timer/counter chip) 
 */
#define PCI180X_COUNTER_CHIPS 1       /* number of counter chips     */
#define PCI180X_COUNTER_CPERC 3       /* number of counter per chip  */
#define PCI180X_COUNTER_TOTAL ( PCI180X_COUNTER_CPERC*PCI180X_COUNTER_CHIPS )
#define PCI180X_COUNTER_BASES { 0x0000 } /* chip address             */
#define PCI180X_COUNTER_DADDR { 0x0000, 0x0004, 0x0008 }
#define PCI180X_COUNTER_CADDR 0x000c  /* COUNTER control             */

/* counter control register words (8254 timer/counter) 
 */
#define PCI180X_COUNTER_BIN   0x00    /* binary count mode             */
#define PCI180X_COUNTER_BCD   0x01    /* BCD count mode                */
#define PCI180X_COUNTER_LATCH 0x00    /* counter latch instruction     */
#define PCI180X_COUNTER_LO    0x10    /* read/write low counter byte   */
#define PCI180X_COUNTER_HI    0x20    /* read/write high counter byte  */
#define PCI180X_COUNTER_LOHI  0x30    /* read/write low, then high     */
#define PCI180X_COUNTER_SEL   { 0x00, 0x40, 0x80 } /* select counter i */
#define PCI180X_COUNTER_RBACK 0xc0    /* counter read back             */

/* in general, there are some 8254 timer/counter with several counters 
 * each. Some counters are used as such, others are used as timers, single
 * or by combination of two counters. For AI, a timer is usually used
 * to serve as pacer. For an application one has:
 */
#define PCI180X_TIMER_BASE    125     /* Clock time in nanosecs        */
#define PCI180X_TIMER_TIMEOUT 10000   /* Timer                         */
#define PCI180X_TIMER_16      2       /* available 16bit timers        */
#define PCI180X_TIMER_16C     {0,1}   /* counter used for 16 bit timer */

#define PCI180X_PACER_ADDR    10


/************************************************************************
 * global values                                                        * 
 ************************************************************************/

struct
    {
    unsigned char di_port[PCI180X_DI_PORTS];
    unsigned char do_port[PCI180X_DO_PORTS];
    unsigned short int ai_mode;
    unsigned short int ai_active;
    unsigned short int ai_range[PCI180X_AI_CHANNELS];    
    unsigned long int  pci_phys; 	    /* physical addresses */
    unsigned long int  counter_phys;	/* physical addresses */
    unsigned long int  control_phys;    /* physical addresses */
    unsigned long int  dio_phys;        /* physical addresses */
    unsigned long int  aio_phys;        /* physical addresses */
    unsigned short int pci_io;          /* 16bit io address */
    unsigned short int counter_io;      /* 16bit io address */
    unsigned short int control_io;      /* 16bit io address */
    unsigned short int dio_io;          /* 16bit io address */
    unsigned short int aio_io;          /* 16bit io address */
    } 
    pci180x;


/***********************************************************************
 * Control functions ...                                               *
 ***********************************************************************/


/* busy wait loop on status register while status byte AND
 * "mask" is 0, otherwise "timeout"
 */
inline int pci180x_wait_while( unsigned char mask, unsigned int timeout )
    {
    int time = 0;
    
    DPRINTK( "%s ..wait_while\n", PCI180X_NAME ); 
    
    while ( ( inb( pci180x.control_io+PCI180X_STATUS_ADDR ) & mask ) == 0 )
        {
        time++;
        if ( time > timeout )
            {
            DPRINTK( "%s %d > %d, wait while loop timeout\n", 
                PCI180X_NAME, time, timeout ); 
            return -ETIME;
            }
        }
    
    DPRINTK( "%s %d < %d, wait while loop OK\n", PCI180X_NAME, 
        time, timeout ); 
    return 0;
    }

/* busy wait loop on status register until status byte AND
 * "mask" is 0, otherwise "timeout"
 */
inline int pci180x_wait_until( unsigned char mask, unsigned int timeout )
    {
    int time = 0;

    DPRINTK( "%s ..wait_until\n", PCI180X_NAME ); 
    
    while ( ( inb( pci180x.control_io+PCI180X_STATUS_ADDR ) & mask ) != 0 )
        {
        time++;
        if ( time > timeout )
            {
            DPRINTK( "%s %d > %d, wait while loop timeout\n", 
                PCI180X_NAME, time, timeout ); 
            return -ETIME;
            }
        }
    DPRINTK( "%s %d < %d, wait while loop OK\n", PCI180X_NAME, 
        time, timeout ); 
    return 0;
    }


/* set counter "counter" with divisor and mode
 */
void pci180x_counter_set( unsigned int counter, 
    unsigned int divisor, unsigned char mode )
    {
    int nchan, nchip;
    unsigned short int counter_bases[PCI180X_COUNTER_CHIPS] = PCI180X_COUNTER_BASES;
    unsigned short int counter_daddr[PCI180X_COUNTER_CPERC] = PCI180X_COUNTER_DADDR;
    unsigned char counter_sel[PCI180X_COUNTER_CPERC] = PCI180X_COUNTER_SEL;
    
    counter %= PCI180X_COUNTER_TOTAL;
    nchip   = counter / PCI180X_COUNTER_CPERC;
    nchan   = counter % PCI180X_COUNTER_CPERC;
    
    outb( (mode|counter_sel[nchan]|PCI180X_COUNTER_LOHI|PCI180X_COUNTER_BIN), 
        pci180x.counter_io+counter_bases[nchip]+PCI180X_COUNTER_CADDR );
    outb( (unsigned char)(divisor),	   
        pci180x.counter_io+counter_bases[nchip]+counter_daddr[nchan] );
    outb( (unsigned char)(divisor>>8),
        pci180x.counter_io+counter_bases[nchip]+counter_daddr[nchan] );
    }

/* start timer using coutners with "time" in nanosecs and mode!
 */
unsigned int pci180x_timer16_start( unsigned int timer,
    unsigned int time, unsigned char mode )
    {
    unsigned int divisor;
    int timer_16c[PCI180X_TIMER_16] = PCI180X_TIMER_16C;
    
    divisor = time/PCI180X_TIMER_BASE;
    
    timer %= PCI180X_TIMER_16;
    pci180x_counter_set( timer_16c[timer], time/PCI180X_TIMER_BASE, mode );
    
    return divisor*PCI180X_TIMER_BASE;
    }

/* stop timer using coutners
 */
void pci180x_timer16_stop( unsigned int timer )
    {
    int timer_16c[PCI180X_TIMER_16] = PCI180X_TIMER_16C;
    
    timer %= PCI180X_TIMER_16;
    pci180x_counter_set( timer_16c[timer], 0x0000, 0x00 );
    }

/* enable timer 2 with delay time in nanosec
 * with real time linux, this is actually redundant
 */
inline int pci180x_timer_delay( unsigned int time )
    {
    DPRINTK( "%s ..timer2_delay\n", PCI180X_NAME ); 

    pci180x_counter_set( 2, time/PCI180X_TIMER_BASE, PCI180X_COUNTER_MODE1 );
    
    if ( pci180x_wait_while( PCI180X_TIMER2_FLAG, PCI180X_TIMER_TIMEOUT ) < 0 )
        {
        DPRINTK( "%s timer 2 timeout\n", PCI180X_NAME ); 
        return -ETIME;
        }
    
    return 0;
    }


/*PIC control
 */
int pci180x_pic_control( unsigned short int word )
    {
    DPRINTK( "%s ..pic_control\n", PCI180X_NAME ); 

    /* check PIC, send recovery and wait until OK
     */
    if ( ( inb( pci180x.control_io+PCI180X_STATUS_ADDR ) 
           & PCI180X_HANDSHAKE ) == 0 )
        outw( PCI180X_RECOVERY, pci180x.control_io+PCI180X_CONTROL_ADDR );   
    
    if ( pci180x_wait_while( PCI180X_HANDSHAKE, PCI180X_TIMEOUT ) < 0 )
        {
        DPRINTK( "%s PIC timeout 1\n", PCI180X_NAME ); 
        return -ETIME;
        }
    
    /* set pic low, and wait while PIC is 1, ie ~0, STRANGE!!!
     */
    word = word & PCI180X_SET_LOW;	
    outw( word, pci180x.control_io+PCI180X_CONTROL_ADDR );
    if ( pci180x_wait_until( PCI180X_HANDSHAKE, PCI180X_TIMEOUT ) < 0 )
        {
        DPRINTK( "%s PIC timeout 2\n", PCI180X_NAME ); 
        return -ETIME;
        }
  
    /* set pic high, and wait until PIC is OK
     */
    outw( ( word | PCI180X_SET_HIGH ), 
        pci180x.control_io+PCI180X_CONTROL_ADDR );
    if ( pci180x_wait_while( PCI180X_HANDSHAKE, PCI180X_TIMEOUT ) < 0 )
        {
        DPRINTK( "%s PIC timeout 3\n", PCI180X_NAME ); 
        return -ETIME;
        }

    return 0;
    }


/***********************************************************************
 * Analog input basic functions ...                                    *
 ***********************************************************************/

/* select channel for successive conversion
 */
int pci180x_ai_adc_select( unsigned short int channel )
    {
    channel %= PCI180X_AI_CHANNELS;
    pci180x.ai_active = channel;
    
    /* program MUX for to use this channel
     */
    if ( pci180x_pic_control( channel | ( pci180x.ai_range[channel] << 6 ) 
        | PCI180X_CONFIG_CHANNEL ) < 0 )
        {
        DPRINTK( "%s PIC timeout\n", PCI180X_NAME ); 
        return -ETIME;
        }

#ifdef PCI180X_HIGH_GAIN
    switch ( range )
        {
        case PCI180X_AI_PM5:
        case PCI180X_AI_PM10:
        case PCI180X_AI_0P10:
            DPRINTK( "%s ... settling time = 23us\n", PCI180X_NAME ); 
            if ( pci180x_timer_delay( 23 ) )
                return -ETIME;
            break;
        
        case PCI180X_AI_PM05:
        case PCI180X_AI_PM1:
        case PCI180X_AI_0P1:
            DPRINTK( "%s ... settling time = 28us\n", PCI180X_NAME ); 
            if ( pci180x_timer_delay( 28 ) )
                return -ETIME;
            break;

        case PCI180X_AI_PM005:
        case PCI180X_AI_0P01:
        case PCI180X_AI_PM01:
            DPRINTK( "%s ... settling time = 140us\n", PCI180X_NAME ); 
            if ( pci180x_timer_delay( 140 ) )
                return -ETIME;
            break;

        case PCI180X_AI_PM0005:
        case PCI180X_AI_PM001:
        case PCI180X_AI_0P001:
            DPRINTK( "%s ... settling time = 1300us\n", PCI180X_NAME ); 
            if ( pci180x_timer_delay( 1300 ) )
                return -ETIME;
            break;

    default:
            /* we should not arrive here
            */
            DPRINTK( "%s range not correct\n", PCI180X_NAME ); 
            break;
        }
#endif
#ifdef PCI180X_LOW_GAIN
    DPRINTK( "%s ... settling time = 3 us\n", PCI180X_NAME ); 
    if ( pci180x_timer_delay( 3 ) )
        return -ETIME;
#endif
    return 0; 
    }

/* program ADC for successive conversion, with both channel 
 * and range as defined in header file.
 */
int pci180x_ai_adc_program( unsigned short int channel,
    unsigned short int range )
    {
    channel %= PCI180X_AI_CHANNELS;
    pci180x.ai_range[channel] = range;

    /* program ADC range/gain for this channel 
     * and set thi channel active on pci180x ! 
     */
    if ( pci180x_ai_adc_select( channel ) < 0 )
        {
        DPRINTK( "%s ..adc_program timeout\n", PCI180X_NAME ); 
        return -ETIME;
        }

    DPRINTK( "%s ..adc_range pci180x.ai_range[%d] = %d\n", PCI180X_NAME, 
        channel, pci180x.ai_range[channel] ); 
    
    return 0;
    }

/* get a single analog value mapped to 16bit independently of ADC,
 * assuming that ADC is terminated ( or/and value is in FIFO).
 */
inline unsigned short int pci180x_ai_get_value( void )
    {
    /* get values from register (or FIFO)
     */
    return ( inw( pci180x.aio_io+PCI180X_AI_ADDR ) << 4 ); /* 12 to 16bit */
    }

/* get a single analog value given that channel is already
 * selected and ints range set, mode is software.
 */
int pci180x_ai_get_value0( unsigned short int *value )
    {
    /* Clear and reset FIFO 
     */
    DPRINTK( "%s ..aget clear FIFO and trigger ADC \n", PCI180X_NAME ); 
    outw( PCI180X_CLEAR_FIFO, pci180x.control_io+PCI180X_CONTROL_ADDR );
    outw( PCI180X_RESET_FIFO, pci180x.control_io+PCI180X_CONTROL_ADDR );
    outw( PCI180X_ADC_START,  pci180x.control_io+PCI180X_SW_ADC_ADDR  );

    /* Busy wait for EOC
     */
    DPRINTK( "%s ..aget wait for ADC ready \n", PCI180X_NAME ); 
    if ( pci180x_wait_while( PCI180X_ADC_BUSY, PCI180X_ADC_TIMEOUT ) < 0 )
        {
        DPRINTK( "%s ADC timeout\n", PCI180X_NAME ); 
        return -1;
        }
    
    *value = pci180x_ai_get_value();

    return 0;
    }

/* get a single analog value( channel ) given that its range set.
 */
int pci180x_ai_get_value1( unsigned short int *value,
    unsigned short int channel )
    {
    pci180x_ai_adc_select( channel );
    
    return pci180x_ai_get_value0( value );
    }

/* get a single analog value( channel, range ).
 */
int pci180x_ai_get_value2( unsigned short int *value, 
    unsigned short int channel, unsigned short int range )
    {
    pci180x_ai_adc_program( channel, range );

    return pci180x_ai_get_value0( value );
    }

/* Add a channel from a list to scan 
 */
int pci180x_add_2_scan( unsigned short int channel )
    {
    unsigned short int word = 0x00;
    
    channel %= PCI180X_AI_CHANNELS;

    word += ( pci180x.ai_range[channel] << 6 );
    word += PCI180X_ADD_2_SCAN;
    word += channel;

    return pci180x_pic_control( word );
    }
         

/* start a channel scan with list predefined and sampling time[us]
 * mode has to be set appropriately, extern (pre,post,middle) or internal
 */
int pci180x_ai_start_scan( unsigned int time, unsigned char mode )
    {
    int result;
    
    /* Program PIC for scan
     */
    result = pci180x_pic_control( PCI180X_START_SCAN );
    if ( result < 0 ) return result;
     
    /* Clear and reset FIFO
     */
    outw( PCI180X_CLEAR_FIFO, pci180x.control_io+PCI180X_CONTROL_ADDR );
    outw( PCI180X_RESET_FIFO, pci180x.control_io+PCI180X_CONTROL_ADDR );
          
    /* Start timer according to mode
     */
    if ( mode == PCI180X_INTERNAL_MODE )
        pci180x_timer16_start( 0, time, PCI180X_COUNTER_MODE2 );
    else if ( mode == PCI180X_EXTERNAL_MODE )
        pci180x_timer16_start( 1, time, PCI180X_COUNTER_MODE2 );

    return 0;
    }

/* Once the scan has been started, the application is responsible
 * for picking up values from the fifos periodically. If an interupt
 * is available, this can be done by the respective routine. The
 * rest depends on the application.
 */
int pci180x_ai_get_scan( unsigned short int *value )
    {
    /* check if FIFO is empty ...
     */
    if ( ( inb( pci180x.control_io+PCI180X_STATUS_ADDR ) & PCI180X_FIFO_EMPTY ) == 0 )
        return 0;
    
    /* if not empty, then :
     */
    *value = pci180x_ai_get_value();

    return 1;
    }

/* Stop analogue input scan, wbasically stop the timer.
 */
int pci180x_ai_stop_scan( void )
    {
    /* Stop timer and reset PIC
     */
    pci180x_timer16_stop( 0 );
    pci180x_timer16_stop( 1 );
    
    return pci180x_pic_control( PCI180X_RESET_PIC );
    }


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
unsigned char pci180x_di_get_byte( unsigned short int nport )
    {
    unsigned short int di_addr[PCI180X_DI_PORTS] = PCI180X_DI_DADDR;

    nport %= PCI180X_DI_PORTS; /* just in case */

    pci180x.di_port[nport] = inb( pci180x.dio_io + di_addr[nport] );
            
    return pci180x.di_port[nport];    
    }

/* get bit #"nbit" from port "nport".
 */ 
unsigned short int pci180x_di_get_bit( unsigned short int nport, 
    unsigned short int nbit )
    {
    nbit  %= PCI180X_DI_NBITS; /* just in case */

    return ( (unsigned short int) 
        ( ( ( pci180x_di_get_byte( nport ) & ( 1 << nbit ) ) >> nbit ) ) );
    } 

/* set byte "byte" on port "nport"
 */ 
void pci180x_do_set_byte( unsigned char byte, 
    unsigned short int nport )
    {
    unsigned short int do_addr[PCI180X_DO_PORTS] = PCI180X_DO_DADDR;

    nport %= PCI180X_DO_PORTS; /* just in case */

    pci180x.do_port[nport] = byte; /* keep a copy */

    outb( pci180x.do_port[nport], pci180x.dio_io + do_addr[nport] );
    }

/* set bit "bit" with #"nbit" on port "nport".
 */ 
void pci180x_do_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit )
    {
    nport %= PCI180X_DO_PORTS; /* just in case */
    nbit  %= PCI180X_DO_NBITS; /* just in case */
    bit   &= 0x01;           /* just in case */

    pci180x.do_port[nport] &= ~(1 << nbit);
    pci180x.do_port[nport] |=  (bit << nbit); 
    
    pci180x_do_set_byte( pci180x.do_port[nport], nport );
    }


/***********************************************************************
 *  pci180x defaults                                                     *
 ***********************************************************************/

static void pci180x_defaults( void )
    {
    int i;
        
    /* reset PIC
     */
    pci180x_pic_control( PCI180X_RESET_PIC );
    printk( "%s DAQ board reset\n", PCI180X_NAME );

    for ( i=0; i<PCI180X_COUNTER_TOTAL; i++ )
        pci180x_counter_set( i, 0x00, 0x00 );

    /* init ranges for analog inputs, if possible
     */
    for ( i=0; i<PCI180X_AI_CHANNELS; i++ )
        pci180x_ai_adc_program( i, PCI180X_AI_PM5 );   

    /* init values for digital output 
     */
    for ( i=0; i<PCI180X_DO_PORTS; i++ )
        pci180x_do_set_byte( 0x00, i );   
    }


/***********************************************************************
 * Linux kernel functions                                              *
 ***********************************************************************/

/*the mandatory init_module.
 */
int init_module( void )
    {
    struct pci_dev *dev;

    /* Check presence of PCI bus, cela se discute.
     */
    if( !pci_present() ) 
        {
        printk("%s PCI bus not present!\n", PCI180X_NAME );
        return -ENODEV;
        }
    
    /* find the first device of this kind
     */
    dev =  pci_find_device( PCI180X_VENDOR_ID, PCI180X_DEVICE_ID, NULL ); 
    if ( !dev )
        {
        printk("%s no board with vendor ID %x and device ID %x found\n", 
            PCI180X_NAME, PCI180X_VENDOR_ID, PCI180X_DEVICE_ID );
        return -ENODEV;
        }

    /* PCI- series has no sub ids.  PCI-1202, 1602, 1800, and 1802 
     * have the same vendor id and device id. 
     */
    printk("%s with vendor ID %x and device ID %x found\n", 
        PCI180X_NAME, PCI180X_VENDOR_ID, PCI180X_DEVICE_ID );
    printk("%s IRQ is selected to be %d (0 = no interrupt)\n", 
        PCI180X_NAME, dev->irq );

    /* request PCI io addresses
     */
    pci180x.pci_phys = dev->base_address[0] & PCI_BASE_ADDRESS_IO_MASK;
    if ( check_region( pci180x.pci_phys, PCI180X_PCI_RANGE ) < 0 ) 
        {
     	printk( "%s could not request pci io addresss\n", PCI180X_NAME );
        return -EIO;
        }
    request_region( pci180x.pci_phys, PCI180X_PCI_RANGE, PCI180X_NAME );
    pci180x.pci_io = (unsigned short int)(pci180x.pci_phys);
    printk( "%s requested pci io address 0x%x\n", PCI180X_NAME, pci180x.pci_io );

    /* request timer io addresses
     */
    pci180x.counter_phys = dev->base_address[1] & PCI_BASE_ADDRESS_IO_MASK;
    if ( check_region( pci180x.counter_phys, PCI180X_COUNTER_RANGE ) < 0 ) 
        {
     	printk( "%s could not request counter io addresses\n", PCI180X_NAME );
        return -EIO;
        }
    request_region( pci180x.counter_phys, PCI180X_COUNTER_RANGE, PCI180X_NAME );
    pci180x.counter_io = (unsigned short int)(pci180x.counter_phys);
    printk( "%s requested counter io address 0x%x\n", PCI180X_NAME, pci180x.counter_io );

    /* request control io addresses
     */
    pci180x.control_phys = dev->base_address[2] & PCI_BASE_ADDRESS_IO_MASK;
    if ( check_region( pci180x.control_phys, PCI180X_CONTROL_RANGE ) < 0 ) 
        {
     	printk( "%s could not request control io addresses\n", PCI180X_NAME );
        return -EIO;
        }
    request_region( pci180x.control_phys, PCI180X_CONTROL_RANGE, PCI180X_NAME );
    pci180x.control_io = (unsigned short int)(pci180x.control_phys); 
    printk( "%s requested control io address 0x%x\n", PCI180X_NAME, pci180x.control_io );

    /* request digital io addresses
     */
    pci180x.dio_phys = dev->base_address[3] & PCI_BASE_ADDRESS_IO_MASK;
    if ( check_region( pci180x.dio_phys, PCI180X_DIO_RANGE ) < 0 ) 
        {
     	printk( "%s could not request digital io addresses\n", PCI180X_NAME );
        return -EIO;
        }
    request_region( pci180x.dio_phys, PCI180X_DIO_RANGE, PCI180X_NAME );
    pci180x.dio_io = (unsigned short int)(pci180x.dio_phys);
    printk( "%s requested digital io address 0x%x\n", PCI180X_NAME, pci180x.dio_io );

    /* request analog io addresses
     */
    pci180x.aio_phys = dev->base_address[4] & PCI_BASE_ADDRESS_IO_MASK;
    if ( check_region( pci180x.aio_phys, PCI180X_AIO_RANGE ) < 0 ) 
        {
     	printk( "%s could not request analog io addresses\n", PCI180X_NAME );
        return -EIO;
        }
    request_region( pci180x.aio_phys, PCI180X_AIO_RANGE, PCI180X_NAME );
    pci180x.aio_io = (unsigned short int)(pci180x.aio_phys);
    printk( "%s requested analog io address 0x%x\n", PCI180X_NAME, pci180x.aio_io );
    
    /* Set board defaults
     */
    pci180x_defaults();

    return 0;
    } 

/*the mandatory cleanup_module.
 */
void cleanup_module( void )
    {
    /* Set board defaults
     */
    pci180x_defaults();
    
    release_region( pci180x.pci_phys, PCI180X_PCI_RANGE );
    printk( "%s released pci io address 0x%x\n", PCI180X_NAME, pci180x.pci_io );
    
    release_region( pci180x.counter_phys, PCI180X_COUNTER_RANGE );
    printk( "%s released timer io address 0x%x\n", PCI180X_NAME, pci180x.counter_io );
    
    release_region( pci180x.control_phys, PCI180X_CONTROL_RANGE );
    printk( "%s released control io address 0x%x\n", PCI180X_NAME, pci180x.control_io );
    
    release_region( pci180x.dio_phys, PCI180X_DIO_RANGE );
    printk( "%s released digital io address 0x%x\n", PCI180X_NAME, pci180x.dio_io );
    
    release_region( pci180x.aio_phys, PCI180X_AIO_RANGE );
    printk( "%s released analog io address 0x%x\n", PCI180X_NAME, pci180x.aio_io );
    
    printk( "%s DAQ board released\n", PCI180X_NAME );
    }
   
