/************************************************************************
 *                                                                      *
 *  pci180x.h  ---  daq inputs header file for pci180x.c.               *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

#define PCI180X_NOOUTPUTS
#define PCI180X_LOW_GAIN

/************************************************************************
 *  control stuff                                                       * 
 ************************************************************************/

/* counter modes
 */
#define PCI180X_COUNTER_MODE0 0x00    /* interrupt on terminal count */
#define PCI180X_COUNTER_MODE1 0x02    /* programmable oneshot        */
#define PCI180X_COUNTER_MODE2 0x04    /* rate generator              */
#define PCI180X_COUNTER_MODE3 0x06    /* square wave generator       */
#define PCI180X_COUNTER_MODE4 0x08    /* software triggered pulse    */
#define PCI180X_COUNTER_MODE5 0x0a    /* hardware triggered pulse    */


/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

/* AI range
 */
#ifdef PCI180X_LOW_GAIN
    #define PCI180X_AI_PM10     0x04     /* [ -10.0,   10.0   ] Volt */
    #define PCI180X_AI_PM5      0x05     /* [  -5.0,    5.0   ] Volt */
    #define PCI180X_AI_PM25     0x06     /* [  -2.5,    2.5   ] Volt */
    #define PCI180X_AI_PM125    0x07     /* [  -1.25,   1.25  ] Volt */
    #define PCI180X_AI_PM0625   0x03     /* [  -0.625,  0.625 ] Volt */
    #define PCI180X_AI_0P10     0x08     /* [   0.0,   10.0   ] Volt */
    #define PCI180X_AI_0P5      0x09     /* [   0.0,    5.0   ] Volt */
    #define PCI180X_AI_0P25     0x0a     /* [   0.0,    2.5   ] Volt */
    #define PCI180X_AI_0P125    0x0b     /* [   0.0,    1.25  ] Volt */
#endif

#ifdef PCI180X_HIGH_GAIN
    #define PCI180X_AI_PM5      0x10      /* [  -5.0,    5.0   ] Volt */
    #define PCI180X_AI_PM05     0x11      /* [  -0.5,    0.5   ] Volt */
    #define PCI180X_AI_PM005    0x12      /* [  -0.05,   0.05  ] Volt */
    #define PCI180X_AI_PM0005   0x13      /* [  -0.005,  0.005 ] Volt */
    #define PCI180X_AI_PM10     0x14      /* [ -10.0,   10.0   ] Volt */
    #define PCI180X_AI_PM1      0x15      /* [  -1.0,    1.0   ] Volt */
    #define PCI180X_AI_PM01     0x16      /* [  -0.1,    0.1   ] Volt */
    #define PCI180X_AI_PM001    0x17      /* [  -0.01,   0.01  ] Volt */
    #define PCI180X_AI_0P10     0x18      /* [   0.0,   10.0   ] Volt */
    #define PCI180X_AI_0P1      0x19      /* [   0.0,    1.0   ] Volt */
    #define PCI180X_AI_0P01     0x1a      /* [   0.0,    0.1   ] Volt */
    #define PCI180X_AI_0P001    0x1b      /* [   0.0,    0.01  ] Volt */
#endif

#define PCI180X_AI_FIFO   8192        /* input FIFO size */

#define PCI180X_AI_EXTPACER   0       /* use timer O as external pacer */
#define PCI180X_AI_INTPACER   1       /* use timer 1 as internal pacer */

/* ADC mode
 */
#define PCI180X_SOFTWARE_MODE     0x00    /* software trigger mode */
#define PCI180X_EXTERNAL_MODE     0x02    /* external trigger mode */
#define PCI180X_INTERNAL_MODE     0x03    /* internal trigger mode */

/* AO range
 */
#define PCI180X_AO_PM5    0x0000      /* [  -5.0,    5.0   ] Volt */
#define PCI180X_AO_PM10   0x0100      /* [ -10.0,   10.0   ] Volt */

/* DIO port type
 */
#define PCI180X_DIO_INPUT     0x00    /* dio port is an output */
#define PCI180X_DIO_OUTPUT    0x01    /* dio port is an input  */


/***********************************************************************
 * Control functions ...                                               *
 ***********************************************************************/

/* start counter "counter" with divisor and mode
 */
extern void pci180x_counter_set( unsigned int counter, 
    unsigned int divisor, unsigned char mode );

/* start timer using coutners with "time" in nanosecs and mode!
 */
extern unsigned int pci180x_timer16_start( unsigned int timer,
    unsigned int time, unsigned char mode );

/* stop timer using coutners
 */
extern void pci180x_timer16_stop( unsigned int timer );


/***********************************************************************
 * Analog input basic functions ...                                    *
 ***********************************************************************/

/* select channel for successive conversion
 */
extern int pci180x_ai_adc_select( unsigned short int channel );

/* program ADC for successive conversion, with both channel 
 * and range as defined in header file.
 */
extern int pci180x_ai_adc_program( unsigned short int channel,
    unsigned short int range );

/* get a single analog value given that channel is already
 * selected and ints range set, mode is software.
 */
extern int pci180x_ai_get_value0( unsigned short int *value );

/* get a single analog value( channel ); given that its range set.
 */
extern int pci180x_ai_get_value1( unsigned short int *value,
    unsigned short int channel );

/* get a single analog value( channel, range );.
 */
extern int pci180x_ai_get_value2( unsigned short int *value,
    unsigned short int channel, unsigned short int range );

/* Add a channel from a list to scan 
 */
extern int pci180x_add_2_scan( unsigned short int channel );

/* start a channel scan with list, number of channels and sampling time[us]
 * mode has to be set appropriately, extern (pre,post,middle) or internal
 */
extern int pci180x_ai_start_scan( unsigned int time, unsigned char mode );

/* Once the scan has been started, the application is responsible
 * for picking up values from the fifos periodically. If an interupt
 * is available, this can be done by the respective routine. The
 * rest depends on the application.
 */
extern int pci180x_ai_get_scan( unsigned short int *value );

/* Stop analogue input scan, wbasically stop the timer.
 */
extern int pci180x_ai_stop_scan( void );


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* activate channel for successive DA conversion
 */
extern void pci180x_ao_dac_select( unsigned short int channel );

/* program DAC for successive DA conversion with both channel 
 * and range as defined in header file.
 */
extern void pci180x_ao_dac_program( unsigned short int channel,
    unsigned short int range );

/* set a single analog value given that channel is already
 * selected and ints range set value mapped to 16bit.
 */
extern void pci180x_ao_set_value0( unsigned short int value );

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
extern void pci180x_ao_set_value1( unsigned short int value, 
    unsigned short int channel );

/* set a single analog value (channel,range).
 * value is mapped to 16bit independently of DAC.
 */
extern void pci180x_ao_set_value2( unsigned short int value,
    unsigned short int channel, unsigned short int range );


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
extern unsigned char pci180x_di_get_byte( unsigned short int nport );

/* get bit #"nbit" from port "nport".
 */ 
extern unsigned short int pci180x_di_get_bit( unsigned short int nport, 
    unsigned short int nbit );

/* set byte "byte" on port "nport"
 */ 
extern void pci180x_do_set_byte( unsigned char byte,
    unsigned short int nport );

/* set bit "bit" with #"nbit" on port "nport".
 */ 
extern void pci180x_do_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit );
