/************************************************************************
 *                                                                      *
 *  pciad16da4.c  ---  real time daq kernel functions.                  *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/io.h>
#include "pciad16da4.h"

/***********************************************************************
 * some general things                                                 *
 ***********************************************************************/

/* Debug macros for DAQ board PCIAD16DA4
 */
//#define DEBUG
#ifdef DEBUG
    #define DPRINTK(args...)    printk(## args)
#else
    #define DPRINTK(args...)
#endif

/************************************************************************
 *                                                                      * 
 *  Settings on the DAQ board the following driver definitions are      *
 *  valid for:                                                          *
 *                                                                      *
 *      JP#    Pos      Meaning                                         *
 *      JP1    SE       Single ended input mode                         *
 *      JP     +-5V                                                     * 
 *                                                                      * 
 ************************************************************************/

#define PCIAD16DA4_NAME  "pciad16da4: "

/************************************************************************
 *  pciad16da4 initialisation                                           * 
 ************************************************************************/

/* vendor and device id and range
 */
#define PCIAD16DA4_VENDOR_ID  0x00001001  /* as used in pci_find_device */
#define PCIAD16DA4_DEVICE_ID  0x00000012  /* as used in pci_find_device */
#define PCIAD16DA4_RANGE      0x0100      /* address range */


/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

/* AI register range and offsets
 */
#define PCIAD16DA4_AI_BASE        0x0000  /* AI read address        */
#define PCIAD16DA4_AI_ADDR        0x0000  /* AI read address        */
#define PCIAD16DA4_AI_START       0x00    /* ADC start              */
#define PCIAD16DA4_AI_READ        0x01    /* ADC high byte set      */
#define PCIAD16DA4_AI_HB          0x01    /* ADC high byte set      */
#define PCIAD16DA4_AI_LB          0x02    /* ADC low byte set       */
#define PCIAD16DA4_AI_EOC_ADDR    0x01    /* EOC polling            */
#define PCIAD16DA4_AI_EOC         0x01    /* EOC flag               */
#define PCIAD16DA4_AI_CHANNELS    16      /* in single ended        */
#define PCIAD16DA4_AI_ADCDELAY    35      /* 35 usecs conversion    */
#define PCIAD16DA4_AI_MUXDELAY    10      /* 10 usecs settling      */
#define PCIAD16DA4_AI_MUXADDR     0x04    /* MUX address            */

/* AO register range and offsets
 */
#define PCIAD16DA4_AO_XFER_DUMMY  0x1d    /* DAC register transfer  */
#define PCIAD16DA4_AO_BASE        0x0020  /* DAC wirte address      */
#define PCIAD16DA4_AO_ZERO        0x7fff  /* zero Volt at output    */
#define PCIAD16DA4_AO_CHANNELS    4       /* number of channels     */
#define PCIAD16DA4_AO_DACDELAY    10      /* 10 usecs conversion    */

/* DI register range and offsets
 */
#define PCIAD16DA4_DI_PORTS       1       /* number of input ports  */
#define PCIAD16DA4_DI_DADDR       { 0x0011 } /* addresses   */
#define PCIAD16DA4_DI_NBITS       1       /* number of channels     */

/* DO register range and offsets
 */
#define PCIAD16DA4_DO_PORTS       1       /* number of output ports */
#define PCIAD16DA4_DO_DADDR       { 0x0010 } /* addresses   */
#define PCIAD16DA4_DO_NBITS       1       /* number of channels     */


/************************************************************************
 * global values                                                        * 
 ************************************************************************/

struct
    {
    unsigned char di_port[PCIAD16DA4_DI_PORTS];
    unsigned char do_port[PCIAD16DA4_DO_PORTS];
    unsigned short int ai_active;
    unsigned short int ao_active;
    unsigned long  int base_phys;
    unsigned short int base_io;
    } 
    pciad16da4;


/***********************************************************************
 * Analog input basic functions ...                                    *
 ***********************************************************************/

/* select channel for successive conversion
 */
void pciad16da4_ai_adc_select( unsigned short int channel )
    {
    channel %= PCIAD16DA4_AI_CHANNELS;
    pciad16da4.ai_active = channel;
    
    /* program MUX for channel
     */
    outb( (unsigned char) ( channel % PCIAD16DA4_AI_CHANNELS ), 
       pciad16da4.base_io + PCIAD16DA4_AI_MUXADDR );

    /* We have to waiting here for 10us, mux settling time.
     */
    udelay( PCIAD16DA4_AI_MUXDELAY );
    }

/* get a single analog value mapped to 16bit independently of ADC,
 * assuming that ADC is terminated ( or/and value is in FIFO).
 */
inline unsigned short int pciad16da4_ai_get_value( void )
    {
    unsigned char hbyte, lbyte;

    /* read low and high bytes of sample
     */
    outb( PCIAD16DA4_AI_HB, pciad16da4.base_io+PCIAD16DA4_AI_BASE );
    hbyte = inb( pciad16da4.base_io+PCIAD16DA4_AI_BASE );
    
    /* read low and high bytes of sample 
     */
    outb( PCIAD16DA4_AI_LB, pciad16da4.base_io+PCIAD16DA4_AI_BASE );
    lbyte = inb( pciad16da4.base_io+PCIAD16DA4_AI_BASE );
    
    return ( (hbyte<<8) | lbyte );
    }

/* get a single analog value given that channel is already
 * selected and ints range set, mode is software.
 */
int pciad16da4_ai_get_value0( unsigned short int *value )
    {
    /* start ADC
     */
    outb( PCIAD16DA4_AI_START, pciad16da4.base_io+PCIAD16DA4_AI_BASE );

    /* wait for EOC
     */
    udelay( PCIAD16DA4_AI_ADCDELAY );

    *value = pciad16da4_ai_get_value();
    return 0;
    }

/* get a single analog value( channel ) given that its range set.
 */
int pciad16da4_ai_get_value1( unsigned short int *value,
    unsigned short int channel )
    {
    pciad16da4_ai_adc_select( channel );
    
    return pciad16da4_ai_get_value0( value );
    }


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* select channel for successive DA conversion
 */
void pciad16da4_ao_dac_select( unsigned short int channel )
    {
    pciad16da4.ao_active = channel;
    }

/* set a single analog value given that channel is already
 * selected and ints range set value mapped to 16bit.
 */
void pciad16da4_ao_set_value0( unsigned short int value )
    {
    unsigned short int ao_addr;
    
    /* addr using pciad16da4.ao_active
     */
    ao_addr = pciad16da4.base_io+PCIAD16DA4_AO_BASE+(pciad16da4.ao_active<<3);
    
    /* outb values by hbyte and lbyte to latch
     */
    outb( (unsigned char)(value),    ao_addr   );
    outb( (unsigned char)(value>>8), ao_addr+4 );
    
    /* read access to dummy register will transfer da-registers
     */
    (void) inb( pciad16da4.base_io+PCIAD16DA4_AO_XFER_DUMMY );

    /* We have to waiting here for 10us, DAC settling time.
     */
    udelay( PCIAD16DA4_AO_DACDELAY );
    }

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
void pciad16da4_ao_set_value1( unsigned short int value, 
    unsigned short int channel )
    {
    pciad16da4_ao_dac_select( channel );
    pciad16da4_ao_set_value0( value );
    }


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
unsigned char pciad16da4_di_get_byte( unsigned short int nport )
    {
    unsigned short int di_addr[PCIAD16DA4_DI_PORTS] = PCIAD16DA4_DI_DADDR;

    nport %= PCIAD16DA4_DI_PORTS; /* just in case */

    pciad16da4.di_port[nport] = inb( pciad16da4.base_io + di_addr[nport] );
            
    return pciad16da4.di_port[nport];    
    }

/* get bit #"nbit" from port "nport".
 */ 
unsigned short int pciad16da4_di_get_bit( unsigned short int nport, 
    unsigned short int nbit )
    {
    nbit  %= PCIAD16DA4_DI_NBITS; /* just in case */

    return ( (unsigned short int) 
        ( ( ( pciad16da4_di_get_byte( nport ) & ( 1 << nbit ) ) >> nbit ) ) );
    } 

/* set byte "byte" on port "nport"
 */ 
void pciad16da4_do_set_byte( unsigned char byte, 
    unsigned short int nport )
    {
    unsigned short int do_addr[PCIAD16DA4_DO_PORTS] = PCIAD16DA4_DO_DADDR;

    nport %= PCIAD16DA4_DO_PORTS; /* just in case */

    pciad16da4.do_port[nport] = byte; /* keep a copy */

    outb( pciad16da4.do_port[nport], pciad16da4.base_io + do_addr[nport] );
    }

/* set bit "bit" with #"nbit" on port "nport".
 */ 
void pciad16da4_do_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit )
    {
    nport %= PCIAD16DA4_DO_PORTS; /* just in case */
    nbit  %= PCIAD16DA4_DO_NBITS; /* just in case */
    bit   &= 0x01;           /* just in case */

    pciad16da4.do_port[nport] &= ~(1 << nbit);
    pciad16da4.do_port[nport] |=  (bit << nbit); 
    
    pciad16da4_do_set_byte( nport, pciad16da4.do_port[nport] );
    }


/***********************************************************************
 *  board defaults                                                     *
 ***********************************************************************/

static void pciad16da4_defaults( void )
    {
    int i;
        
    /* init values for analog output 
     */
    for ( i=0; i<PCIAD16DA4_AO_CHANNELS; i++ )
        pciad16da4_ao_set_value1( PCIAD16DA4_AO_ZERO, i );   

    /* init values for digital output 
     */
    for ( i=0; i<PCIAD16DA4_DO_PORTS; i++ )
        pciad16da4_do_set_byte( 0x00, i );   
    }


/***********************************************************************
 * Linux kernel functions                                              *
 ***********************************************************************/

/*the mandatory init_module.
 */
int init_module( void )
    {
    struct pci_dev *dev;

    /* Check presence of PCI bus, ca se discute.
     */
    if( !pci_present() ) 
        {
        printk("%s PCI bus not present!\n", PCIAD16DA4_NAME );
        return -ENODEV;
        }
    
    /* find the first device of this kind
     */
    dev =  pci_find_device( PCIAD16DA4_VENDOR_ID, PCIAD16DA4_DEVICE_ID, NULL ); 
    if ( !dev )
        {
        printk("%s no %s with vendor ID %x and device ID %x found\n", 
            PCIAD16DA4_NAME, PCIAD16DA4_NAME, PCIAD16DA4_VENDOR_ID, PCIAD16DA4_DEVICE_ID );
        return -ENODEV;
        }
    printk("%s %s with vendor ID %x and device ID %x found\n", 
        PCIAD16DA4_NAME, PCIAD16DA4_NAME, PCIAD16DA4_VENDOR_ID, PCIAD16DA4_DEVICE_ID );

    /* request PCI io addresses
     */
    pciad16da4.base_phys = dev->base_address[0] & PCI_BASE_ADDRESS_IO_MASK;
    if ( check_region( pciad16da4.base_phys, PCIAD16DA4_RANGE ) < 0 ) 
        {
     	printk( "%s could not request base io addresss\n", PCIAD16DA4_NAME );
        return -EIO;
        }
    request_region( pciad16da4.base_phys, PCIAD16DA4_RANGE, PCIAD16DA4_NAME );

    pciad16da4.base_io = (unsigned short int)(pciad16da4.base_phys);
    printk( "%s requested base io address 0x%x\n", PCIAD16DA4_NAME, 
        pciad16da4.base_io );

    /* Set pciad16da4 defaults
     */
    pciad16da4_defaults();

    return 0;
    } 

/*the mandatory cleanup_module.
 */
void cleanup_module( void )
    {
    /* Set pciad16da4 defaults
     */
    pciad16da4_defaults();
    
    release_region( pciad16da4.base_phys, PCIAD16DA4_RANGE );
    printk( "%s released base io address 0x%x\n", PCIAD16DA4_NAME, 
        pciad16da4.base_io );
    
    printk( "%s DAQ board released\n", PCIAD16DA4_NAME );
    }
