/************************************************************************
 *                                                                      *
 *  pcl818.c  ---  real time daq kernel functions.                      *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/io.h>
#include "pcl818.h"

/***********************************************************************
 * some general things                                                 *
 ***********************************************************************/

/* Debug macros for DAQ board PCL818
 */
//#define DEBUG
#ifdef DEBUG
    #define DPRINTK(args...)    printk(## args)
#else
    #define DPRINTK(args...)
#endif

/************************************************************************
 *                                                                      * 
 *  Settings on the DAQ board the following driver definitions are      *
 *  valid for:                                                          *
 *                                                                      *
 *      JP#    Pos     Meaning                                          *
 *                                                                      *
 *      SW1    10Mhz  Timer clock                                      *
 *      SW2    3       DMA channel 3                                    *
 *      SW3    0x300   I/O base address                                 *
 *      SW4    REM     REM AD range                                     *
 *      SW5    S/E     Single ended                                     *       
 *      SW6    remote  AD range remote                                  *
 *                                                                      *
 *      JP1    C       TRIG0 connected to DI0                           *
 *      JP2    NC      GATE0 not connected                              * 
 *      JP3    INT     internal voltage reference                       *
 *      JP4    INT     internal voltage reference                       *
 *      JP5    -5V     DAC voltage reference                            *
 *      JP6    3       default                                          *
 *                                                                      * 
 ************************************************************************/

#define PCL818_NAME  "pcl818: "

/************************************************************************
 *  pcl818 initialisation                                               * 
 ************************************************************************/

/* for ISA pcl818s, base addr and irq
 */
#define PCL818_BASE  0x300   /* io base */
#define PCL818_RANGE 0x10    /* address range */
#define PCL818_IRQ   7


/************************************************************************
 *  control stuff                                                       * 
 ************************************************************************/

#define PCL818_CTRL_ADDR            (9)

/* values for IRQ
 */
#define PCL818_IRQ_ZEROMASK         (0x8f)
#define PCL818_IRQ_SHIFT            (4)
#define PCL818_IRQ_ENABLEMASK       (0x7f)

#define PCL818_INT_CLEAR_BYTE       (0x00)
#define PCL818_INT_CLEAR_ADDR       (8)

/* values for DMA data transfer
 */
#define PCL818_DMA_CHANNEL          (3)
#define PCL818_DMA_ZEROMASK         (0xfb)


/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

/* AI register range and offsets
 */
#define PCL818_AI_BASE              (0)
#define PCL818_AI_CHANNELS          (16) /* in single ended        */
#define PCL818_AI_LADDR             (0)  /* AI read address        */
#define PCL818_AI_HADDR             (1)  /* AI read address        */
#define PCL818_AI_SHIFT             (4)
#define PCL818_AI_START_BYTE        (0x00)
#define PCL818_AI_EOC_ADDR          (8)
#define PCL818_AI_EOC_MASK          (0x80)
#define PCL818_AI_EOC_BYTE          (0x00)
#define PCL818_AI_MODEMASK          (0xfc)
#define PCL818_AI_RANGE_ADDR        (1)
#define PCL818_AI_MUX_ADDR          (2)

/* AO register range and offsets
 */
#define PCL818_AO_BASE        0x04    /* DAC wirte address      */
#define PCL818_AO_ZERO        0x7fff  /* zero Volt at output    */
#define PCL818_AO_CHANNELS    4       /* number of channels     */

/* DI register range and offsets
 */
#define PCL818_DI_PORTS       2       /* number of input ports  */
#define PCL818_DI_DADDR       { 3, 11 } /* addresses   */
#define PCL818_DI_NBITS       8       /* number of channels     */

/* DO register range and offsets
 */
#define PCL818_DO_PORTS       2       /* number of output ports */
#define PCL818_DO_DADDR       { 3, 11 } /* addresses   */
#define PCL818_DO_NBITS       8       /* number of channels     */


/************************************************************************
 *  timer/counter stuff                                                 * 
 ************************************************************************/

/* counter chips (8254 timer/counter chip) 
 */
#define PCL818_COUNTER_CHIPS  1       /* number of counter chips     */
#define PCL818_COUNTER_CPERC  3       /* number of counter per chip  */
#define PCL818_COUNTER_TOTAL  ( PCL818_COUNTER_CPERC*PCL818_COUNTER_CHIPS )
#define PCL818_COUNTER_BASES  { 0x00e0 } /* chip address             */
#define PCL818_COUNTER_DADDR  { 0x0000, 0x0004, 0x0008 }
#define PCL818_COUNTER_CADDR  0x000c  /* COUNTER control             */

/* counter control register words (8254 timer/counter) 
 */
#define PCL818_COUNTER_BIN    0x00  /* binary count mode */
#define PCL818_COUNTER_BCD    0x01  /* BCD count mode    */
#define PCL818_COUNTER_LATCH  0x00  /* counter latch instruction    */
#define PCL818_COUNTER_LO     0x10  /* read/write low counter byte  */
#define PCL818_COUNTER_HI     0x20  /* read/write high counter byte */
#define PCL818_COUNTER_LOHI   0x30  /* read/write low, then high counter */
#define PCL818_COUNTER_SEL    { 0x00, 0x40, 0x80 } /* select counter i */
#define PCL818_COUNTER_RBACK  0xc0  /* counter read back */

/* in general, there are some 8254 timer/counter with several counters 
 * each. Some counters are used as such, others are used as timers, single
 * or by combination of two counters. For AI, a timer is usually used
 * to serve as pacer. For an application one has:
 */
#define PCL818_TIMER_16       1       /* available 16bit timers        */
#define PCL818_TIMER_16C      {0}     /* counter used for 16 bit timer */
#define PCL818_TIMER_16BASE   10000   /* Clock time in nanosecs        */
#define PCL818_TIMER_32       1       /* available 32bit timers        */
#define PCL818_TIMER_32C      {1,2}   /* counter used for 32 bit timer */
#define PCL818_TIMER_32BASE   100     /* Clock time in nanosecs        */

#define PCL818_PACER_ADDR    (10)


/************************************************************************
 * global values                                                        * 
 ************************************************************************/

struct
    {
    unsigned char di_port[PCL818_DI_PORTS];
    unsigned char do_port[PCL818_DO_PORTS];
    unsigned short int ai_active;
    unsigned short int ai_range[PCL818_AI_CHANNELS];    
    unsigned short int ao_active;
    unsigned short int base_io;
    unsigned short int irq;
    unsigned short int dma;
    unsigned char ctrl_status;
    } 
    pcl818;


/***********************************************************************
 * Control functions ...                                               *
 ***********************************************************************/

/* set interrupt status, enable or disable
 */
void pcl818_irq_status( unsigned char state  )
    {
    pcl818.ctrl_status &= PCL818_IRQ_ENABLEMASK;
    pcl818.ctrl_status |= state;

    outb( pcl818.ctrl_status, pcl818.base_io+PCL818_CTRL_ADDR );
    }

/* set number of interrupt line, for ISA
 */
void pcl818_irq_set( void )
    {
    pcl818.ctrl_status &= PCL818_IRQ_ZEROMASK;
    pcl818.ctrl_status |= (pcl818.irq << PCL818_IRQ_SHIFT);

    outb( pcl818.ctrl_status, pcl818.base_io+PCL818_CTRL_ADDR );
    }

/* get number of interrupt line, irq handler can be defined outside
 */
int pcl818_irq_get( void )
    {
    return pcl818.irq;
    }

/*irq handler for pcl818 
 */
unsigned int pcl818_irq_handler( unsigned int irq_number, 
    struct pt_regs *p )
    {
    /* nothing done her, use this template in your real time application !
     */

    return 0;
    }

/* pcl818_int_clear  ---  clear interrupt on DAQ board. 
 */ 
void pcl818_int_clear( void )
    {
    outb( PCL818_INT_CLEAR_BYTE, pcl818.base_io+PCL818_INT_CLEAR_ADDR );
    }

/* dma enable/disable
 */ 
void pcl818_dma_status( unsigned char state )
    {
    pcl818.ctrl_status &= PCL818_DMA_ZEROMASK;
    pcl818.ctrl_status |= state;
    
    outb( pcl818.ctrl_status, pcl818.base_io+PCL818_CTRL_ADDR );
    }

/* start counter "counter" with divisor and mode
 */
void pcl818_counter_set( unsigned int counter, 
    unsigned int divisor, unsigned char mode )
    {
    int nchan, nchip;
    unsigned short int counter_caddr[PCL818_COUNTER_CHIPS] = PCL818_COUNTER_BASES;
    unsigned short int counter_daddr[PCL818_COUNTER_CPERC] = PCL818_COUNTER_DADDR;
    unsigned char counter_sel[PCL818_COUNTER_CPERC] = PCL818_COUNTER_SEL;
    
    counter %= PCL818_COUNTER_TOTAL;
    nchip   = counter / PCL818_COUNTER_CPERC;
    nchan   = counter % PCL818_COUNTER_CPERC;
    
    DPRINTK("counter[%d] with %d \n",counter,divisor);

    outb( (mode|counter_sel[nchan]|PCL818_COUNTER_LOHI|PCL818_COUNTER_BIN), 
        pcl818.base_io+counter_caddr[nchip]+PCL818_COUNTER_CADDR );
    outb( (unsigned char)(divisor),	   
        pcl818.base_io+counter_caddr[nchip]+counter_daddr[nchan] );
    outb( (unsigned char)(divisor>>8),
        pcl818.base_io+counter_caddr[nchip]+counter_daddr[nchan] );
    }

/* start timer using coutners, time in nanosecs!
 */
unsigned int pcl818_timer16_start( unsigned int timer,
    unsigned int time, unsigned char mode )
    {
    unsigned int divisor;
    int timer_16c[PCL818_TIMER_16] = PCL818_TIMER_16C;
    
    divisor = time/PCL818_TIMER_32BASE;
    
    timer %= PCL818_TIMER_16;
    pcl818_counter_set( timer_16c[timer], time/PCL818_TIMER_16BASE, mode );
    
    return divisor*PCL818_TIMER_32BASE;
    }

/* stop timer using coutners
 */
void pcl818_timer16_stop( unsigned int timer )
    {
    int timer_16c[PCL818_TIMER_16] = PCL818_TIMER_16C;
    
    timer %= PCL818_TIMER_16;
    pcl818_counter_set( timer_16c[timer], 0x0000, 0x00 );
    }

/* start timer using two coutners, with "time" in nanosecs and mode!
 */
unsigned int pcl818_timer32_start( unsigned int timer,
    unsigned int time, unsigned char mode )
    {
    unsigned int divisor,d1,d2;
    int timer_32c[PCL818_TIMER_32*2] = PCL818_TIMER_32C;
    
    timer  %= PCL818_TIMER_32;
    divisor = time/PCL818_TIMER_32BASE;
    DPRINTK("divisor[%d] = %d\n",timer,divisor);
    
    d1 = divisor/0x0ffff+1;
    if ( d1 < 0x00002 ) d1 = 2;
    if ( d1 > 0x0ffff ) d1 = 0x0ffff;
    d2 = divisor/d1;
    if ( d2 < 0x00002 ) d2 = 2;
    if ( d2 > 0x0ffff ) d2 = 0x0ffff;

    time    = d1*d2*PCL818_TIMER_32BASE;   
    DPRINTK("timer[%d] at %d ns\n",timer,time);
    
    pcl818_counter_set( timer_32c[timer*2], 
        (unsigned short int)(d1), mode );
    pcl818_counter_set( timer_32c[timer*2+1], 
        (unsigned short int)(d2), mode );

    return time;
    }

/* stop timer using coutners
 */
void pcl818_timer32_stop( unsigned int timer )
    {
    int timer_32c[PCL818_TIMER_32*2] = PCL818_TIMER_32C;
    
    timer %= PCL818_TIMER_32;
    pcl818_counter_set( timer_32c[timer*2],   0x0000, 0x00 );
    pcl818_counter_set( timer_32c[timer*2+1], 0x0000, 0x00 );
    }

/* program pacer to use timer
 */
void pcl818_pacer_set( unsigned char mode )
    {
    outb( mode, pcl818.base_io+PCL818_PACER_ADDR );
    }


/***********************************************************************
 * Analog input basic functions ...                                    *
 ***********************************************************************/

/* select adc mode
 */
void pcl818_ai_set_mode( unsigned char mode )
    {
    pcl818.ctrl_status &= PCL818_AI_MODEMASK;
    pcl818.ctrl_status |= mode;

    /* program baord for mode
     */
    outb( pcl818.ctrl_status, pcl818.base_io+PCL818_CTRL_ADDR );
    }

/* select channel for successive conversion
 */
void pcl818_ai_adc_select( unsigned short int channel )
    {
    channel %= PCL818_AI_CHANNELS;
    pcl818.ai_active = channel;
        
    /* Programming the mux on DAQ board PCL818 for start
     * and stop channel (int) from 0-15.
     */ 
    outb( (unsigned char) ( channel | ( channel << 4 ) ), 
        pcl818.base_io+PCL818_AI_MUX_ADDR );
    }

/* program ADC for successive conversion, with both channel 
 * and range as defined in header file.
 */
void pcl818_ai_adc_program( unsigned short int channel,
    unsigned short int range )
    {
    channel %= PCL818_AI_CHANNELS;
    pcl818.ai_active = channel;
    pcl818.ai_range[channel] = range;
    
    pcl818_ai_adc_select( channel );
    
    /* Setting the range.
     */ 
    outb( range, pcl818.base_io+PCL818_AI_RANGE_ADDR );
    }

/* get a single analog value mapped to 16bit independently of ADC,
 * assuming that ADC is terminated ( or/and value is in FIFO).
 */
inline unsigned short int pcl818_ai_get_value( void )
    {
    unsigned char lbyte, hbyte;

    /* Now we get high and low byte
     */
    lbyte = inb( pcl818.base_io + PCL818_AI_LADDR );    
    hbyte = inb( pcl818.base_io + PCL818_AI_HADDR );

    /* value is a pointer to (unsigned short int), and board offers
     * D11 D10 D9 D8 D7 D6 D5 D4 | D3 D2 D1 D0 x x x x
     * so we take lbyte masking out the channel and shift hbyte 8 up
     */
    return ( (unsigned short int) ( hbyte << 8   ) + 
             (unsigned short int) ( lbyte & 0xf0 ) );  
    }

/* get a single analog value given that channel is already
 * selected and ints range set, mode is software.
 */
int pcl818_ai_get_value0( unsigned short int *value )
    {
    unsigned char conversion_status;

    /* Start conversion by writing anything to PCL818_AI_BASE
     */
    outb( PCL818_AI_START_BYTE, pcl818.base_io + PCL818_AI_BASE );

    /* We have to keep the CPU busy waiting here, which is 
     * quite stupid. Has to be replaced by something more
     * sophisticated.
     */
    while(1)
        {
        conversion_status  = inb( pcl818.base_io + PCL818_AI_EOC_ADDR );
        conversion_status &= PCL818_AI_EOC_MASK;
        if ( conversion_status == PCL818_AI_EOC_BYTE )
            return -1;
        }

    *value = pcl818_ai_get_value();
    return 0;
    }

/* get a single analog value( channel ) given that its range set.
 */
int pcl818_ai_get_value1( unsigned short int *value,
    unsigned short int channel )
    {
    pcl818_ai_adc_select( channel );
    
    return pcl818_ai_get_value0( value );
    }

/* get a single analog value( channel, range ).
 */
int pcl818_ai_get_value2( unsigned short int *value,
    unsigned short int channel, unsigned short int range )
    {
    pcl818_ai_adc_program( channel, range );

    return pcl818_ai_get_value0( value );
    }

/* start a channel scan with list, number of channels and sampling time[us]
 * mode has to be set appropriately, extern (pre,post,middle) or internal
 */
int pcl818_ai_start_scan( unsigned short int *channellist,
    unsigned short int nchannels, unsigned int time, unsigned char mode )
    {
    /* Program channel list on pcl818
     */
     
    /* Stop timer and clear FIFO
     */
     
    /* Program PIC for scan
     */
     
    /* Start timer/pacer according to mode
     */
    
    return 0;
    }

/* Once the scan has been started, the application is responsible
 * for picking up values from the fifos periodically. If an interupt
 * is available, this can be done by the respective routine. The
 * rest depends on the application.
 */
int pcl818_ai_get_scan( unsigned short int *value )
    {
    /* check end of FIFO if yes return 0, if not empty, then */
    
    *value = pcl818_ai_get_value();
    
    return 1;
    }

/* Stop analogue input scan, wbasically stop the timer.
 */
void pcl818_ai_stop_scan( void )
    {
    /* stop all timers 
     */
    }


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* select channel for successive DA conversion
 */
void pcl818_ao_dac_select( unsigned short int channel )
    {
    pcl818.ao_active = channel;
    }

/* set a single analog value given that channel is already
 * selected and ints range set value mapped to 16bit.
 */
void pcl818_ao_set_value0( unsigned short int value )
    {
    unsigned short int ao_addr;
    
    /* addr using pciad16da4.ao_active
     */
    ao_addr = pcl818.base_io+PCL818_AO_BASE+(pcl818.ao_active<<1);
     
    /* outb values by hbyte and lbyte to latch
     */
    outb( (unsigned char)(value),    ao_addr   );
    outb( (unsigned char)(value>>8), ao_addr+1 );
    }

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
void pcl818_ao_set_value1( unsigned short int value, 
    unsigned short int channel )
    {
    pcl818_ao_dac_select( channel );
    pcl818_ao_set_value0( value );
    }


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
unsigned char pcl818_di_get_byte( unsigned short int nport )
    {
    unsigned short int di_addr[PCL818_DI_PORTS] = PCL818_DI_DADDR;

    nport %= PCL818_DI_PORTS; /* just in case */

    pcl818.di_port[nport] = inb( pcl818.base_io + di_addr[nport] );
            
    return pcl818.di_port[nport];    
    }

/* get bit #"nbit" from port "nport".
 */ 
unsigned short int pcl818_di_get_bit( unsigned short int nport, 
    unsigned short int nbit )
    {
    nbit  %= PCL818_DI_NBITS; /* just in case */

    return ( (unsigned short int) 
        ( ( ( pcl818_di_get_byte( nport ) & ( 1 << nbit ) ) >> nbit ) ) );
    } 

/* set byte "byte" on port "nport"
 */ 
void pcl818_do_set_byte( unsigned char byte,
    unsigned short int nport )
    {
    unsigned short int do_addr[PCL818_DO_PORTS] = PCL818_DO_DADDR;

    nport %= PCL818_DO_PORTS; /* just in case */

    pcl818.do_port[nport] = byte; /* keep a copy */

    outb( pcl818.do_port[nport], pcl818.base_io + do_addr[nport] );
    }

/* set bit "bit" with #"nbit" on port "nport".
 */ 
void pcl818_do_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit )
    {
    nport %= PCL818_DO_PORTS; /* just in case */
    nbit  %= PCL818_DO_NBITS; /* just in case */
    bit   &= 0x01;           /* just in case */

    pcl818.do_port[nport] &= ~(1 << nbit);
    pcl818.do_port[nport] |=  (bit << nbit); 
    
    pcl818_do_set_byte( nport, pcl818.do_port[nport] );
    }


/***********************************************************************
 *  board defaults                                                     *
 ***********************************************************************/

static void pcl818_defaults( void )
    {
    int i;
        
    /* init ranges for analog inputs, if possible
     */
    for ( i=0; i<PCL818_AI_CHANNELS; i++ )
        pcl818_ai_adc_program( i, PCL818_AI_PM5 );   

    /* init values for analog output 
     */
    for ( i=0; i<PCL818_AO_CHANNELS; i++ )
        pcl818_ao_set_value1( PCL818_AO_ZERO, i );   

    /* init values for digital output 
     */
    for ( i=0; i<PCL818_DO_PORTS; i++ )
        pcl818_do_set_byte( 0x00, i );   
    }


/***********************************************************************
 * Linux kernel functions                                              *
 ***********************************************************************/

/*the mandatory init_module.
 */
int init_module( void )
    {
    /* Initialize addresses and other stuff 
     */
    pcl818.base_io     = PCL818_BASE;
    pcl818.dma         = PCL818_DMA_CHANNEL;
    pcl818.ctrl_status = 0x00;

    /* request base io addresses
     */
    if ( check_region( pcl818.base_io, PCL818_RANGE ) < 0 ) 
        {
     	printk( "%s could not request base io addresss\n", PCL818_NAME );
        return -EIO;
        }
    request_region( pcl818.base_io, PCL818_RANGE, PCL818_NAME );
    printk( "%s requested base io address 0x%x\n", PCL818_NAME, 
        pcl818.base_io );

    pcl818.irq = PCL818_IRQ;
    printk("%s IRQ is selected to be %d (0 = no interrupt)\n", 
        PCL818_NAME, pcl818.irq );

    /* Set pcl818 defaults
     */
    pcl818_defaults();

    return 0;
    } 

/*the mandatory cleanup_module.
 */
void cleanup_module( void )
    {
    /* Set pcl818 defaults
     */
    pcl818_defaults();
    
    release_region( pcl818.base_io, PCL818_RANGE );
    printk( "%s released base io address 0x%x\n", PCL818_NAME, 
        pcl818.base_io );
    
    printk( "%s DAQ board released\n", PCL818_NAME );
    }
