/************************************************************************
 *                                                                      *
 *  pcl818.h  ---  daq inputs header file for pcl818.c.                 *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *  control stuff                                                       * 
 ************************************************************************/

/* dma
 */
#define PCL818_DMA_ENABLE     0x04    /* enable DMA transfer    */
#define PCL818_DMA_DISABLE    0x00    /* disable DMA transfer   */

/* interrupt dis/enable
 */
#define PCL818_IRQ_DISABLE    0x00  /* disable all interrupts */
#define PCL818_IRQ_ENABLE     0x80  /* enable interrupt       */

/* counter modes
 */
#define PCL818_COUNTER_MODE0  0x00  /* interrupt on terminal count */
#define PCL818_COUNTER_MODE1  0x02  /* programmable oneshot */
#define PCL818_COUNTER_MODE2  0x04  /* rate generator */
#define PCL818_COUNTER_MODE3  0x06  /* square wave generator */
#define PCL818_COUNTER_MODE4  0x08  /* software triggered pulse */
#define PCL818_COUNTER_MODE5  0x0a  /* hardware triggered pulse */

/* pacer modes
 */
#define PCL818_PACER_MODE0  0x00
#define PCL818_PACER_MODE1  0x01
#define PCL818_PACER_MODE2  0x02
#define PCL818_PACER_MODE3  0x03


/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

/* AI range
 */
#define PCL818_AI_PM5       (0x0000)    /* [  -5.0,    5.0   ] Volt */
#define PCL818_AI_PM05      (0x0040)    /* [  -0.5,    0.5   ] Volt */
#define PCL818_AI_PM005     (0x0080)    /* [  -0.05,   0.05  ] Volt */
#define PCL818_AI_PM0005    (0x00c0)    /* [  -0.005,  0.005 ] Volt */
#define PCL818_AI_PM10      (0x0100)    /* [ -10.0,   10.0   ] Volt */
#define PCL818_AI_PM1       (0x0140)    /* [  -1.0,    1.0   ] Volt */
#define PCL818_AI_PM01      (0x0180)    /* [  -0.1,    0.1   ] Volt */
#define PCL818_AI_PM001     (0x01c0)    /* [  -0.01,   0.01  ] Volt */
#define PCL818_AI_0P10      (0x0200)    /* [   0.0,   10.0   ] Volt */
#define PCL818_AI_0P1       (0x0240)    /* [   0.0,    1.0   ] Volt */
#define PCL818_AI_0P01      (0x0280)    /* [   0.0,    0.1   ] Volt */
#define PCL818_AI_0P001     (0x02c0)    /* [   0.0,    0.01  ] Volt */

#define PCL818_AI_FIFO      (1024)        /* input FIFO size */

#define PCL818_AI_SOFTWARE  (0x00)    
#define PCL818_AI_EXTERNAL  (0x02)
#define PCL818_AI_PACER     (0x03)

/* ADC mode
 */
#define PCL818_SOFTWARE_MODE  0x00    
#define PCL818_EXTERNAL_MODE  0x02
#define PCL818_PACER_MODE     0x03


/***********************************************************************
 * Control functions ...                                               *
 ***********************************************************************/

/* set interrupt status, enable or disable
 */
extern void pcl818_irq_status( unsigned char state );

/* set number of interrupt line, for ISA
 */
extern void pcl818_irq_set( void );

/* get number of interrupt line, irq handler can be defined outside
 */
extern int pcl818_irq_get( void );

/* pcl818_int_clear  ---  clear interrupt on DAQ board. 
 */ 
extern void pcl818_int_clear( void );

/* dma enable/disable
 */ 
extern void pcl818_dma_status( unsigned char state );

/* start counter "counter" with divisor and mode
 */
extern void pcl818_counter_set( unsigned int counter, 
    unsigned int divisor, unsigned char mode );

/* start timer using coutners, time in msecs!
 */
extern unsigned int pcl818_timer16_start( unsigned int timer,
    unsigned int time, unsigned char mode );

/* stop timer using coutners
 */
extern void pcl818_timer16_stop( unsigned int timer );

/* start timer using two coutners, time in usecs!
 */
extern unsigned int pcl818_timer32_start( unsigned int timer,
    unsigned int time, unsigned char mode );

/* stop timer using coutners
 */
extern void pcl818_timer32_stop( unsigned int timer );

/* program pacer to use timer
 */
extern void pcl818_pacer_set( unsigned char mode );


/***********************************************************************
 * Analog input basic functions ...                                    *
 ***********************************************************************/

/* select adc mode
 */
extern void pcl818_ai_set_mode( unsigned char mode );

/* select channel for successive conversion
 */
extern void pcl818_ai_adc_select( unsigned short int channel );

/* program ADC for successive conversion, with both channel 
 * and range as defined in header file.
 */
extern void pcl818_ai_adc_program( unsigned short int channel,
    unsigned short int range );

/* get a single analog value given that channel is already
 * selected and ints range set, mode is software.
 */
extern int pcl818_ai_get_value0( unsigned short int *value );

/* get a single analog value( channel ); given that its range set.
 */
extern int pcl818_ai_get_value1( unsigned short int *value,
    unsigned short int channel );

/* get a single analog value( channel, range );.
 */
extern int pcl818_ai_get_value2( unsigned short int *value,
    unsigned short int channel, unsigned short int range );

/* start a channel scan with list, number of channels and sampling time[us]
 * mode has to be set appropriately, extern (pre,post,middle) or internal
 */
extern int pcl818_ai_start_scan( unsigned short int *channellist,
    unsigned short int nchannels, unsigned int time, unsigned char mode );

/* Once the scan has been started, the application is responsible
 * for picking up values from the fifos periodically. If an interupt
 * is available, this can be done by the respective routine. The
 * rest depends on the application.
 */
extern int pcl818_ai_get_scan( unsigned short int *value );

/* Stop analogue input scan, wbasically stop the timer.
 */
extern void pcl818_ai_stop_scan( void );


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* select channel for successive DA conversion
 */
extern void pcl818_ao_dac_select( unsigned short int channel );

/* set a single analog value given that channel is already
 * selected and ints range set value mapped to 16bit.
 */
extern void pcl818_ao_set_value0( unsigned short int value );

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
extern void pcl818_ao_set_value1( unsigned short int value, 
    unsigned short int channel );


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
extern unsigned char pcl818_di_get_byte( unsigned short int nport );

/* get bit #"nbit" from port "nport".
 */ 
extern unsigned short int pcl818_di_get_bit( unsigned short int nport, 
    unsigned short int nbit );

/* set byte "byte" on port "nport"
 */ 
extern void pcl818_do_set_byte( unsigned char byte,
    unsigned short int nport );

/* set bit "bit" with #"nbit" on port "nport".
 */ 
extern void pcl818_do_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit );
