/************************************************************************
 *                                                                      *
 *  piodax.c  ---  real time daq kernel functions.                      *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/io.h>
#include "piodax.h"

/***********************************************************************
 * some general things                                                 *
 ***********************************************************************/

/* Debug macros for DAQ piodax PIODAX
 */
//#define DEBUG
#ifdef DEBUG
    #define DPRINTK(args...)    printk(## args)
#else
    #define DPRINTK(args...)
#endif

/************************************************************************
 *                                                                      * 
 *  Settings on the DAQ piodax the following driver definitions are      *
 *  valid for:                                                          *
 *                                                                      *
 *    JP#    Pos       Meaning                                          *
 *                                                                      * 
 ************************************************************************/

#define PIODAX_NAME  "piodax: "

/************************************************************************
 *  piodax initialisation                                                * 
 ************************************************************************/

/* define board type
 */
#define PIODA4

/* PCI stuff
 */
#define PIODAX_VENDOR_ID    0x0000e159  /* as used in pci_find_device */
#define PIODAX_DEVICE_ID    0x00000002  /* as used in pci_find_device */
#define PIODAX_SUBAUX_ID    0x00800400  /* as used in driver */
#define PIODAX_SUBAUX_MASK  0xf0        /* look into code */
#define PIODAX_SUBAUX_ADDR  0x0007      /* look into code */

/* register range and offsets
 */
#define PIODAX_RANGE              0x0100    /* base address range */
#define PIODAX_RESET_ADDR         0x0000    /* reset register     */
#define PIODAX_CONTROL_ADDR       0x0000    /* control register */
#define PIODAX_AUX_CTRL_ADDR      0x0002    /* auxiliary control register */
#define PIODAX_AUX_DATA_ADDR      0x0003    /* auxiliary data register */
#define PIODAX_AUX_STAT_ADDR      0x0007    /* auxiliary pin status register */
#define PIODAX_INT_MASK_ADDR      0x0005    /* interrupt mask register */
#define PIODAX_INT_POLARITY_ADDR  0x002a    /* interrput polarity register*/

#define PIODAX_INT_MASK_SHIFT     0x04


/************************************************************************
 *  control stuff                                                       * 
 ************************************************************************/

/* board enable
 */
#define PIODAX_ENABLE   0x11    /* after boot, enable board */
#define PIODAX_DISABLE  0x00    /* disable at shutdown      */

/************************************************************************
 *  io stuff                                                            * 
 ************************************************************************/

/* AO register range and offsets
 */
#define PIODAX_AO_SEL_ADDR   0x00e0   /* DA chip select      */
#define PIODAX_AO_LO_ADDR    0x00f0   /* write DA low byte   */
#define PIODAX_AO_HI_ADDR    0x00f4   /* write DA high byte  */

#ifdef PIODA4
    #define PIODAX_AO_CHANNELS    4
#endif
#ifdef PIODA8
    #define PIODAX_AO_CHANNELS    8
#endif
#ifdef PIODA16
    #define PIODAX_AO_CHANNELS    16
#endif

/* DI register range and offsets
 */
#define PIODAX_DI_PORTS      2       /* number of input ports */
#define PIODAX_DI_DADDR      { 0x00f8, 0x00fc } /* addresses  */
#define PIODAX_DI_NBITS      8       /* number of channels    */

/* DO register range and offsets
 */
#define PIODAX_DO_PORTS      2       /* number of output ports */
#define PIODAX_DI_DADDR      { 0x00f8, 0x00fc } /* addresses   */
#define PIODAX_DO_NBITS      8       /* number of channels     */


/************************************************************************
 *  timer/counter stuff                                                 * 
 ************************************************************************/

/* counter chips (8254 timer/counter chip) 
 */
#define PIODAX_COUNTER_CHIPS 1       /* number of counter chips     */
#define PIODAX_COUNTER_CPERC 3       /* number of counter per chip  */
#define PIODAX_COUNTER_TOTAL ( PIODAX_COUNTER_CPERC*PIODAX_COUNTER_CHIPS )
#define PIODAX_COUNTER_BASES { 0x00c0 } /* chip address             */
#define PIODAX_COUNTER_DADDR { 0x0000, 0x0004, 0x0008 }
#define PIODAX_COUNTER_CADDR 0x000c  /* COUNTER control             */

/* counter control register words (8254 timer/counter) 
 */
#define PIODAX_COUNTER_BIN   0x00    /* binary count mode             */
#define PIODAX_COUNTER_BCD   0x01    /* BCD count mode                */
#define PIODAX_COUNTER_LATCH 0x00    /* counter latch instruction     */
#define PIODAX_COUNTER_LO    0x10    /* read/write low counter byte   */
#define PIODAX_COUNTER_HI    0x20    /* read/write high counter byte  */
#define PIODAX_COUNTER_LOHI  0x30    /* read/write low, then high     */
#define PIODAX_COUNTER_SEL   { 0x00, 0x40, 0x80 } /* select counter i */
#define PIODAX_COUNTER_RBACK 0xc0    /* counter read back             */

/* in general, there are some 8254 timer/counter with several counters 
 * each. Some counters are used as such, others are used as timers, single
 * or by combination of two counters. For AI, a timer is usually used
 * to serve as pacer. For an application one has:
 */
#define PIODAX_TIMER_BASE    250     /* Clock time in nanosecs        */
#define PIODAX_TIMER_16      1       /* available 16bit timers        */
#define PIODAX_TIMER_16C     {0}     /* counter used for 16 bit timer */
#define PIODAX_TIMER_32      1       /* available 32bit timers        */
#define PIODAX_TIMER_32C     {1,2}   /* counter used for 32 bit timer */


/************************************************************************
 * global values                                                        * 
 ************************************************************************/

struct
    {
    unsigned char di_port[PIODAX_DI_PORTS];
    unsigned char do_port[PIODAX_DO_PORTS];
    unsigned short int ao_active;
    unsigned short int ao_range[PIODAX_AO_CHANNELS];
    unsigned long  int base_phys;
    unsigned short int base_io;
    unsigned short int irq;
    unsigned int  int_mask;
    unsigned char int_source;
    unsigned int  int_polarity;
    } 
    piodax;


/***********************************************************************
 * Control functions ...                                               *
 ***********************************************************************/

/* program interrupt with mask
 */
void piodax_int_select( unsigned int mask )
    {
    piodax.int_mask = mask;
    piodax.int_polarity = PIODAX_INTX_NONINV;

    outb( piodax.int_polarity, 
        piodax.base_io+PIODAX_INT_POLARITY_ADDR );
    outb( piodax.int_mask, 
        piodax.base_io+PIODAX_INT_MASK_ADDR );
    }

/* invert polarity
 */
unsigned int piodax_inv_polarity( void )
    {
    unsigned int mask;      /* mask for test */
    unsigned int status;    /* AUX Pin Status Register */
    unsigned int source;    /* interrupt source  */
    unsigned int edge;      /* interrupt source edge */

    edge   = 0;
    source = 0;
    status = inb( piodax.base_io+PIODAX_AUX_STAT_ADDR ) & 0x0f;

    for ( mask = 1; mask < PIODAX_INT_MASK_SHIFT; mask <<= 1 )
        {
        if ( piodax.int_polarity & mask ) /* non-inverse mode */
            {
            if ( status & mask )
                {
                source |= mask;    /* interrupt source */
                edge   |= mask;    /* positive edge    */
                }
            }
        else /* inverse mode */
            {
            if (!(status & mask))
                {
                source |= mask;    /* interrupt source */
                edge   &= !mask;   /* negative edge    */
                }
            }
        }

    piodax.int_polarity ^= source; /* Inverts the polarity of interrupted */
    outb( piodax.int_polarity, 
        piodax.base_io + PIODAX_INT_POLARITY_ADDR ); /* source for next edge. */

    return edge;
    }

/* get number of interrupt line, irq handler can be defined outside
 */
int piodax_irq_get( void )
    {
    return piodax.irq;
    }

/*irq handler for piodax 
 */
unsigned int piodax_irq_handler( unsigned int irq_number, 
    struct pt_regs *p )
    {
    /* has to be done since PCI interrupt is level triggered 
     */
    piodax_inv_polarity();

     /* nothing done her, use this template in your real time application !
     */
    
    return 0;
    }

/* start counter "counter" with divisor and mode
 */
void piodax_counter_set( unsigned int counter, 
    unsigned int divisor, unsigned char mode )
    {
    int nchan, nchip;
    unsigned short int counter_caddr[PIODAX_COUNTER_CHIPS] = PIODAX_COUNTER_BASES;
    unsigned short int counter_daddr[PIODAX_COUNTER_CPERC] = PIODAX_COUNTER_DADDR;
    unsigned char counter_sel[PIODAX_COUNTER_CPERC] = PIODAX_COUNTER_SEL;
    
    counter %= PIODAX_COUNTER_TOTAL;
    nchip   = counter / PIODAX_COUNTER_CPERC;
    nchan   = counter % PIODAX_COUNTER_CPERC;
    
    DPRINTK("counter[%d] with %d \n",counter,divisor);

    outb( (mode|counter_sel[nchan]|PIODAX_COUNTER_LOHI|PIODAX_COUNTER_BIN), 
        piodax.base_io+counter_caddr[nchip]+PIODAX_COUNTER_CADDR );
    outb( (unsigned char)(divisor),	   
        piodax.base_io+counter_caddr[nchip]+counter_daddr[nchan] );
    outb( (unsigned char)(divisor>>8),
        piodax.base_io+counter_caddr[nchip]+counter_daddr[nchan] );
    }

/* start timer using coutners, with "time" in nanosecs and mode!
 */
unsigned int piodax_timer16_start( unsigned int timer,
    unsigned int time, unsigned char mode )
    {
    unsigned int divisor;
    int timer_16c[PIODAX_TIMER_16] = PIODAX_TIMER_16C;
    
    divisor = time/PIODAX_TIMER_BASE;
    
    timer %= PIODAX_TIMER_16;
    piodax_counter_set( timer_16c[timer], time/PIODAX_TIMER_BASE, mode );
    
    return divisor*PIODAX_TIMER_BASE;
    }

/* stop timer using coutners
 */
void piodax_timer16_stop( unsigned int timer )
    {
    int timer_16c[PIODAX_TIMER_16] = PIODAX_TIMER_16C;
    
    timer %= PIODAX_TIMER_16;
    piodax_counter_set( timer_16c[timer], 0x0000, 0x00 );
    }

/* start timer using two coutners, with "time" in nanosecs and mode!
 */
unsigned int piodax_timer32_start( unsigned int timer,
    unsigned int time, unsigned char mode )
    {
    unsigned int divisor,d1,d2;
    int timer_32c[PIODAX_TIMER_32*2] = PIODAX_TIMER_32C;
    
    timer  %= PIODAX_TIMER_32;
    divisor = time/PIODAX_TIMER_BASE;
    DPRINTK("divisor[%d] = %d\n",timer,divisor);
    
    d1 = divisor/0x0ffff+1;
    if ( d1 < 0x00002 ) d1 = 2;
    if ( d1 > 0x0ffff ) d1 = 0x0ffff;
    d2 = divisor/d1;
    if ( d2 < 0x00002 ) d2 = 2;
    if ( d2 > 0x0ffff ) d2 = 0x0ffff;

    time    = d1*d2*PIODAX_TIMER_BASE;   
    DPRINTK("timer[%d] at %d ns\n",timer,time);
    
    piodax_counter_set( timer_32c[timer*2], 
        (unsigned short int)(d1), mode );
    piodax_counter_set( timer_32c[timer*2+1], 
        (unsigned short int)(d2), mode );

    return time;
    }

/* stop timer using coutners
 */
void piodax_timer32_stop( unsigned int timer )
    {
    int timer_32c[PIODAX_TIMER_32*2] = PIODAX_TIMER_32C;
    
    timer %= PIODAX_TIMER_32;
    piodax_counter_set( timer_32c[timer*2],   0x0000, 0x00 );
    piodax_counter_set( timer_32c[timer*2+1], 0x0000, 0x00 );
    }


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* activate channel for successive DA conversion
 */
void piodax_ao_dac_select( unsigned short int channel )
    {
    piodax.ao_active = channel;
    }

/* program DAC for successive DA conversion with both channel 
 * and range as defined in header file.
 */
void piodax_ao_dac_program( unsigned short int channel,
    unsigned short int range )
    {
    channel %= PIODAX_AO_CHANNELS;
    piodax.ao_range[channel] = range;
    piodax_ao_dac_select( channel );
    }

/* set a single analog value given that channel is already
 * selected and ints range set value mapped to 16bit.
 */
void piodax_ao_set_value0( unsigned short int value )
    {
    unsigned short int output;

    DPRINTK( "%s output channel = %d\n",   PIODAX_NAME, piodax.ao_active ); 
    DPRINTK( "%s virgin output  = 0x%x\n", PIODAX_NAME, value ); 

    /* This hardware always outputs [-10V, 10V] mapped
       to 14 bit [0x0000, 0x3fff]. Depending on the output
       voltage range selected, only less bits are used,
       e.g. the range [0x0fff, 0x2fff] is used to output 
       the output [-5V, 5V]. In order to offer always
       virtual 16 bit for any range, we map the "value"
       to the bit range according to the selected range
     */
    switch( piodax.ao_range[piodax.ao_active] )
        {
        case PIODAX_AO_PM10: 
            output = ( value >> 2 );          /* reduce to 14 bit */
            break;
        
        case PIODAX_AO_PM5:
            output = ( value >> 3 ) + 0x0fff; /* reduce to 13 bit and shift */      
            break;

        case PIODAX_AO_0P10:
            output = ( value >> 3 ) + 0x1fff; /* reduce to 13 bit and shift */                  
            break;

        case PIODAX_AO_0P5:
            output = ( value >> 4 ) + 0x1fff; /* reduce to 12 bit and shift */                  
            break;

        default:
            output = 0; /* should not arrive here */
        }
    DPRINTK( "%s shifted output = 0x%x\n", PIODAX_NAME, output ); 

    /* now, output is mapped to the correct range, add channel info 
     */
    output |= (piodax.ao_active<<14);   
    outb( (unsigned char)(output>>8), piodax.base_io + PIODAX_AO_HI_ADDR );
    outb( (unsigned char)(output),    piodax.base_io + PIODAX_AO_LO_ADDR );
    
    /* select DAC by writing 0xff to appropriate address
     */
    outb( 0xff, piodax.base_io + PIODAX_AO_SEL_ADDR + (piodax.ao_active>>2) ); 

    DPRINTK( "%s DAC addr = 0x%x\n", PIODAX_NAME, 
        piodax.base_io + PIODAX_AO_SEL_ADDR + (piodax.ao_active>>2) );

    DPRINTK( "%s output = 0x%x\n", PIODAX_NAME, output ); 
    DPRINTK( "%s lbyte 0x%x to address 0x%x\n", PIODAX_NAME, 
        (output&0x00ff), piodax.base_io + PIODAX_AO_LO_ADDR ); 
    DPRINTK( "%s hbyte 0x%x to address 0x%x\n", PIODAX_NAME, 
        ((output&0xff00)>>8), piodax.base_io + PIODAX_AO_HI_ADDR ); 
    }

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
void piodax_ao_set_value1( unsigned short int value, 
    unsigned short int channel )
    {
    piodax_ao_dac_select( channel );
    piodax_ao_set_value0( value );
    }

/* set a single analog value (channel,range).
 * value is mapped to 16bit independently of DAC.
 */
void piodax_ao_set_value2( unsigned short int value,
    unsigned short int channel, unsigned short int range )
    {
    piodax_ao_dac_program( channel, range );
    piodax_ao_set_value0( value );
    }


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
unsigned char piodax_di_get_byte( unsigned short int nport )
    {
    unsigned short int di_addr[PIODAX_DI_PORTS] = PIODAX_DI_DADDR;

    nport %= PIODAX_DI_PORTS; /* just in case */

    piodax.di_port[nport] = inb( piodax.base_io + di_addr[nport] );
            
    return piodax.di_port[nport];    
    }

/* get bit #"nbit" from port "nport".
 */ 
unsigned short int piodax_di_get_bit( unsigned short int nport, 
    unsigned short int nbit )
    {
    nbit  %= PIODAX_DI_NBITS; /* just in case */

    return ( (unsigned short int) 
        ( ( ( piodax_di_get_byte( nport ) & ( 1 << nbit ) ) >> nbit ) ) );
    } 

/* set byte "byte" on port "nport"
 */ 
void piodax_do_set_byte( unsigned char byte, 
    unsigned short int nport )
    {
    unsigned short int do_addr[PIODAX_DO_PORTS] = PIODAX_DI_DADDR;

    nport %= PIODAX_DO_PORTS; /* just in case */

    piodax.do_port[nport] = byte; /* keep a copy */

    outb( piodax.do_port[nport], piodax.base_io + do_addr[nport] );
    }

/* set bit "bit" with #"nbit" on port "nport".
 */ 
void piodax_do_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit )
    {
    nport %= PIODAX_DO_PORTS; /* just in case */
    nbit  %= PIODAX_DO_NBITS; /* just in case */
    bit   &= 0x01;           /* just in case */

    piodax.do_port[nport] &= ~(1 << nbit);
    piodax.do_port[nport] |=  (bit << nbit); 
    
    piodax_do_set_byte( piodax.do_port[nport], nport );
    }


/***********************************************************************
 *  piodax defaults                                                     *
 ***********************************************************************/

static void piodax_defaults( void )
    {
    int i;
        
    /* init ranges for analog outputs, if possible
     */
    for ( i=0; i<PIODAX_AO_CHANNELS; i++ )
        piodax_ao_dac_program( i, PIODAX_AO_PM5 );   

    /* init values for analog output 
     */
    for ( i=0; i<PIODAX_AO_CHANNELS; i++ )
        piodax_ao_set_value1( PIODAX_AO_PM5_ZERO, i );   

    /* init values for digital output 
     */
    for ( i=0; i<PIODAX_DO_PORTS; i++ )
        piodax_do_set_byte( 0x00, i );   
    }


/***********************************************************************
 * Linux kernel functions                                              *
 ***********************************************************************/

/*the mandatory init_module.
 */
int init_module( void )
    {
    struct pci_dev *dev = NULL;
    unsigned char  sub_ven, sub_dev, sub_aux;
    unsigned int   csid;  /* composed sub-ids */

    /* Check presence of PCI bus, cela se discute.
     */
    if( !pci_present() ) 
        {
        printk("%s PCI bus not present!\n", PIODAX_NAME );
        return -ENODEV;
        }
    
    /* find the first device of this kind
     */
    while(1)
        {
        dev =  pci_find_device( PIODAX_VENDOR_ID, PIODAX_DEVICE_ID, dev );
        if ( !dev )
            {
            printk("%s no PCI device with vendor ID %x and device ID %x found\n",
                PIODAX_NAME, PIODAX_VENDOR_ID, PIODAX_DEVICE_ID );
            return -ENODEV;
            }
        printk("%s PCI device with vendor ID %x and device ID %x found\n",
            PIODAX_NAME, PIODAX_VENDOR_ID, PIODAX_DEVICE_ID );

        /* get sub-vendor-id and sub-device-id, sub-aux-id,
         * a little bit annoying make composed sub-ids and identify this device
         */
        if ( pci_read_config_byte( dev, PCI_SUBSYSTEM_VENDOR_ID, &sub_ven ) ||
		     pci_read_config_byte( dev, PCI_SUBSYSTEM_ID,        &sub_dev ) )
            {
		    printk( "%s getting subsystem ID failed!\n", PIODAX_NAME );
		    return -ENXIO; /* something went wrong */
            }

        /* get physical address, request it and make io address
         */
        piodax.base_phys =  dev->base_address[0] & PCI_BASE_ADDRESS_IO_MASK;
        if ( check_region( piodax.base_phys, PIODAX_RANGE ) < 0 )
            {
         	printk( "%s could not request base address\n", PIODAX_NAME );
            continue; /* look for another board */
            }
        request_region( piodax.base_phys, PIODAX_RANGE, PIODAX_NAME );
        piodax.base_io = (unsigned short int)(piodax.base_phys);
        printk( "%s requested base io address 0x%x\n", 
            PIODAX_NAME, piodax.base_io );

        sub_aux = ( inb( piodax.base_io + PIODAX_SUBAUX_ADDR ) & 
                    PIODAX_SUBAUX_MASK );
        csid = (sub_ven<<16) + (sub_dev<<8) + (sub_aux);
        if( csid == PIODAX_SUBAUX_ID )
            {
            printk("%s PCI device with composed sub ID %x is a %s\n", 
                PIODAX_NAME, PIODAX_SUBAUX_ID, PIODAX_NAME );
            
            piodax.irq = dev->irq;
            printk("%s IRQ is selected to be %d (0 = no interrupt)\n",
                PIODAX_NAME, dev->irq );
 
            /* Board has to be enabled before, not explicitely in manual!!
             */
            outb( PIODAX_ENABLE, piodax.base_io + PIODAX_RESET_ADDR );
            printk( "%s DAQ board enabled\n", PIODAX_NAME );

            /* Set board defaults
             */
            piodax_defaults();
            
            break;
            }
        else
            {
            printk("%s PCI device with composed sub ID %x is no %s\n",
                PIODAX_NAME, PIODAX_SUBAUX_ID, PIODAX_NAME );
            release_region( piodax.base_phys, PIODAX_RANGE );
            printk( "%s released base io address 0x%x\n", 
                PIODAX_NAME, piodax.base_io );
            }
        }
        
    return 0;
    } 
     
/*the mandatory cleanup_module.
 */
void cleanup_module( void )
    {
    /* Set board defaults
     */
    piodax_defaults();

    /* disable board
     */
    outb( PIODAX_DISABLE, piodax.base_io + PIODAX_RESET_ADDR );
    printk( "%s DAQ board disabled\n", PIODAX_NAME );
       
    release_region( piodax.base_phys, PIODAX_RANGE );
    printk( "%s released base io address 0x%x\n", 
        PIODAX_NAME, piodax.base_io );
 	
    printk( "%s DAQ board released\n", PIODAX_NAME );
    } 
