/************************************************************************
 *                                                                      *
 *  piodax.h  ---  daq inputs header file for piodax.c.                 *
 *                                                                      *
 *  This file is open source software; you can redistribute it and/or   *
 *  modify it under the terms of the version 2 of GNU General Public    *
 *  License as published by the Free Software Foundation.               *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with this package; if not, write to the Free Software         *
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.           *
 *  or look at http://www.gnu.org/fsf/fsf.html                          *
 *                                                                      *
 *  Copyleft, Peter Wurmsdobler, peterw@thinkingnerds.com               *
 *                                                                      *
 ************************************************************************/

/************************************************************************ 
 *  values for public functions                                         *
 ************************************************************************/

/* interrupt control values to mask register
 */
#define PIODAX_INTx_DISABLE     0x00    /* disable all interrupt  */
#define PIODAX_INT0_ENABLE      0x01    /* enable interrupt INT0  */
#define PIODAX_INT1_ENABLE      0x02    /* enable interrupt INT1  */
#define PIODAX_INTx_ENABLE      0x03    /* enable interrupt INT0 + INT1 */

/* interrupt polarity values to polarity register
 */
#define PIODAX_INTX_INVERSE     0x00    /* inverse polarity for all  */
#define PIODAX_INT0_NONINV      0x01    /* INT0 non-inverse polarity */
#define PIODAX_INT1_NONINV      0x02    /* INT1 non-inverse polarity */
#define PIODAX_INTX_NONINV      0x03    /* all non-inverse polarity  */

/* analog output ranges
 */
#define PIODAX_AO_PM10          (0x0000)     /* [ -10,  10 ] Volt */
#define PIODAX_AO_PM5           (0x0001)     /* [  -5,   5 ] Volt */
#define PIODAX_AO_0P10          (0x0002)     /* [   0,  10 ] Volt */
#define PIODAX_AO_0P5           (0x0003)     /* [   0,   5 ] Volt */
#define PIODAX_AO_PM10_ZERO     (0x7fff)     /* zero Volt for [ -10,  10 ] */
#define PIODAX_AO_PM5_ZERO      (0x7fff)     /* zero Volt for [  -5,   5 ] */
#define PIODAX_AO_0P10_ZERO     (0x0000)     /* zero Volt for [   0,  10 ] */
#define PIODAX_AO_0P5_ZERO      (0x0000)     /* zero Volt for [   0,   5 ] */

/* counter modes
 */
#define PIODAX_COUNTER_MODE0    0x00    /* interrupt on terminal count */
#define PIODAX_COUNTER_MODE1    0x02    /* programmable oneshot        */
#define PIODAX_COUNTER_MODE2    0x04    /* rate generator              */
#define PIODAX_COUNTER_MODE3    0x06    /* square wave generator       */
#define PIODAX_COUNTER_MODE4    0x08    /* software triggered pulse    */
#define PIODAX_COUNTER_MODE5    0x0a    /* hardware triggered pulse    */


/***********************************************************************
 * Control functions ...                                               *
 ***********************************************************************/

/* program interrupt with mask
 */
extern void piodax_int_select( unsigned int mask );

/* invert polarity
 */
extern unsigned int piodax_inv_polarity( void );

/* get number of interrupt line, irq handler can be defined outside
 */
extern int piodax_irq_get( void );

/* set counter "counter" with divisor and mode
 */
extern void piodax_counter_set( unsigned int counter, 
    unsigned int divisor, unsigned char mode );

/* start timer using coutners with "time" in nanosecs and mode!
 */
extern unsigned int piodax_timer16_start( unsigned int timer,
    unsigned int time, unsigned char mode );

/* stop timer using coutners
 */
extern void piodax_timer16_stop( unsigned int timer );

/* start timer using two coutners, with "time" in nanosecs and mode!
 */
extern unsigned int piodax_timer32_start( unsigned int timer,
    unsigned int time, unsigned char mode );

/* stop timer using coutners
 */
extern void piodax_timer32_stop( unsigned int timer );


/***********************************************************************
 * Analog output basic functions ...                                   *
 ***********************************************************************/

/* activate channel for successive DA conversion
 */
extern void piodax_ao_dac_select( unsigned short int channel );

/* program DAC for successive DA conversion with both channel 
 * and range as defined in header file.
 */
extern void piodax_ao_dac_program( unsigned short int channel,
    unsigned short int range );

/* set a single analog value given that channel is already
 * selected and ints range set value mapped to 16bit.
 */
extern void piodax_ao_set_value0( unsigned short int value );

/* set a single analog value (channel) given that its range set.
 * value is mapped to 16bit independently of DAC.
 */
extern void piodax_ao_set_value1( unsigned short int value, 
    unsigned short int channel );

/* set a single analog value (channel,range).
 * value is mapped to 16bit independently of DAC.
 */
extern void piodax_ao_set_value2( unsigned short int value,
    unsigned short int channel, unsigned short int range );


/***********************************************************************
 * Digital input/output basic functions ...                            *
 ***********************************************************************/

/* get byte from port "nport".
 */ 
extern unsigned char piodax_di_get_byte( unsigned short int nport );

/* get bit #"nbit" from port "nport".
 */ 
extern unsigned short int piodax_di_get_bit( unsigned short int nport, 
    unsigned short int nbit );

/* set byte "byte" on port "nport"
 */ 
extern void piodax_do_set_byte( unsigned char byte, 
    unsigned short int nport );

/* set bit "bit" with #"nbit" on port "nport".
 */ 
extern void piodax_do_set_bit( unsigned short int bit,
    unsigned short int nport, unsigned short int nbit );
